/**
  *  \file client/dialogs/exit.cc
  *  \brief Exit Confirmation Dialog
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/dialogs/exit.h"
#include "cpluslib/holder.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/msgbox.h"
#include "ui/widgets/button.h"
#include "ui/widgets/keytranslator.h"
#include "ui/widgets/statictext.h"
#include "util/nls.h"

namespace {
    class TwoLineListbox : public UIListbox {
     public:
        struct Item {
            int id;
            UIColor::Color bottomColor;
            string_t top;
            string_t bottom;
            Item(int id, UIColor::Color bottomColor, string_t top, string_t bottom)
                : id(id), bottomColor(bottomColor), top(top), bottom(bottom)
                { }
        };

        TwoLineListbox(const std::vector<Item>& content);
        virtual void drawPart(GfxCanvas& can, int from, int to);
        int getCurrentId() const;

     private:
        std::vector<Item> m_content;
    };

    int computeWidth(const std::vector<TwoLineListbox::Item>& content)
    {
        int max = 0;
        for (size_t i = 0; i < content.size(); ++i) {
            max = std::max(max, fonts[FONT_TITLE]->getTextWidth(content[i].top));
            max = std::max(max, fonts[FONT_NORMAL]->getTextWidth(content[i].bottom));
        }
        return max + 30;
    }
}

TwoLineListbox::TwoLineListbox(const std::vector<Item>& content)
    : UIListbox(0, int(content.size()), computeWidth(content), int(content.size())),
      m_content(content)
{
    setItemHeight(font_heights[FONT_TITLE] + font_heights[FONT_NORMAL]);
}

void
TwoLineListbox::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);

    const int EXTRA = 20;
    for (int i = from; i < to; ++i) {
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);
        size_t index = size_t(i + getTop());
        if (index < m_content.size()) {
            ctx.setColor(UIColor::tc_Static);
            ctx.useFont(*fonts[FONT_TITLE]);
            outTextF(ctx, area.x+EXTRA, area.y, area.w-EXTRA, m_content[i].top);

            if (index == size_t(getCurrent())) {
                int center = area.y + font_heights[FONT_TITLE]/2;
                for (int i = 0; i < 5; ++i) {
                    drawVLine(ctx, area.x + 7 + i, center - 4 + i, center + 3 - i);
                }
            }

            ctx.setColor(m_content[i].bottomColor);
            ctx.useFont(*fonts[FONT_NORMAL]);
            outTextF(ctx, area.x+EXTRA, area.y+font_heights[FONT_TITLE], area.w-EXTRA, m_content[i].bottom);
        }
    }
}

int
TwoLineListbox::getCurrentId() const
{
    size_t index = size_t(getCurrent());
    if (index < m_content.size()) {
        return m_content[index].id;
    } else {
        return 0;
    }
}

ExitResult
askExitConfirmation()
{
    std::vector<TwoLineListbox::Item> content;
    content.push_back(TwoLineListbox::Item(ExitNormally,      UIColor::tc_Faded, _("Save & Exit"),         _("Return to race selection")));
    content.push_back(TwoLineListbox::Item(ExitSaveOnly,      UIColor::tc_Faded, _("Save only"),           _("Save and keep playing")));
    content.push_back(TwoLineListbox::Item(ExitWithoutSaving, UIColor::tc_Red,   _("Exit without Saving"), _("Discard all changes")));

    ObjectHolder h;
    UIWindow& win = h.add(new UIWindow(_("Exit Game"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
    win.setLayout(UIVBoxLayout::instance5);

    win.add(h.add(new UIStaticText(_("Choose an action:"), UIStaticText::StaticNormal)));

    TwoLineListbox& box = h.add(new TwoLineListbox(content));
    win.add(box);

    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    g.add(h.add(new UISpacer()));
    g.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    g.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    g.add(h.add(new UISpacer()));
    win.add(g);
    win.add(h.add(new UIKeyTranslator()).addStop(' ', 1));
    win.pack();

    ui_root->centerWidget(win);
    ui_root->add(win);
    if (win.run() == 0) {
        return ExitCancelled;
    }

    ExitResult result = ExitResult(box.getCurrentId());
    if (result == ExitWithoutSaving) {
        if (!UIMessageBox(_("Do you really want to exit without saving?"), _("Exit Game")).doYesNoDialog()) {
            return ExitCancelled;
        }
    }

    return result;
}
