/**
  *  \file client/dialogs/hullsel.cc
  *  \brief Class WHullSelectionDialog
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/hullsel.h"
#include "client/widgets/helpwidget.h"
#include "cpluslib/format.h"
#include "game/racename.h"
#include "game/spec.h"
#include "game/struct.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/quit.h"
#include "util/nls.h"

WHullSelectionDialog::WHullSelectionDialog()
    : m_raceLimit(0),
      m_sorted(true),
      m_data(),
      m_list(0, 20, 20*font_ems[FONT_NORMAL], m_data),
      m_filterButton("#", '#'),
      m_filterText("", UIStaticText::StaticNormal)
{
    m_filterButton.setState(UIButton::bf_SmallSize, true);
    m_filterText.setFlexible(true);
    m_filterButton.sig_fired.add(this, &WHullSelectionDialog::onMenuButton);
    m_list.sig_menu.add(this, &WHullSelectionDialog::onMenuClick);
    m_list.setState(UIListbox::lb_KeyboardMenu, true);
}

bool
WHullSelectionDialog::run(string_t title, string_t help, int& current)
{
    // Window [VBox]
    //   UIFrameGroup [HBox]
    //     m_list
    //     UIScrollbar
    //   HBox
    //     m_filterButton
    //     m_filterText
    //   HBox
    //     UIButton "OK"
    //     UIButton "Cancel"
    //     UISpacer
    //     UIButton "Help"
    ObjectHolder h;

    UIWindow& w = h.add(new UIWindow(title, UIWindow::BLUE_WINDOW, COLOR_GRAY));
    w.setLayout(UIVBoxLayout::instance5);

    UIFrameGroup& g1 = h.add(new UIFrameGroup(UIHBoxLayout::instance0));
    g1.add(m_list);
    g1.add(h.add(new UIScrollbar(m_list)));
    w.add(g1);

    UIGroup& g2 = h.add(new UIGroup(UIHBoxLayout::instance5));
    g2.add(m_filterButton);
    g2.add(m_filterText);
    w.add(g2);

    UIGroup& g3 = h.add(new UIGroup(UIHBoxLayout::instance5));
    g3.add(h.add(new UIButton(1, _("OK"),     SDLK_RETURN)).addStop());
    g3.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    g3.add(h.add(new UISpacer()));
    g3.add(h.add(new UIButton(_("Help"), 'H')));
    w.add(g3);
    w.add(h.add(new WHelpWidget(help)));
    w.add(h.add(new UIQuit(0)));

    buildContent();
    w.pack();

    m_list.scrollToKey(current);
    m_list.setFocusedWidget();

    ui_root->centerWidget(w);
    bool ok = (ui_root->runModal(w) != 0);
    if (ok) {
        current = m_list.getKey();
    }
    return ok;
}

void
WHullSelectionDialog::onMenuButton()
{
    doMenu(m_filterButton);
}

void
WHullSelectionDialog::onMenuClick()
{
    doMenu(m_list);
}

void
WHullSelectionDialog::doMenu(UIBaseWidget& refWidget)
{
    UIStringList menu;
    menu.add(0, _("Show all ship types"));
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        menu.add(i, format(_("Show %s ship types"), player_racenames.getAdjName(i)));
    }
    menu.add(1000, _("Unsorted"));
    menu.add(1001, _("Sort by Name"));

    UIStandardListbox box(0, menu.getCount(), getTextWidth(*fonts[FONT_NORMAL], menu) + 20, menu);
    box.scrollToKey(m_raceLimit);
    if (box.doStandardMenu(refWidget)) {
        int raceLimit = m_raceLimit;
        bool sorted = m_sorted;
        int choice = box.getKey();
        if (choice >= 0 && choice <= NUM_PLAYERS) {
            raceLimit = choice;
        } else if (choice == 1000) {
            sorted = false;
        } else {
            sorted = true;
        }
        if (raceLimit != m_raceLimit || sorted != m_sorted) {
            m_raceLimit = raceLimit;
            m_sorted = sorted;
            buildContent();
        }
    }
}

void
WHullSelectionDialog::buildContent()
{
    // Remember old selection
    int oldSelection = m_data.getCount() == 0 ? 0 : m_list.getKey();

    // New list content
    m_data = UIStringList();
    if (m_raceLimit == 0) {
        for (int i = 1; i <= getNumHulls(); ++i) {
            m_data.add(i, getHull(i).getName());
        }
    } else {
        for (int i = 1; i <= NUM_HULLS_PER_PLAYER; ++i) {
            if (int h = getTruehull(m_raceLimit, i)) {
                m_data.add(h, getHull(h).getName());
            }
        }
    }
    if (m_sorted) {
        m_data.sortAlphabetically();
    }
    if (m_raceLimit == 0) {
        m_data.addAtFront(0, _("Custom Ship"));
    }

    // Update the list box
    m_list.onContentChange();
    m_list.scrollToKey(oldSelection);

    // Label
    string_t label;
    if (m_raceLimit == 0) {
        label = _("all ship types");
    } else {
        label = format(_("%s ship types"), player_racenames.getAdjName(m_raceLimit));
    }
    if (m_sorted) {
        label += _(", sorted");
    }
    m_filterText.setText(label);
}

