/**
  *  \file client/tiles/shipequipment.cc
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "client/tiles/shipequipment.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "game/ship.h"
#include "game/spec.h"
#include "ui/uiglobal.h"
#include "util/misc.h"
#include "game/config.h"
#include "game/cmds.h"
#include "game/global.h"
#include "ui/util.h"
#include "client/remote.h"
#include "game/host.h"
#include "client/wcolorframe.h"
#include "ui/widgets/button.h"
#include "game/format.h"
#include "util/unicodechars.h"
#include "game/planet.h"
#include "game/mission.h"

/*************************** WShipEquipmentTile **************************/

WShipEquipmentTile::WShipEquipmentTile(GObjectSelection& selection, bool forHistory)
    : WObjectSelectionObserverTile(GfxPoint(315, 6*font_heights[FONT_NORMAL]+3), selection),
      cf_r(0),
      cf_g(0),
      cf_s(0),
      forHistory(forHistory)
{
    setLayout(RightJustified, 0, false);

    /* Some buttons: */
    int x = 0;
    if (!forHistory && host.isPHost() && config.CPEnableRemote()) {
        /* Remote control */
        cf_r = &h.add(new WColorFrame());
        cf_r->add(h.add(new UIButton("R", 'r')));
        addControl(GfxRect(x, 0, 24, 24), *cf_r);
        x += 25;
    }
    if (!forHistory && host.isPHost() && config.CPEnableGive()) {
        /* Give ship */
        cf_g = &h.add(new WColorFrame());
        cf_g->add(h.add(new UIButton("G", 'g')));
        addControl(GfxRect(x, 0, 24, 24), *cf_g);
        x += 25;
    }

    /* Ship spec sheet */
    cf_s = &h.add(new WColorFrame());
    cf_s->add(h.add(new UIButton("S", 's')));
    addControl(GfxRect(x, 0, 24, 24), *cf_s);

    /* Cargo */
    if (forHistory) {
        addControl(GfxRect(x+2, 24, 20, 20), h.add(new UIButton("C", 'c')));
    }
}

WShipEquipmentTile::~WShipEquipmentTile()
{ }

void
WShipEquipmentTile::drawShipEquipmentTile(UIAbstractTableWriter& tw)
{
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Engines:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Primary:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Secondary:"));
    tw.writeFixedCell(UIColor::tc_Static, string_t());
    tw.writeFixedCell(UIColor::tc_Static, _("Damage:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Crew:"));

    tw.newLColumn();

    const GShip* s = dynamic_cast<GShip*>(getCurrentObject());
    if (!s || !s->getHullId().isKnown())
        return;

    const GShip& ship = *s;
    const GHull& hull = ship.getHull();

    UIColor::Color equipColor = (ship.getHistoryTimestamp(sts_ArmsDamage) == getCurrentTurn().getTurnNumber() ? UIColor::tc_Green : UIColor::tc_Yellow);
    const char* unknownLabel = forHistory ? _("unknown") : "";

    /* engines */
    if (ship.getEngineType().isKnown()) {
        string_t en = getEngine(ship.getEngineType()).getName();
        if (ship.canDoSpecial(hf_HardenedEngines))
            en = format(_("Hard. %s"), en);
        if (ship.canDoSpecial(hf_Gravitonic))
            en = format(_("Grav. %s"), en);
        if (hull.getNumEngines() > 1)
            en = format("%d " UTF_TIMES " %s", hull.getNumEngines(), en);
        tw.writeVariableCell(equipColor, en);
    } else {
        tw.writeVariableCell(UIColor::tc_Yellow, unknownLabel);
    }

    /* primary */
    if (ship.getNumBeams().isKnown() && ship.getBeamType().isKnown()) {
        if (ship.getNumBeams() && ship.getBeamType()) {
            string_t bm = getBeam(ship.getBeamType()).getName();
            if (ship.getNumBeams() != 1)
                bm = format("%d " UTF_TIMES " %s", ship.getNumBeams(), bm);
            tw.writeVariableCell(equipColor, bm);
        } else {
            tw.writeVariableCell(equipColor, _("none"));
        }
    } else {
        if (hull.getMaxBeams() > 0) {
            tw.writeVariableCell(UIColor::tc_Yellow, format(_("up to %d beam%!1{s%}"), hull.getMaxBeams()));
        } else if (hull.getMaxBeams() == 0) {
            tw.writeVariableCell(UIColor::tc_Yellow, _("none"));
        } else {
            tw.writeVariableCell(UIColor::tc_Yellow, unknownLabel);
        }
    }

    /* secondary */
    if (!(ship.getTorpType().isKnown() && ship.getNumTorpLaunchers().isKnown() && ship.getNumBays().isKnown())) {
        if (hull.getMaxLaunchers() > 0) {
            tw.writeVariableCell(UIColor::tc_Yellow, format(_("up to %d torpedo launcher%!1{s%}"), hull.getMaxLaunchers()));
            if (ship.getAmmoRaw().isKnown()) {
                string_t tc = (ship.getAmmoRaw()
                               ? string_t(format(_("%d torpedo%!1{es%} aboard"), numToString(ship.getAmmoRaw())))
                               : string_t(_("no torpedoes aboard")));
                tw.writeVariableCell(UIColor::tc_Yellow, tc);
            } else {
                tw.writeVariableCell(UIColor::tc_Yellow, unknownLabel);
            }
        } else if (hull.getNumBays() > 0) {
            tw.writeVariableCell(UIColor::tc_Yellow, format(_("%d fighter bay%!1{s%}"), hull.getNumBays()));
            if (ship.getAmmoRaw().isKnown()) {
                string_t tc = (ship.getAmmoRaw()
                               ? string_t(format(_("%d fighter%!1{s%} ready"), numToString(ship.getAmmoRaw())))
                               : string_t(_("no fighters aboard")));
                tw.writeVariableCell(UIColor::tc_Yellow, tc);
            } else {
                tw.writeVariableCell(UIColor::tc_Yellow, unknownLabel);
            }
        } else {
            tw.writeVariableCell(UIColor::tc_Yellow, _("none"));
            tw.writeVariableCell(UIColor::tc_Green, string_t());
        }
    } else if (ship.getTorpType() && ship.getNumTorpLaunchers()) {
        string_t tl = format(_("%d \xC3\x97 %s launcher%!1{s%}"),
                             ship.getNumTorpLaunchers(),
                             getTorp(ship.getTorpType()).getName());
        string_t tc = (ship.getAmmoRaw().isKnown()
                       ? (ship.getAmmoRaw()
                          ? string_t(format(_("%d torpedo%!1{es%} aboard"), numToString(ship.getAmmoRaw())))
                          : string_t(_("no torpedoes aboard")))
                       : string_t());
        tw.writeVariableCell(equipColor, tl);
        tw.writeVariableCell(equipColor, tc);
    } else if (ship.getNumBays()) {
        string_t tl = format(_("%d fighter bay%!1{s%}"), ship.getNumBays());
        string_t tc = (ship.getAmmoRaw().isKnown()
                       ? (ship.getAmmoRaw()
                          ? string_t(format(_("%d fighter%!1{s%} ready"), numToString(ship.getAmmoRaw())))
                          : string_t(_("no fighters aboard")))
                       : string_t());
        tw.writeVariableCell(equipColor, tl);
        tw.writeVariableCell(equipColor, tc);
    } else {
        tw.writeVariableCell(equipColor, _("none"));
        tw.writeVariableCell(equipColor, string_t());
    }

    /* damage */
    bool fixed = false;
    if (ship.getOrbitPlanetId() != 0) {
        const GPlanet& pl = getCurrentUniverse()->getPlanet(ship.getOrbitPlanetId());
        if (pl.isPlayable(GObject::Playable)
            && pl.hasFullBaseData()
            && pl.getBaseShipyardOrder() == sya_Fix
            && pl.getBaseShipyardId() == ship.getId())
        {
            fixed = true;
        }
    }

    if (!ship.getDamage().isKnown()) {
        tw.writeVariableCell(UIColor::tc_Yellow, unknownLabel);
    } else if (ship.getDamage()) {
        if (fixed) {
            tw.writeVariableCell(UIColor::tc_Yellow, format(_("%d%% (fixed by base)"), ship.getDamage()));
        } else {
            int i = ship.getCargo(el_Supplies).isKnown() ? ship.getCargo(el_Supplies) / 5 : 0;
            if (raceMissionId(ship.getOwner()) == 6
                && ship.getMission().isKnown()
                && (ship.getMission() == GMission::msn_Special
                    || (host.isPHost() && ship.getMission() == config.ExtMissionsStartAt() + GMission::pmsn_Special)))
            {
                // Borg self repair
                i += 10;
            }
            int j = ship.getDamage() - i;
            UISkin::Color c = UIColor::tc_Red;
            if (j <= 0) {
                j = 0;
                c = UIColor::tc_Yellow;
            }
            if (i) {
                tw.writeVariableCell(c, format("%d%% (-%d: %d%%)", ship.getDamage(), i, j));
            } else {
                tw.writeVariableCell(c, format("%d%%", ship.getDamage()));
            }
        }
    } else {
        tw.writeVariableCell(equipColor, _("none"));
    }

    /* crew */
    UIColor::Color c;
    string_t line;
    if (!ship.getCrew().isKnown()) {
        c = UIColor::tc_Yellow;
        line = format(_("up to %d"), numToString(hull.getCrew()));
    } else if (ship.getCrew() != hull.getCrew()) {
        c = (ship.getCrew() > hull.getCrew() || fixed) ? UIColor::tc_Yellow : UIColor::tc_Red;
        line = format(_("%d of %d"), numToString(ship.getCrew()), numToString(hull.getCrew()));
    } else {
        c = equipColor;
        line = numToString(ship.getCrew());
    }

    int16_t exp = ship.unit_scores.getScore(ship_score_definitions.lookupScore(ScoreId_ExpPoints));
    if (exp >= 0) {
        if (ship.getMission().isKnown() && ship.getMission() == config.ExtMissionsStartAt() + GMission::pmsn_Training) {
            // Ship is attempting to train
            int sup = ship.getInterceptId().isKnown() ? ship.getInterceptId() : 0;
            double points = sup < 25 ? sup : 25 + std::sqrt(8.0 * (sup - 25));
            int32_t acaBonus = ship.canDoSpecial(hf_Academy) ? config.EPAcademyScale(ship.getRealOwner()) : 100;
            int32_t rate = config.EPTrainingScale(ship.getRealOwner()) * acaBonus / 100;
            int32_t added = int32_t(rate * points / (std::sqrt(double(hull.getCrew())) + 1));
            line += format(_(" (%d EP, +%d)"), numToString(exp), numToString(added));

            // Warning for mission failure if...
            // ...not at a planet
            // ...not enough supplies on played planet
            int pid = ship.getOrbitPlanetId();
            GUniverse* univ = getCurrentUniverse();
            if (univ == 0
                || !univ->isValidPlanetId(pid)
                || (univ->ty_played_planets.isValidIndex(pid)
                    && sup > univ->getPlanet(pid).getCargoRaw(el_Supplies)))
            {
                if (c == UIColor::tc_Green) {
                    c = UIColor::tc_Yellow;
                }
            }
        } else {
            // Normal ship
            line += format(_(" (%d EP)"), numToString(exp));
        }
    }

    tw.writeVariableCell(c, line);
    tw.finish();
}

void
WShipEquipmentTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getInnerSize(5, 0), FONT_NORMAL);
    drawShipEquipmentTile(tw);

    /* Timestamp */
    if (forHistory) {
        ctx.useFont(*fonts[FONT_NORMAL]);
        ctx.setSolidBackground();
        ctx.setTextAlign(2, 0);

        const GShip* s = dynamic_cast<GShip*>(getCurrentObject());
        string_t info;
        if (s == 0 || s->getHistoryTimestamp(sts_ArmsDamage) <= 0) {
            ctx.setColor(UIColor::tc_Green);
        } else {
            ctx.setColor(s->getHistoryTimestamp(sts_ArmsDamage) == getCurrentTurn().getTurnNumber()
                         ? UIColor::tc_Green
                         : UIColor::tc_Yellow);
            info = formatTurnNumber(s->getHistoryTimestamp(sts_ArmsDamage));
        }

        GfxRect is = getInnerSize(5, 0);
        outTextF(ctx, is.w, is.h - font_heights[FONT_NORMAL], font_ems[FONT_NORMAL]*9, info);
    }

    /* Some frames: */
    GShip* sh = dynamic_cast<GShip*>(getCurrentObject());
    if (sh != 0) {
        GCommandContainer& con = getDisplayedTurn().getCommands(sh->getOwner());
        if (cf_g != 0) {
            /* "give" button */
            cf_g->setColor((con.getCommand(GCommand::phc_GiveShip, sh->getId()) != 0
                            ? cf_Yellow
                            : cf_None));
        }
        if (cf_r != 0) {
            /* "remote" button */
            cf_r->setColor(getRemoteControlColor(getNewRemoteControlState(*sh, con)));
        }
        if (cf_s != 0) {
            /* "spec" button */
            cf_s->setColor(sh->hasAnyShipSpecificSpecials() ? cf_Green : cf_None);
        }
    } else {
        /* No ship selected */
        if (cf_g != 0)
            cf_g->setColor(cf_None);
        if (cf_r != 0)
            cf_r->setColor(cf_None);
        if (cf_s != 0)
            cf_s->setColor(cf_None);
    }
}

/************************ WNarrowShipEquipmentTile ***********************/

WNarrowShipEquipmentTile::WNarrowShipEquipmentTile(GObjectSelection& selection)
    : WObjectSelectionObserverTile(GfxPoint(250, 5 * font_heights[FONT_NORMAL]), selection)
{ }

void
WNarrowShipEquipmentTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setSolidBackground();

    // Fetch the ship
    GShip* sh = dynamic_cast<GShip*>(getCurrentObject());
    if (!sh) {
        drawBackground(ctx, getInnerSize(0, 0));
        return;
    }

    // Coordinates
    int x = 0;
    int y = 0;
    int w = getInnerSize(0, 0).w;
    int dy = font_heights[FONT_NORMAL];

    // Engines
    if (sh->getEngineType().isKnown()) {
        string_t en = getEngine(sh->getEngineType()).getName();
        if (sh->canDoSpecial(hf_HardenedEngines))
            en = format(_("Hard. %s"), en);
        if (sh->canDoSpecial(hf_Gravitonic))
            en = format(_("Grav. %s"), en);
        if (sh->getHullId().isKnown())
            if (sh->getHull().getNumEngines() > 1)
                en = format("%d " UTF_TIMES " %s", sh->getHull().getNumEngines(), en);
        ctx.setColor(UIColor::tc_Green);
        outTextF(ctx, x, y, w, en);
    } else if (sh->getHullId().isKnown()) {
        string_t en;
        int n = sh->getHull().getNumEngines();
        if (sh->canDoSpecial(hf_Gravitonic))
            en = format(_("%d grav. engine%!1{s%}"), n);
        else
            en = format(_("%d engine%!1{s%}"), n);
        ctx.setColor(UIColor::tc_Static);
        outTextF(ctx, x, y, w, en);
    } else {
        ctx.setColor(UIColor::tc_Static);
        outTextF(ctx, x, y, w, string_t());
    }
    y += dy;

    // Beams
    if (sh->getBeamType().isKnown()) {
        if (sh->getBeamType() == 0 || (sh->getNumBeams().isKnown() && sh->getNumBeams() == 0)) {
            ctx.setColor(UIColor::tc_Green);
            outTextF(ctx, x, y, w, _("no beam weapons"));
        } else if (sh->getNumBeams().isKnown()) {
            // type and count
            ctx.setColor(UIColor::tc_Green);
            outTextF(ctx, x, y, w, format(_("%!d%!1{%0$d \xC3\x97 %}%1$s"), sh->getNumBeams(), getBeam(sh->getBeamType()).getName()));
        } else if (sh->getHullId().isKnown()) {
            // just type
            ctx.setColor(UIColor::tc_Static);
            outTextF(ctx, x, y, w, format(_("up to %d %s"), sh->getHull().getMaxBeams(), getBeam(sh->getBeamType()).getName()));
        } else {
            ctx.setColor(UIColor::tc_Static);
            outTextF(ctx, x, y, w, getBeam(sh->getBeamType()).getName());
        }
    } else {
        if (sh->getHullId().isKnown()) {
            ctx.setColor(UIColor::tc_Static);
            if (sh->getHull().getMaxBeams() == 0) {
                outTextF(ctx, x, y, w, _("no beam weapons"));
            } else {
                outTextF(ctx, x, y, w, format(_("up to %d beam%!1{s%}"), sh->getHull().getMaxBeams()));
            }
        } else {
            ctx.setColor(UIColor::tc_Static);
            outTextF(ctx, x, y, w, string_t());
        }
    }
    y += dy;

    // Torps/Fighters
    if (sh->getNumTorpLaunchers().isKnown() && sh->getNumTorpLaunchers() > 0) {
        // Ship has torpedoes
        if (sh->getTorpType().isKnown()) {
            ctx.setColor(UIColor::tc_Green);
            if (sh->getAmmoRaw().isKnown()) {
                outTextF(ctx, x, y, w, format(_("%!d%!1{%0$d \xC3\x97 %}%1$s, %d torpedo%!1{es%}"),
                                              sh->getNumTorpLaunchers(),
                                              getTorp(sh->getTorpType()).getName(),
                                              numToString(sh->getAmmoRaw())));
            } else {
                outTextF(ctx, x, y, w, format(_("%!d%!1{%0$d \xC3\x97 %}%1$s"),
                                              sh->getNumTorpLaunchers(),
                                              getTorp(sh->getTorpType()).getName()));
            }
        } else {
            ctx.setColor(UIColor::tc_Static);
            if (sh->getAmmoRaw().isKnown()) {
                outTextF(ctx, x, y, w, format(_("%d launcher%!1{s%}, %d torpedo%!1{es%}"),
                                              sh->getNumTorpLaunchers(),
                                              numToString(sh->getAmmoRaw())));
            } else {
                outTextF(ctx, x, y, w, format(_("%d torpedo launcher%!1{s%}"),
                                              sh->getNumTorpLaunchers()));
            }
        }
    } else if (sh->getNumBays().isKnown() && sh->getNumBays() > 0) {
        // Ship has fighters
        ctx.setColor(UIColor::tc_Green);
        if (sh->getAmmoRaw().isKnown()) {
            outTextF(ctx, x, y, w, format(_("%d bay%!1{s%}, %d fighter%!1{s%}"),
                                          sh->getNumBays(),
                                          numToString(sh->getAmmoRaw())));
        } else {
            outTextF(ctx, x, y, w, format(_("%d fighter bay%!1{s%}"),
                                          sh->getNumBays()));
        }
    } else if (sh->getNumTorpLaunchers().isKnown() && sh->getNumBays().isKnown()) {
        // No torpedo launchers or fighters
        ctx.setColor(UIColor::tc_Green);
        outTextF(ctx, x, y, w, _("no secondary weapons"));
    } else if (sh->getHullId().isKnown()) {
        // Hull is known
        ctx.setColor(UIColor::tc_Static);
        if (sh->getHull().getMaxLaunchers() > 0) {
            outTextF(ctx, x, y, w, format(_("up to %d torpedo launcher%!1{s%}"), sh->getHull().getMaxLaunchers()));
        } else if (sh->getHull().getNumBays() > 0) {
            outTextF(ctx, x, y, w, format(_("%d fighter bay%!1{s%}"),
                                          sh->getHull().getNumBays()));
        } else {
            outTextF(ctx, x, y, w, _("no secondary weapons"));
        }
    } else {
        ctx.setColor(UIColor::tc_Static);
        outTextF(ctx, x, y, w, string_t());
    }
    y += dy;

    // Crew [of Crew]
    if (sh->getCrew().isKnown()) {
        if (sh->getHullId().isKnown() && sh->getCrew() != sh->getHull().getCrew()) {
            ctx.setColor(UIColor::tc_Red);
            outTextF(ctx, x, y, w, format(_("%d crew (of %d)"), numToString(sh->getCrew()), numToString(sh->getHull().getCrew())));
        } else {
            ctx.setColor(UIColor::tc_Green);
            outTextF(ctx, x, y, w, format(_("%d crew"), numToString(sh->getCrew())));
        }
    } else if (sh->getHullId().isKnown()) {
        ctx.setColor(UIColor::tc_Static);
        outTextF(ctx, x, y, w, format(_("up to %d crew"), numToString(sh->getHull().getCrew())));
    } else {
        ctx.setColor(UIColor::tc_Static);
        outTextF(ctx, x, y, w, string_t());
    }
    y += dy;

    // Damage
    if (sh->getDamage().isKnown()) {
        if (sh->getDamage() == 0)
            ctx.setColor(UIColor::tc_Green);
        else
            ctx.setColor(UIColor::tc_Red);
        outTextF(ctx, x, y, w, format(_("%d%% damage"), sh->getDamage()));
    } else {
        ctx.setColor(UIColor::tc_Static);
        outTextF(ctx, x, y, w, string_t());
    }
    y += dy;
}
