/**
  *  \file client/widgets/chartlockconfig.cc
  *  \brief WChartLockConfig widget
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/chartlockconfig.h"
#include "cpluslib/assert.h"
#include "util/nls.h"
#include "resmgr/resmgr.h"
#include "game/pref.h"

namespace {
    struct Option {
        uint16_t    flag;
        const char* name;
    };

    // Options for each mode
    const Option flags[] = {
        { li_Planet,       N_("Planets") },
        { li_Ship,         N_("Ships") },
        { li_Ufo,          N_("Ufos") },
        { li_Marker,       N_("Markers") },
        { li_Minefield,    N_("Minefields") }
    };

    // Modes (corresponds to TLockMode)
    enum Mode {
        MODE_LEFT,
        MODE_RIGHT,
        MODE_WHEEL
    };

    const char*const WHEEL_MODES[] = {
        N_("Zoom (+/-)"),
        N_("Browse list (Tab)"),
        N_("Cycle units (PgUp/PgDn)"),
    };
    CTASSERT(countof(WHEEL_MODES) == GUserPreferences::cw_MAX+1);

    const char*const modes[] = {
        N_("Left mouse button / space bar locks on..."),
        N_("Right mouse button / Enter locks on..."),
    };
    CTASSERT(lm_Left == 0);
    CTASSERT(lm_Right == 1);
}

/*
 *  Tree node Ids. A tree node has a 32-bit Id. This folds two pieces of information
 *  into one such value:
 *  - the mode index into the modes[]/opts[] array
 *  - the sub-index for the actual flag:
 *      0 - node
 *      1,2,3 - index into flags[], plus one, or wheel mode, plus one
 */

static inline int32_t
makePair(Mode mode, uint_t sub)
{
    return (mode << 5) + sub;
}

static inline Mode
getModeFromPair(int32_t pair)
{
    return Mode(pair >> 5);
}

static inline uint_t
getSubIndexFromPair(int32_t pair)
{
    return (pair & 31);
}

/**************************** WChartLockConfig ***************************/

/** Constructor.
    \param id Widget id */
WChartLockConfig::WChartLockConfig(int id)
    : UITreeWidget(id, 16 /*countof(modes)*(countof(flags)+1) + 1+countof(WHEEL_MODES)*/, 25*font_ems[FONT_NORMAL])
{
    init();
}

WChartLockConfig::~WChartLockConfig()
{ }

void
WChartLockConfig::drawItem(GfxContext& ctx, GfxRect area, const Item& it)
{
    UITreeWidget::drawItem(ctx, area, it);
    if (area.w >= 20 && area.h >= 20) {
        uint_t mode = getModeFromPair(it.id);
        uint_t sub = getSubIndexFromPair(it.id);
        switch (mode) {
         case MODE_LEFT:
         case MODE_RIGHT:
            if (sub != 0) {
                bool value = (opts[mode] & flags[sub-1].flag) != 0;
                Ptr<GfxPixmap> pix = requestPixmap(ResId("cb", 1+value));
                if (pix) {
                    pix->blit(ctx.getCanvas(), area.x + area.w - 18, area.y + area.h/2 - 8);
                }
            }
            break;

         case MODE_WHEEL:
            if (sub != 0) {
                bool value = (wheel == int(sub-1));
                Ptr<GfxPixmap> pix = requestPixmap(ResId("radio", 1+value));
                if (pix) {
                    pix->blit(ctx.getCanvas(), area.x + area.w - 20, area.y + area.h/2 - 10);
                }
            }
            break;
        }
    }
}

bool
WChartLockConfig::handleEvent(const UIEvent& event, bool second_pass)
{
    if (hasState(st_Focused) && event.type == ev_Key && event.key.code == ' ') {
        onClick(getItemId(getCurrentItemIndex()));
        return true;
    } else {
        return UITreeWidget::handleEvent(event, second_pass);
    }
}

void
WChartLockConfig::onItemClick(int item, int x)
{
    if (x >= getExtent().w - 20) {
        onClick(getItemId(getCurrentItemIndex()));
    } else {
        UITreeWidget::onItemClick(item, x);
    }
}

void
WChartLockConfig::load()
{
    for (int i = 0; i < lm_MAX; ++i) {
        opts[i] = getLockMode(TLockMode(i));
    }
    wheel = getUserPreferences().ChartWheel();
}

void
WChartLockConfig::save()
{
    for (int i = 0; i < lm_MAX; ++i) {
        setLockMode(TLockMode(i), opts[i]);
    }
    getUserPreferences().ChartWheel.set(wheel);
}

void
WChartLockConfig::init()
{
    // Content
    for (int i = 0; i < lm_MAX; ++i) {
        opts[i] = 0;
    }
    wheel = 0;

    // Configure list
    setItemHeight(std::max(font_heights[FONT_NORMAL] + 4, 20));

    // Configure tree
    for (uint_t i = 0; i < lm_MAX; ++i) {
        addItem(makePair(Mode(i), 0), 0, _(modes[i]), true);
        for (uint_t j = 0; j < countof(flags); ++j) {
            addItem(makePair(Mode(i), j+1), 1, _(flags[j].name));
        }
    }
    addItem(makePair(MODE_WHEEL, 0), 0, _("Mouse wheel does..."), true);
    for (size_t i = 0; i < countof(WHEEL_MODES); ++i) {
        addItem(makePair(MODE_WHEEL, i+1), 1, _(WHEEL_MODES[i]));
    }

    updateAfterModification();
}

void
WChartLockConfig::onClick(int32_t id)
{
    Mode mode = getModeFromPair(id);
    uint_t sub  = getSubIndexFromPair(id);
    if (sub != 0) {
        switch (mode) {
         case MODE_LEFT:
         case MODE_RIGHT:
            opts[mode] ^= flags[sub-1].flag;
            drawWidget();
            break;

         case MODE_WHEEL:
            wheel = sub-1;
            drawWidget();
            break;
        }
    }
}
