/**
  *  \file game/bum.cc
  *  \brief Beam up Multiple
  *
  *  Beam up multiple is implemented by using two custom transfer
  *  objects, completely unrelated to the normal cargo transfer
  *  partners. The planet half contains 10000 (minus existing command)
  *  of everything; the ship half contains 0 (plus existing command).
  *  Display is adjusted to actual content using display offsets, so
  *  users can (and should) easily overdraw.
  *
  *  In PCC 1.x, the BUM halves are a little closer related, making things
  *  a little more complex but sharing the "init from planet" part.
  *  PCC 1.x also has a "beam up+down" mode which is a cross of BUM
  *  and a regular cargo unload; I believe this feature to be very
  *  hard to understand, so I left it out here.
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/bum.h"
#include "game/ship.h"
#include "game/planet.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "game/cmds.h"
#include "game/cspec.h"
#include "game/global.h"
#include "game/config.h"
#include "game/mission.h"
#include "game/fleet.h"
#include "game/actions/preconditions.h"

/** Parse "beam up multiple" command.
    This will find a command for the specified ship, and if one exists,
    add its cargo to %counts, multiplied by %factor.
    \param trn      [in] Turn to operate from
    \param counts   [in/out] Produce result here
    \param sid      [in] Ship Id
    \param factor   [in] */
static void
parseBumCommand(GGameTurn& trn, int32_t counts[], const int sid, const int factor)
{
    int shipOwner = trn.getCurrentUniverse().getShip(sid).getOwner();
    const GCommand* cmd = trn.getCommands(shipOwner).getCommand(GCommand::phc_Beamup, sid);
    if (!cmd)
        return;

    GCargoSpec sp(cmd->getArg(), true);
    counts[el_Neutronium] += factor * sp.get(el_Neutronium);
    counts[el_Tritanium]  += factor * sp.get(el_Tritanium);
    counts[el_Duranium]   += factor * sp.get(el_Duranium);
    counts[el_Molybdenum] += factor * sp.get(el_Molybdenum);
    counts[el_Colonists]  += factor * sp.get(el_Colonists);
    counts[el_Supplies]   += factor * sp.get(el_Supplies);
    counts[el_Money]      += factor * sp.get(el_Money);
}

/**************************** GShipBumTransfer ***************************/

/** Constructor.
    \param trn Turn to operate from
    \param sid Ship Id */
GShipBumTransfer::GShipBumTransfer(GGameTurn& trn, int sid)
    : trn(trn),
      sid(sid)
{
    mustBePlayed(trn.getCurrentUniverse().getShip(sid));
    init();
}

GShipBumTransfer::~GShipBumTransfer()
{
    if (!isFinished()) {
        doCancel();
    }
}

bool
GShipBumTransfer::canHaveCargo(GCargoType cargo_id)
{
    switch (cargo_id) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
     case el_Money:
     case el_Supplies:
        return true;
     case el_Colonists:
        return config.AllowBeamUpClans();
     default:
        return false;
    }
}

int32_t
GShipBumTransfer::getMaxCargo(GCargoType cargo_id)
{
    const GShip& ship = trn.getCurrentUniverse().getShip(sid);
    const GHull& hull = ship.getHull();
    if (hasFlag(can_Overload)) {
        return 10000;
    } else if (cargo_id == el_Neutronium) {
        return hull.getFuel() - ship.getCargo(el_Neutronium);
    } else if (cargo_id == el_Money) {
        return 10000 - ship.getCargo(el_Money);
    } else {
        int32_t result = ship.getFreeCargo();
        for (int i = 0; i < el_COUNT; ++i)
            if (i != el_Neutronium && i != el_Money && i != cargo_id)
                result -= getCargo(GCargoType(i));
        return result;
    }
}

int32_t
GShipBumTransfer::getDisplayOffset(GCargoType cargo_id)
{
    return display_offsets[cargo_id];
}

string_t
GShipBumTransfer::getName() const
{
    return trn.getCurrentUniverse().getShip(sid).getName(GObject::PlainName);
}

int32_t
GShipBumTransfer::getCargoOnObject(GCargoType type)
{
    return have[type] - reserved[type];
}

void
GShipBumTransfer::changeCargoOnObject(GCargoType type, int32_t amount)
{
    have[type] += amount;
}

void
GShipBumTransfer::changeReservedCargo(GCargoType type, int32_t amount)
{
    reserved[type] += amount;
}

void
GShipBumTransfer::commit()
{
    GCargoContainer::commit();

    /* Now we have the cargo amount to transfer in have */
    GCargoSpec cs;
    for (int i = 0; i < el_COUNT; ++i) {
        if (cs.isValidType(GCargoType(i))) {
            cs.set(GCargoType(i), have[i]);
        }
    }

    GShip& ship = trn.getCurrentUniverse().getShip(sid);
    const int bumMission = config.ExtMissionsStartAt() + GMission::pmsn_BeamUpMultiple;
    GCommandContainer& cmds = trn.getCommands(ship.getOwner());
    if (cs.isEmpty()) {
        cmds.removeCommand(GCommand::phc_Beamup, sid);
        if (ship.getMission() == bumMission) {
            /* Ship has mission "Beam up multiple". Try to set it away */
            const GShip& oldShip = trn.getPreviousUniverse().getShip(sid);
            if (oldShip.getShipKind() != GShip::CurrentShip
                || oldShip.getMission() == bumMission
                || !setFleetMission(trn.getCurrentUniverse(), sid,
                                    oldShip.getMission(),
                                    oldShip.getInterceptId(),
                                    oldShip.getTowId()))
            {
                /* oldShip is not a current ship (should not happen),
                   it also has a "Beam up multiple" mission,
                   or resetting failed (maybe because the ship started
                   the turn with Intercept and is now member of a fleet),
                   so clear it. */
                setFleetMission(trn.getCurrentUniverse(), sid, 0, 0, 0);
            }
        }
    } else {
        cmds.setCommand(GCommand::phc_Beamup, sid, cs.toPHostString());
        setFleetMission(trn.getCurrentUniverse(), sid, bumMission, 0, 0);
    }
    ship.markDirty();
}

void
GShipBumTransfer::init()
{
    /* Initialize */
    GShip& ship = trn.getCurrentUniverse().getShip(sid);
    for (int i = 0; i < el_COUNT; ++i) {
        have[i] = 0;
        display_offsets[i] = 0;
        reserved[i] = 0;
        if (canHaveCargo(GCargoType(i))) {
            display_offsets[i] += ship.getCargo(GCargoType(i));
        }
    }

    /* Add existing command message */
    parseBumCommand(trn, have, sid, +1);

    setFlags(can_Unload);
}

/*************************** GPlanetBumTransfer **************************/

/** Constructor.
    \param trn Turn to operate from
    \param pl  Planet
    \param sid Ship Id */
GPlanetBumTransfer::GPlanetBumTransfer(GGameTurn& trn, GPlanet& pl, int sid)
    : trn(trn),
      pl(pl)
{
    /* No need to verify the ship; we don't access it.
       It will be verified by the other half (GShipBumTransfer). */
    init(sid);
}

GPlanetBumTransfer::~GPlanetBumTransfer()
{
    if (!isFinished()) {
        doCancel();
    }
}

bool
GPlanetBumTransfer::canHaveCargo(GCargoType cargo_id)
{
    switch (cargo_id) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
     case el_Money:
     case el_Supplies:
        return true;
     case el_Colonists:
        return config.AllowBeamUpClans();
     default:
        return false;
    }
}

int32_t
GPlanetBumTransfer::getMaxCargo(GCargoType /*cargo_id*/)
{
    return 10000;
}

int32_t
GPlanetBumTransfer::getDisplayOffset(GCargoType cargo_id)
{
    return display_offsets[cargo_id];
}

string_t
GPlanetBumTransfer::getName() const
{
    return format(_("Beam up from %s"), pl.getName(GObject::PlainName));
}

int32_t
GPlanetBumTransfer::getCargoOnObject(GCargoType type)
{
    return have[type] - reserved[type];
}

void
GPlanetBumTransfer::changeCargoOnObject(GCargoType type, int32 amount)
{
    have[type] += amount;
}

void
GPlanetBumTransfer::changeReservedCargo(GCargoType type, int32 amount)
{
    reserved[type] += amount;
}

void
GPlanetBumTransfer::init(int sid)
{
    /* Initialize */
    for (int i = 0; i < el_COUNT; ++i) {
        have[i] = 10000;
        display_offsets[i] = -10000;
        reserved[i] = 0;
        if (canHaveCargo(GCargoType(i)) && pl.getCargo(GCargoType(i)).isKnown()) {
            display_offsets[i] += pl.getCargo(GCargoType(i));
        }
    }

    /* Add existing command message */
    parseBumCommand(trn, have, sid, -1);

    setFlags(can_GetUnload);
}
