/**
  *  \file game/config.cc
  *  \brief Game Configuration (HConfig/PConfig) Access
  *
  *  This file contains functions to manage and parse PConfig. HConfig
  *  is a strict subset of PConfig (with the single exception of
  *  AllowVPAFeatures), so it is handled here, too.
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include <cstring>
#include <cstdio>
#include <vector>
#include "cpluslib/assert.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/gamedirs.h"
#include "game/host.h"
#include "game/struct.h"
#include "io/io.h"
#include "util/confpars.h"
#include "util/console.h"
#include "util/except.h"
#include "util/misc.h"
#include "util/nls.h"
#include "game/playerset.h"
#include "util/modinit.h"
#include "game/exp.h"

namespace {
    enum {
        MAX_HCONFIG_SIZE = 1000
    };
}

/**************************** ConfigCostOption ***************************/

/** Constructor.
    \param container Parent
    \param name Name of option */
ConfigCostOption::ConfigCostOption(Config& container, const char* name)
    : ConfigOption(container, name)
{ }

/** Destructor. */
ConfigCostOption::~ConfigCostOption()
{ }

/** Assign value from string. */
void
ConfigCostOption::set(string_t value)
{
    int count = 0;
    GCost last_value;
    string_t::size_type n;
    while (count < NUM_PLAYERS-1 && (n = value.find(',')) != string_t::npos) {
        last_value = GCost::fromString(value.substr(0, n));
        value.erase(0, n+1);
        data[count++] = last_value;
    }
    last_value = GCost::fromString(value);
    data[count++] = last_value;
    while (count < NUM_PLAYERS)
        data[count++] = last_value;
}

/** Convert to string. */
string_t
ConfigCostOption::toString() const
{
    string_t result = data[0].toPHostString();
    bool diff = false;
    for (int i = 1; i < NUM_PLAYERS; ++i) {
        if (data[i] != data[0]) {
            diff = true;
            break;
        }
    }
    if (diff) {
        for (int i = 1; i < NUM_PLAYERS; ++i) {
            result += ",";
            result += data[i].toPHostString();
        }
    }
    return result;
}

/** Set single value.
    \param index Player slot to set [1,NUM_PLAYERS]
    \param value Value to assign */
void
ConfigCostOption::set(int index, const GCost& value)
{
    ASSERT(index > 0 && index <= NUM_PLAYERS);
    data[index-1] = value;
}

/** Element access.
    \param player Player slot to query [1,NUM_PLAYERS], values outside this range folded into range
    \return value */
GCost
ConfigCostOption::operator()(int player) const
{
    if (player > 0 && player <= NUM_PLAYERS)
        return data[player-1];
    else
        return data[NUM_PLAYERS-1];
}

/**************************** ConfigStdOption ****************************/

string_t
ConfigStdOption::toString() const
{
    if (isAllTheSame()) {
        return getParser().toString(operator()(1));
    } else {
        return ConfigIntArrayOption<NUM_PLAYERS>::toString();
    }
}

/****************************** GGameConfig ******************************/

/** Game configuration (hconfig/pconfig). */
GGameConfig config;

/** Construct blank config object. */
GGameConfig::GGameConfig()
    : ConfigLevel(*this, "ConfigLevel", ValueIntParser::instance),
      PlayerRace(*this, "PlayerRace", ValueIntParser::instance),
      RecycleRate(*this, "RecycleRate", ValueIntParser::instance),
      RandomMeteorRate(*this, "RandomMeteorRate", ValueIntParser::instance),
      AllowMinefields(*this, "AllowMinefields", ValueBoolParser::instance),
      AllowAlchemy(*this, "AllowAlchemy", ValueBoolParser::instance),
      DeleteOldMessages(*this, "DeleteOldMessages", ValueBoolParser::instance),
      DisablePasswords(*this, "DisablePasswords", ValueBoolParser::instance),
      GroundKillFactor(*this, "GroundKillFactor", ValueIntParser::instance),
      GroundDefenseFactor(*this, "GroundDefenseFactor", ValueIntParser::instance),
      FreeFighters(*this, "FreeFighters", ValueIntParser::instance),
      RaceMiningRate(*this, "RaceMiningRate", ValueIntParser::instance),
      AllowBuildFighters(*this, "AllowBuildFighters", ValueBoolParser::instance),
      RebelsBuildFighters(*this, "RebelsBuildFighters", ValueBoolParser::instance),
      ColoniesBuildFighters(*this, "ColoniesBuildFighters", ValueBoolParser::instance),
      RobotsBuildFighters(*this, "RobotsBuildFighters", ValueBoolParser::instance),
      CloakFailureRate(*this, "CloakFailureRate", ValueIntParser::instance),
      RobCloakedShips(*this, "RobCloakedShips", ValueBoolParser::instance),
      ScanRange(*this, "ScanRange", ValueIntParser::instance),
      SensorRange(*this, "SensorRange", ValueIntParser::instance),
      DarkSenseRange(*this, "DarkSenseRange", ValueIntParser::instance),
      AllowHiss(*this, "AllowHiss", ValueBoolParser::instance),
      AllowRebelGroundAttack(*this, "AllowRebelGroundAttack", ValueBoolParser::instance),
      AllowSuperRefit(*this, "AllowSuperRefit", ValueBoolParser::instance),
      AllowWebMines(*this, "AllowWebMines", ValueBoolParser::instance),
      CloakFuelBurn(*this, "CloakFuelBurn", ValueIntParser::instance),
      AllowNewNatives(*this, "AllowNewNatives", ValueBoolParser::instance),
      AllowPlanetAttacks(*this, "AllowPlanetAttacks", ValueBoolParser::instance),
      BorgAssimilationRate(*this, "BorgAssimilationRate", ValueIntParser::instance),
      MineDecayRate(*this, "MineDecayRate", ValueIntParser::instance),
      WebMineDecayRate(*this, "WebMineDecayRate", ValueIntParser::instance),
      MaximumMinefieldRadius(*this, "MaximumMinefieldRadius", ValueIntParser::instance),
      MaximumWebMinefieldRadius(*this, "MaximumWebMinefieldRadius", ValueIntParser::instance),
      TransuraniumDecayRate(*this, "TransuraniumDecayRate", ValueIntParser::instance),
      StructureDecayPerTurn(*this, "StructureDecayPerTurn", ValueIntParser::instance),
      StructureDecayOnUnowned(*this, "StructureDecayOnUnowned", ValueIntParser::instance),
      ClimateLimitsPopulation(*this, "ClimateLimitsPopulation", ValueBoolParser::instance),
      AllowEatingSupplies(*this, "AllowEatingSupplies", ValueBoolParser::instance),
      AllowNoFuelMovement(*this, "AllowNoFuelMovement", ValueBoolParser::instance),
      MineHitOdds(*this, "MineHitOdds", ValueIntParser::instance),
      WebMineHitOdds(*this, "WebMineHitOdds", ValueIntParser::instance),
      MineScanRange(*this, "MineScanRange", ValueIntParser::instance),
      AllowMinesDestroyMines(*this, "AllowMinesDestroyMines", ValueBoolParser::instance),
      AllowEngineShieldBonus(*this, "AllowEngineShieldBonus", ValueBoolParser::instance),
      EngineShieldBonusRate(*this, "EngineShieldBonusRate", ValueIntParser::instance),
      FighterSweepRate(*this, "FighterSweepRate", ValueIntParser::instance),
      AllowColoniesSweepWebs(*this, "AllowColoniesSweepWebs", ValueBoolParser::instance),
      MineSweepRate(*this, "MineSweepRate", ValueIntParser::instance),
      WebMineSweepRate(*this, "WebMineSweepRate", ValueIntParser::instance),
      HissEffectRate(*this, "HissEffectRate", ValueIntParser::instance),
      RobFailureOdds(*this, "RobFailureOdds", ValueIntParser::instance),
      PlanetsAttackRebels(*this, "PlanetsAttackRebels", ValueBoolParser::instance),
      PlanetsAttackKlingons(*this, "PlanetsAttackKlingons", ValueBoolParser::instance),
      MineSweepRange(*this, "MineSweepRange", ValueIntParser::instance),
      WebMineSweepRange(*this, "WebMineSweepRange", ValueIntParser::instance),
      AllowScienceMissions(*this, "AllowScienceMissions", ValueBoolParser::instance),
      MineHitOddsWhenCloakedX10(*this, "MineHitOddsWhenCloakedX10", ValueIntParser::instance),
      DamageLevelForCloakFail(*this, "DamageLevelForCloakFail", ValueIntParser::instance),
      AllowFedCombatBonus(*this, "AllowFedCombatBonus", ValueBoolParser::instance),
      MeteorShowerOdds(*this, "MeteorShowerOdds", ValueIntParser::instance),
      MeteorShowerOreRanges(*this, "MeteorShowerOreRanges", ValueIntParser::instance),
      LargeMeteorsImpacting(*this, "LargeMeteorsImpacting", ValueIntParser::instance),
      LargeMeteorOreRanges(*this, "LargeMeteorOreRanges", ValueIntParser::instance),
      AllowMeteorMessages(*this, "AllowMeteorMessages", ValueBoolParser::instance),
      AllowOneEngineTowing(*this, "AllowOneEngineTowing", ValueBoolParser::instance),
      AllowHyperWarps(*this, "AllowHyperWarps", ValueBoolParser::instance),
      ClimateDeathRate(*this, "ClimateDeathRate", ValueIntParser::instance),
      AllowGravityWells(*this, "AllowGravityWells", ValueBoolParser::instance),
      CrystalsPreferDeserts(*this, "CrystalsPreferDeserts", ValueBoolParser::instance),
      AllowMinesDestroyWebs(*this, "AllowMinesDestroyWebs", ValueBoolParser::instance),
      MaxPlanetaryIncome(*this, "MaxPlanetaryIncome", ValueIntParser::instance),
      IonStormActivity(*this, "IonStormActivity", ValueIntParser::instance),
      AllowChunneling(*this, "AllowChunneling", ValueBoolParser::instance),
      AllowDeluxeSuperSpy(*this, "AllowDeluxeSuperSpy", ValueBoolParser::instance),
      IonStormsHideMines(*this, "IonStormsHideMines", ValueBoolParser::instance),
      AllowGloryDevice(*this, "AllowGloryDevice", ValueBoolParser::instance),
      AllowAntiCloakShips(*this, "AllowAntiCloakShips", ValueBoolParser::instance),
      AllowGamblingShips(*this, "AllowGamblingShips", ValueBoolParser::instance),
      AllowCloakedShipsAttack(*this, "AllowCloakedShipsAttack", ValueBoolParser::instance),
      AllowShipCloning(*this, "AllowShipCloning", ValueBoolParser::instance),
      AllowBoardingParties(*this, "AllowBoardingParties", ValueBoolParser::instance),
      AllowImperialAssault(*this, "AllowImperialAssault", ValueBoolParser::instance),
      RamScoopFuelPerLY(*this, "RamScoopFuelPerLY", ValueIntParser::instance),
      AllowAdvancedRefinery(*this, "AllowAdvancedRefinery", ValueBoolParser::instance),
      AllowBioscanners(*this, "AllowBioscanners", ValueBoolParser::instance),
      HullTechNotSlowedByMines(*this, "HullTechNotSlowedByMines", ValueIntParser::instance),
      UseAccurateFuelModel(*this, "UseAccurateFuelModel", ValueBoolParser::instance),
      DefenseForUndetectable(*this, "DefenseForUndetectable", ValueIntParser::instance),
      DefenseToBlockBioscan(*this, "DefenseToBlockBioscan", ValueIntParser::instance),
      FactoriesForDetectable(*this, "FactoriesForDetectable", ValueIntParser::instance),
      MinesForDetectable(*this, "MinesForDetectable", ValueIntParser::instance),
      FighterSweepRange(*this, "FighterSweepRange", ValueIntParser::instance),
      MineHitDamageFor100KT(*this, "MineHitDamageFor100KT", ValueIntParser::instance),
      WebHitDamageFor100KT(*this, "WebHitDamageFor100KT", ValueIntParser::instance),
      AllowRegisteredFunctions(*this, "AllowRegisteredFunctions", ValueBoolParser::instance),
      GravityWellRange(*this, "GravityWellRange", ValueIntParser::instance),
      Language(*this, "Language", 0, NUM_PLAYERS+1),
      AllowPlayerMessages(*this, "AllowPlayerMessages", ValueBoolParser::instance),
      ScoringMethod(*this, "ScoringMethod"),
      TowedShipsBreakFree(*this, "TowedShipsBreakFree", ValueBoolParser::instance),
      NativeClimateDeathRate(*this, "NativeClimateDeathRate", ValueIntParser::instance),
      AllowMoreThan50Targets(*this, "AllowMoreThan50Targets", ValueBoolParser::instance),
      CrystalSinTempBehavior(*this, "CrystalSinTempBehavior", ValueBoolParser::instance),
      RGANeedsBeams(*this, "RGANeedsBeams", ValueBoolParser::instance),
      AllowRGAOnUnowned(*this, "AllowRGAOnUnowned", ValueBoolParser::instance),
      CPEnableLanguage(*this, "CPEnableLanguage", ValueBoolParser::instance),
      CPEnableBigTargets(*this, "CPEnableBigTargets", ValueBoolParser::instance),
      CPEnableRaceName(*this, "CPEnableRaceName", ValueBoolParser::instance),
      CPEnableAllies(*this, "CPEnableAllies", ValueBoolParser::instance),
      CPEnableMessage(*this, "CPEnableMessage", ValueBoolParser::instance),
      AllowAnonymousMessages(*this, "AllowAnonymousMessages", ValueBoolParser::instance),
      DelayAllianceCommands(*this, "DelayAllianceCommands", ValueBoolParser::instance),
      TerraformRate(*this, "TerraformRate", ValueIntParser::instance),
      MaxColTempSlope(*this, "MaxColTempSlope", ValueIntParser::instance),
      WebDrainFuelLoss(*this, "WebDrainFuelLoss", ValueIntParser::instance),
      WebHitFuelLoss(*this, "WebHitFuelLoss", ValueIntParser::instance),
      AllowWormholes(*this, "AllowWormholes", ValueBoolParser::instance),
      WrmDisplacement(*this, "WrmDisplacement", ValueIntParser::instance),
      WrmRandDisplacement(*this, "WrmRandDisplacement", ValueIntParser::instance),
      WrmStabilityAddX10(*this, "WrmStabilityAddX10", ValueIntParser::instance),
      WrmRandStability(*this, "WrmRandStability", ValueIntParser::instance),
      WrmMassAdd(*this, "WrmMassAdd", ValueIntParser::instance),
      WrmRandMass(*this, "WrmRandMass", ValueIntParser::instance),
      WrmVoluntaryTravel(*this, "WrmVoluntaryTravel", ValueBoolParser::instance),
      WrmTravelDistDivisor(*this, "WrmTravelDistDivisor", ValueIntParser::instance),
      WrmTravelWarpSpeed(*this, "WrmTravelWarpSpeed", ValueIntParser::instance),
      WrmTravelCloaked(*this, "WrmTravelCloaked", ValueBoolParser::instance),
      WrmEntryPowerX100(*this, "WrmEntryPowerX100", ValueIntParser::instance),
      CPEnableGive(*this, "CPEnableGive", ValueBoolParser::instance),
      AllowTowCloakedShips(*this, "AllowTowCloakedShips", ValueBoolParser::instance),
      RobCloakedChance(*this, "RobCloakedChance", ValueIntParser::instance),
      UnitsPerTorpRate(*this, "UnitsPerTorpRate", ValueIntParser::instance),
      UnitsPerWebRate(*this, "UnitsPerWebRate", ValueIntParser::instance),
      AllowESBonusAgainstPlanets(*this, "AllowESBonusAgainstPlanets", ValueBoolParser::instance),
      ShipCloneCostRate(*this, "ShipCloneCostRate", ValueIntParser::instance),
      AllowHyperjumpGravWells(*this, "AllowHyperjumpGravWells", ValueBoolParser::instance),
      NativeCombatSurvivalRate(*this, "NativeCombatSurvivalRate", ValueIntParser::instance),
      AllowPrivateerTowCapture(*this, "AllowPrivateerTowCapture", ValueBoolParser::instance),
      AllowCrystalTowCapture(*this, "AllowCrystalTowCapture", ValueBoolParser::instance),
      RoundGravityWells(*this, "RoundGravityWells", ValueBoolParser::instance),
      CPEnableSend(*this, "CPEnableSend", ValueBoolParser::instance),
      CumulativePillaging(*this, "CumulativePillaging", ValueBoolParser::instance),
      AllowInterceptAttack(*this, "AllowInterceptAttack", ValueBoolParser::instance),
      RaceGrowthRate(*this, "RaceGrowthRate", ValueIntParser::instance),
      ProductionRate(*this, "ProductionRate", ValueIntParser::instance),
      MineOddsWarpBonusX100(*this, "MineOddsWarpBonusX100", ValueIntParser::instance),
      CloakMineOddsWarpBonusX100(*this, "CloakMineOddsWarpBonusX100", ValueIntParser::instance),
      WebMineOddsWarpBonusX100(*this, "WebMineOddsWarpBonusX100", ValueIntParser::instance),
      MineTravelSafeWarp(*this, "MineTravelSafeWarp", ValueIntParser::instance),
      CloakedMineTravelSafeWarp(*this, "CloakedMineTravelSafeWarp", ValueIntParser::instance),
      WebMineTravelSafeWarp(*this, "WebMineTravelSafeWarp", ValueIntParser::instance),
      AllowCloakFailMessages(*this, "AllowCloakFailMessages", ValueBoolParser::instance),
      TonsScoreCountsPlanets(*this, "TonsScoreCountsPlanets", ValueBoolParser::instance),
      AllowExtendedMissions(*this, "AllowExtendedMissions", ValueBoolParser::instance),
      ExtMissionsStartAt(*this, "ExtMissionsStartAt", ValueIntParser::instance),
      WormholeUFOsStartAt(*this, "WormholeUFOsStartAt", ValueIntParser::instance),
      MaxShipsHissing(*this, "MaxShipsHissing", ValueIntParser::instance),
      SpyDetectionChance(*this, "SpyDetectionChance", ValueIntParser::instance),
      MapTruehullByPlayerRace(*this, "MapTruehullByPlayerRace", ValueBoolParser::instance),
      AllowWraparoundMap(*this, "AllowWraparoundMap", ValueBoolParser::instance),
      WraparoundRectangle(*this, "WraparoundRectangle", ValueIntParser::instance),
      CPEnableRemote(*this, "CPEnableRemote", ValueBoolParser::instance),
      AllowAlliedChunneling(*this, "AllowAlliedChunneling", ValueBoolParser::instance),
      ColonistTaxRate(*this, "ColonistTaxRate", ValueIntParser::instance),
      NativeTaxRate(*this, "NativeTaxRate", ValueIntParser::instance),
      AllowAlternativeTowing(*this, "AllowAlternativeTowing", ValueBoolParser::instance),
      AllowBeamUpClans(*this, "AllowBeamUpClans", ValueBoolParser::instance),
      AllowBeamUpMultiple(*this, "AllowBeamUpMultiple", ValueBoolParser::instance),
      DamageLevelForTerraformFail(*this, "DamageLevelForTerraformFail", ValueIntParser::instance),
      DamageLevelForAntiCloakFail(*this, "DamageLevelForAntiCloakFail", ValueIntParser::instance),
      DamageLevelForChunnelFail(*this, "DamageLevelForChunnelFail", ValueIntParser::instance),
      DamageLevelForHyperjumpFail(*this, "DamageLevelForHyperjumpFail", ValueIntParser::instance),
      MinimumChunnelDistance(*this, "MinimumChunnelDistance", ValueIntParser::instance),
      TowStrengthEngineScale(*this, "TowStrengthEngineScale", ValueIntParser::instance),
      TowStrengthDistanceScale(*this, "TowStrengthDistanceScale", ValueIntParser::instance),
      AllowPriorityBuild(*this, "AllowPriorityBuild", ValueBoolParser::instance),
      SBQBuildPALBoost(*this, "SBQBuildPALBoost", ValueIntParser::instance),
      SBQNewBuildPALBoost(*this, "SBQNewBuildPALBoost", ValueIntParser::instance),
      SBQPointsForAging(*this, "SBQPointsForAging", ValueIntParser::instance),
      SBQBuildChangePenalty(*this, "SBQBuildChangePenalty", ValueIntParser::instance),
      SBQBoostExpX100(*this, "SBQBoostExpX100", ValueIntParser::instance),
      BuildChangeRelativePenalty(*this, "BuildChangeRelativePenalty", ValueIntParser::instance),
      PALDecayPerTurn(*this, "PALDecayPerTurn", ValueIntParser::instance),
      PALPlayerRate(*this, "PALPlayerRate", ValueIntParser::instance),
      PALCombatAggressor(*this, "PALCombatAggressor", ValueIntParser::instance),
      PALAggressorPointsPer10KT(*this, "PALAggressorPointsPer10KT", ValueIntParser::instance),
      PALOpponentPointsPer10KT(*this, "PALOpponentPointsPer10KT", ValueIntParser::instance),
      PALAggressorKillPointsPer10KT(*this, "PALAggressorKillPointsPer10KT", ValueIntParser::instance),
      PALOpponentKillPointsPer10KT(*this, "PALOpponentKillPointsPer10KT", ValueIntParser::instance),
      PALCombatPlanetScaling(*this, "PALCombatPlanetScaling", ValueIntParser::instance),
      PALCombatBaseScaling(*this, "PALCombatBaseScaling", ValueIntParser::instance),
      PALShipCapturePer10Crew(*this, "PALShipCapturePer10Crew", ValueIntParser::instance),
      PALRecyclingPer10KT(*this, "PALRecyclingPer10KT", ValueIntParser::instance),
      PALBoardingPartyPer10Crew(*this, "PALBoardingPartyPer10Crew", ValueIntParser::instance),
      PALGroundAttackPer100Clans(*this, "PALGroundAttackPer100Clans", ValueIntParser::instance),
      PALGloryDevice(*this, "PALGloryDevice", ValueIntParser::instance),
      PALGloryDevicePer10KT(*this, "PALGloryDevicePer10KT", ValueIntParser::instance),
      PALGloryDamagePer10KT(*this, "PALGloryDamagePer10KT", ValueIntParser::instance),
      PALGloryKillPer10KT(*this, "PALGloryKillPer10KT", ValueIntParser::instance),
      PALImperialAssault(*this, "PALImperialAssault", ValueIntParser::instance),
      PALRGA(*this, "PALRGA", ValueIntParser::instance),
      PALPillage(*this, "PALPillage", ValueIntParser::instance),
      PALIncludesESB(*this, "PALIncludesESB", ValueBoolParser::instance),
      FilterPlayerMessages(*this, "FilterPlayerMessages", ValueBoolParser::instance),
      AlternativeAntiCloak(*this, "AlternativeAntiCloak", ValueBoolParser::instance),
      AntiCloakImmunity(*this, "AntiCloakImmunity", ValueBoolParser::instance),
      AllowMoreThan500Minefields(*this, "AllowMoreThan500Minefields", ValueBoolParser::instance),
      NumMinefields(*this, "NumMinefields", ValueIntParser::instance),
      PALShipMinekillPer10KT(*this, "PALShipMinekillPer10KT", ValueIntParser::instance),
      MaximumMinefieldsPerPlayer(*this, "MaximumMinefieldsPerPlayer", ValueIntParser::instance),
      MineIdNeedsPermission(*this, "MineIdNeedsPermission", ValueBoolParser::instance),
      BuildQueue(*this, "BuildQueue"),
      PBPCostPer100KT(*this, "PBPCostPer100KT", ValueIntParser::instance),
      PBPMinimumCost(*this, "PBPMinimumCost", ValueIntParser::instance),
      PBPCloneCostRate(*this, "PBPCloneCostRate", ValueIntParser::instance),
      AllowShipNames(*this, "AllowShipNames", ValueBoolParser::instance),
      BuildPointReport(*this, "BuildPointReport", ValueBoolParser::instance),
      AlternativeMinesDestroyMines(*this, "AlternativeMinesDestroyMines", ValueBoolParser::instance),
      NumShips(*this, "NumShips", ValueIntParser::instance),
      ExtendedSensorSweep(*this, "ExtendedSensorSweep", ValueBoolParser::instance),
      ColonistCombatSurvivalRate(*this, "ColonistCombatSurvivalRate", ValueIntParser::instance),
      NewNativesPerTurn(*this, "NewNativesPerTurn", ValueIntParser::instance),
      NewNativesPopulationRange(*this, "NewNativesPopulationRange", ValueIntParser::instance),
      NewNativesRaceRate(*this, "NewNativesRaceRate", ValueIntParser::instance),
      NewNativesGovernmentRate(*this, "NewNativesGovernmentRate", ValueIntParser::instance),
      PlayerSpecialMission(*this, "PlayerSpecialMission", ValueIntParser::instance),
      TowedShipsCooperate(*this, "TowedShipsCooperate", ValueBoolParser::instance),
      WrmScanRange(*this, "WrmScanRange", ValueIntParser::instance),
      FuelUsagePerFightFor100KT(*this, "FuelUsagePerFightFor100KT", ValueIntParser::instance),
      FuelUsagePerTurnFor100KT(*this, "FuelUsagePerTurnFor100KT", ValueIntParser::instance),
      CPEnableEnemies(*this, "CPEnableEnemies", ValueBoolParser::instance),
      CPEnableShow(*this, "CPEnableShow", ValueBoolParser::instance),
      CPEnableRefit(*this, "CPEnableRefit", ValueBoolParser::instance),
      AllowIncompatibleConfiguration(*this, "AllowIncompatibleConfiguration", ValueBoolParser::instance),
      FreeFighterCost(*this, "FreeFighterCost"),
      StarbaseCost(*this, "StarbaseCost"),
      BaseFighterCost(*this, "BaseFighterCost"),
      ShipFighterCost(*this, "ShipFighterCost"),
      MaximumFightersOnBase(*this, "MaximumFightersOnBase", ValueIntParser::instance),
      MaximumDefenseOnBase(*this, "MaximumDefenseOnBase", ValueIntParser::instance),
      NumExperienceLevels(*this, "NumExperienceLevels", ValueIntParser::instance),
      ExperienceLevels(*this, "ExperienceLevels", ValueIntParser::instance),
      ExperienceLevelNames(*this, "ExperienceLevelNames", 0, NUM_EXPERIENCE_LEVELS+1),
      ExperienceLimit(*this, "ExperienceLimit", ValueIntParser::instance),
      EPRecrewScaling(*this, "EPRecrewScaling", ValueIntParser::instance),
      EPShipAging(*this, "EPShipAging", ValueIntParser::instance),
      EPPlanetAging(*this, "EPPlanetAging", ValueIntParser::instance),
      EPPlanetGovernment(*this, "EPPlanetGovernment", ValueIntParser::instance),
      EPShipMovement100LY(*this, "EPShipMovement100LY", ValueIntParser::instance),
      EPShipHyperjump(*this, "EPShipHyperjump", ValueIntParser::instance),
      EPShipChunnel(*this, "EPShipChunnel", ValueIntParser::instance),
      EPShipIonStorm100MEV(*this, "EPShipIonStorm100MEV", ValueIntParser::instance),
      EPCombatKillScaling(*this, "EPCombatKillScaling", ValueIntParser::instance),
      EPCombatDamageScaling(*this, "EPCombatDamageScaling", ValueIntParser::instance),
      EPShipAlchemy100KT(*this, "EPShipAlchemy100KT", ValueIntParser::instance),
      EPCombatBoostRate(*this, "EPCombatBoostRate", ValueIntParser::instance),
      EPCombatBoostLevel(*this, "EPCombatBoostLevel", ValueIntParser::instance),
      EPTrainingScale(*this, "EPTrainingScale", ValueIntParser::instance),
      ExactExperienceReports(*this, "ExactExperienceReports", ValueBoolParser::instance),
      EModBayRechargeRate(*this, "EModBayRechargeRate", ValueIntParser::instance),
      EModBayRechargeBonus(*this, "EModBayRechargeBonus", ValueIntParser::instance),
      EModBeamRechargeRate(*this, "EModBeamRechargeRate", ValueIntParser::instance),
      EModBeamRechargeBonus(*this, "EModBeamRechargeBonus", ValueIntParser::instance),
      EModTubeRechargeRate(*this, "EModTubeRechargeRate", ValueIntParser::instance),
      EModBeamHitFighterCharge(*this, "EModBeamHitFighterCharge", ValueIntParser::instance),
      EModTorpHitOdds(*this, "EModTorpHitOdds", ValueIntParser::instance),
      EModBeamHitOdds(*this, "EModBeamHitOdds", ValueIntParser::instance),
      EModBeamHitBonus(*this, "EModBeamHitBonus", ValueIntParser::instance),
      EModStrikesPerFighter(*this, "EModStrikesPerFighter", ValueIntParser::instance),
      EModFighterBeamExplosive(*this, "EModFighterBeamExplosive", ValueIntParser::instance),
      EModFighterBeamKill(*this, "EModFighterBeamKill", ValueIntParser::instance),
      EModFighterMovementSpeed(*this, "EModFighterMovementSpeed", ValueIntParser::instance),
      EModMaxFightersLaunched(*this, "EModMaxFightersLaunched", ValueIntParser::instance),
      EModTorpHitBonus(*this, "EModTorpHitBonus", ValueIntParser::instance),
      EModTubeRechargeBonus(*this, "EModTubeRechargeBonus", ValueIntParser::instance),
      EModExtraFighterBays(*this, "EModExtraFighterBays", ValueIntParser::instance),
      EModEngineShieldBonusRate(*this, "EModEngineShieldBonusRate", ValueIntParser::instance),
      EModShieldDamageScaling(*this, "EModShieldDamageScaling", ValueIntParser::instance),
      EModShieldKillScaling(*this, "EModShieldKillScaling", ValueIntParser::instance),
      EModHullDamageScaling(*this, "EModHullDamageScaling", ValueIntParser::instance),
      EModCrewKillScaling(*this, "EModCrewKillScaling", ValueIntParser::instance),
      EModPlanetaryTorpsPerTube(*this, "EModPlanetaryTorpsPerTube", ValueIntParser::instance),
      EModMineHitOddsBonus(*this, "EModMineHitOddsBonus", ValueIntParser::instance),
      GameName(*this, "GameName"),
      BayRechargeRate(*this, "BayRechargeRate", ValueIntParser::instance),
      BayRechargeBonus(*this, "BayRechargeBonus", ValueIntParser::instance),
      BeamRechargeRate(*this, "BeamRechargeRate", ValueIntParser::instance),
      BeamRechargeBonus(*this, "BeamRechargeBonus", ValueIntParser::instance),
      TubeRechargeRate(*this, "TubeRechargeRate", ValueIntParser::instance),
      BeamHitFighterCharge(*this, "BeamHitFighterCharge", ValueIntParser::instance),
      BeamHitShipCharge(*this, "BeamHitShipCharge", ValueIntParser::instance),
      TorpFiringRange(*this, "TorpFiringRange", ValueIntParser::instance),
      BeamFiringRange(*this, "BeamFiringRange", ValueIntParser::instance),
      TorpHitOdds(*this, "TorpHitOdds", ValueIntParser::instance),
      BeamHitOdds(*this, "BeamHitOdds", ValueIntParser::instance),
      BeamHitBonus(*this, "BeamHitBonus", ValueIntParser::instance),
      StrikesPerFighter(*this, "StrikesPerFighter", ValueIntParser::instance),
      FighterKillOdds(*this, "FighterKillOdds", ValueIntParser::instance),
      FighterBeamExplosive(*this, "FighterBeamExplosive", ValueIntParser::instance),
      FighterBeamKill(*this, "FighterBeamKill", ValueIntParser::instance),
      ShipMovementSpeed(*this, "ShipMovementSpeed", ValueIntParser::instance),
      FighterMovementSpeed(*this, "FighterMovementSpeed", ValueIntParser::instance),
      BayLaunchInterval(*this, "BayLaunchInterval", ValueIntParser::instance),
      MaxFightersLaunched(*this, "MaxFightersLaunched", ValueIntParser::instance),
      AllowAlternativeCombat(*this, "AllowAlternativeCombat", ValueBoolParser::instance),
      StandoffDistance(*this, "StandoffDistance", ValueIntParser::instance),
      PlanetsHaveTubes(*this, "PlanetsHaveTubes", ValueBoolParser::instance),
      PlanetaryTorpsPerTube(*this, "PlanetaryTorpsPerTube", ValueIntParser::instance),
      FireOnAttackFighters(*this, "FireOnAttackFighters", ValueBoolParser::instance),
      TorpHitBonus(*this, "TorpHitBonus", ValueIntParser::instance),
      TubeRechargeBonus(*this, "TubeRechargeBonus", ValueIntParser::instance),
      ShieldDamageScaling(*this, "ShieldDamageScaling", ValueIntParser::instance),
      HullDamageScaling(*this, "HullDamageScaling", ValueIntParser::instance),
      CrewKillScaling(*this, "CrewKillScaling", ValueIntParser::instance),
      ShieldKillScaling(*this, "ShieldKillScaling", ValueIntParser::instance),
      ExtraFighterBays(*this, "ExtraFighterBays", ValueIntParser::instance),
      BeamHitFighterRange(*this, "BeamHitFighterRange", ValueIntParser::instance),
      FighterFiringRange(*this, "FighterFiringRange", ValueIntParser::instance),
      AllowVPAFeatures(*this, "AllowVPAFeatures", ValueIntParser::instance),
      CPEnableRumor(*this, "CPEnableRumor", AllowAnonymousMessages),
      RaceTaxRate(*this, "RaceTaxRate", ColonistTaxRate),
      CPNumMinefields(*this, "CPNumMinefields", NumMinefields),
      NativeClansRange(*this, "NativeClansRange", NewNativesPopulationRange),
      NativeTypeFrequencies(*this, "NativeTypeFrequencies", NewNativesRaceRate),
      NativeGovFrequencies(*this, "NativeGovFrequencies", NewNativesGovernmentRate),
      MinimumHappiness(*this, "MinimumHappiness", ValueIntParser::instance),
      ColonistCombatCaptureRate(*this, "ColonistCombatCaptureRate", ValueIntParser::instance),
      EPAcademyScale(*this, "EPAcademyScale", ValueIntParser::instance),
      UseBaseTorpsInCombat(*this, "UseBaseTorpsInCombat", ValueBoolParser::instance),
      BaseTechCost(*this, "BaseTechCost", ValueIntParser::instance),
      EPShipBuild1000TorpUnits(*this, "EPShipBuild1000TorpUnits", ValueIntParser::instance),
      EPShipBuild10Fighters(*this, "EPShipBuild10Fighters", ValueIntParser::instance)
{ }

/** Destructor. */
GGameConfig::~GGameConfig()
{ }

/** Assign default values to all options. These defaults are (mostly) the PHost default
    values. A HCONFIG.HST file is never incomplete, so we need not deal with partially
    populated Tim-Host configuration at this point. */
void
GGameConfig::assignDefaults()
{
    ConfigLevel.set(0);
    PlayerRace.set("1,2,3,4,5,6,7,8,9,10,11");
    RecycleRate.set(75);
    RandomMeteorRate.set(2);
    AllowMinefields.set(1);
    AllowAlchemy.set(1);
    DeleteOldMessages.set(1);
    DisablePasswords.set(0);
    GroundKillFactor.set("1,30,1,15,1,1,1,1,1,1,1");
    GroundDefenseFactor.set("1,10,1,5,1,1,1,1,1,1,1");
    FreeFighters.set("0,0,0,0,0,0,0,10,0,0,0");
    RaceMiningRate.set("70,200,100,100,100,100,100,100,100,100,100");
    AllowBuildFighters.set("No,No,No,No,No,No,No,No,Yes,Yes,Yes");
    RebelsBuildFighters.set(1);
    ColoniesBuildFighters.set(1);
    RobotsBuildFighters.set(1);
    CloakFailureRate.set(1);
    RobCloakedShips.set(0);
    ScanRange.set(300);
    SensorRange.set(200);
    DarkSenseRange.copyFrom(SensorRange);
    AllowHiss.set(1);
    AllowRebelGroundAttack.set(1);
    AllowSuperRefit.set(1);
    AllowWebMines.copyFrom(AllowMinefields);
    CloakFuelBurn.set(5);
    AllowNewNatives.set(1);
    AllowPlanetAttacks.set(1);
    BorgAssimilationRate.set(100);
    MineDecayRate.set(5);
    WebMineDecayRate.copyFrom(MineDecayRate);
    MaximumMinefieldRadius.set(150);
    MaximumWebMinefieldRadius.copyFrom(MaximumMinefieldRadius);
    TransuraniumDecayRate.set(5);
    StructureDecayPerTurn.set(1);
    StructureDecayOnUnowned.set(1);
    ClimateLimitsPopulation.set(1);
    AllowEatingSupplies.set(0);
    AllowNoFuelMovement.set(1);
    MineHitOdds.set(1);
    WebMineHitOdds.set(5);
    MineScanRange.copyFrom(SensorRange);
    AllowMinesDestroyMines.set(1);
    AllowEngineShieldBonus.set(0);
    EngineShieldBonusRate.set(0);
    FighterSweepRate.set("0,0,0,0,0,0,0,0,0,0,20");
    AllowColoniesSweepWebs.set(0);
    MineSweepRate.set(4);
    WebMineSweepRate.set(3);
    HissEffectRate.set(5);
    RobFailureOdds.set(1);
    PlanetsAttackRebels.set(0);
    PlanetsAttackKlingons.set(0);
    MineSweepRange.set(10);
    WebMineSweepRange.set(5);
    AllowScienceMissions.set(1);
    MineHitOddsWhenCloakedX10.set(5);
    DamageLevelForCloakFail.set(1);
    AllowFedCombatBonus.set(1);
    MeteorShowerOdds.set(0);
    MeteorShowerOreRanges.set("10,10,10,10,200,200,200,200");
    LargeMeteorsImpacting.set(0);
    LargeMeteorOreRanges.set("100,100,100,100,10000,9000,9000,7000");
    AllowMeteorMessages.set(1);
    AllowOneEngineTowing.set(0);
    AllowHyperWarps.set(1);
    ClimateDeathRate.set(10);
    AllowGravityWells.set(1);
    CrystalsPreferDeserts.set(1);
    AllowMinesDestroyWebs.set(0);
    MaxPlanetaryIncome.set(5000);
    IonStormActivity.set(0);
    AllowChunneling.set(1);
    AllowDeluxeSuperSpy.set(1);
    IonStormsHideMines.set(1);
    AllowGloryDevice.set(1);
    AllowAntiCloakShips.set(1);
    AllowGamblingShips.set(1);
    AllowCloakedShipsAttack.set(1);
    AllowShipCloning.set(1);
    AllowBoardingParties.set(1);
    AllowImperialAssault.set(1);
    RamScoopFuelPerLY.set(2);
    AllowAdvancedRefinery.set(1);
    AllowBioscanners.set(1);
    HullTechNotSlowedByMines.set(7);
    UseAccurateFuelModel.set(0);
    DefenseForUndetectable.set(15);
    DefenseToBlockBioscan.set(20);
    FactoriesForDetectable.set(15);
    MinesForDetectable.set(20);
    FighterSweepRange.set("0,0,0,0,0,0,0,0,0,0,100");
    MineHitDamageFor100KT.set(100);
    WebHitDamageFor100KT.set(10);
    AllowRegisteredFunctions.set(1);
    GravityWellRange.set(3);
    Language.set("English");
    AllowPlayerMessages.set(1);
    ScoringMethod.set("Compatible");
    TowedShipsBreakFree.set(0);
    NativeClimateDeathRate.set(0);
    AllowMoreThan50Targets.set(0);
    CrystalSinTempBehavior.set(0);
    RGANeedsBeams.set(0);
    AllowRGAOnUnowned.set(0);
    CPEnableLanguage.set(1);
    CPEnableBigTargets.set(1);
    CPEnableRaceName.set(1);
    CPEnableAllies.set(1);
    CPEnableMessage.set(1);
    AllowAnonymousMessages.copyFrom(AllowPlayerMessages);
    DelayAllianceCommands.set(0);
    TerraformRate.set(1);
    MaxColTempSlope.set(1000);
    WebDrainFuelLoss.set(25);
    WebHitFuelLoss.set(50);
    AllowWormholes.set(1);
    WrmDisplacement.set(1);
    WrmRandDisplacement.set(0);
    WrmStabilityAddX10.set(0);
    WrmRandStability.set(0);
    WrmMassAdd.set(0);
    WrmRandMass.set(0);
    WrmVoluntaryTravel.set(1);
    WrmTravelDistDivisor.set(100);
    WrmTravelWarpSpeed.set(9);
    WrmTravelCloaked.set(0);
    WrmEntryPowerX100.set(25);
    CPEnableGive.set(1);
    AllowTowCloakedShips.set(0);
    RobCloakedChance.set(0);
    UnitsPerTorpRate.set("100,100,100,100,100,100,100,100,400,100,100");
    UnitsPerWebRate.copyFrom(UnitsPerTorpRate);
    AllowESBonusAgainstPlanets.set(1);
    ShipCloneCostRate.set("200,200,200,200,32767,200,32767,200,200,200,200");
    AllowHyperjumpGravWells.copyFrom(AllowGravityWells);
    NativeCombatSurvivalRate.set(75);
    AllowPrivateerTowCapture.copyFrom(AllowBoardingParties);
    AllowCrystalTowCapture.copyFrom(AllowBoardingParties);
    RoundGravityWells.set(0);
    CPEnableSend.set(1);
    CumulativePillaging.set(1);
    AllowInterceptAttack.set(1);
    RaceGrowthRate.set(100);
    ProductionRate.set(100);
    MineOddsWarpBonusX100.set(0);
    CloakMineOddsWarpBonusX100.set(0);
    WebMineOddsWarpBonusX100.set(0);
    MineTravelSafeWarp.set(0);
    CloakedMineTravelSafeWarp.copyFrom(MineTravelSafeWarp);
    WebMineTravelSafeWarp.copyFrom(MineTravelSafeWarp);
    AllowCloakFailMessages.set(1);
    TonsScoreCountsPlanets.set(0);
    AllowExtendedMissions.set(1);
    ExtMissionsStartAt.set(20);
    WormholeUFOsStartAt.set(51);
    MaxShipsHissing.set(500);
    SpyDetectionChance.set(20);
    MapTruehullByPlayerRace.set(0);
    AllowWraparoundMap.set(0);
    WraparoundRectangle.set("1000,1000,3000,3000");
    CPEnableRemote.set(1);
    AllowAlliedChunneling.set(1);
    ColonistTaxRate.set("200,100,100,100,100,100,100,100,100,100,100");
    NativeTaxRate.copyFrom(ColonistTaxRate);
    AllowAlternativeTowing.set(0);
    AllowBeamUpClans.set(1);
    AllowBeamUpMultiple.set(1);
    DamageLevelForTerraformFail.set(100);
    DamageLevelForAntiCloakFail.set(20);
    DamageLevelForChunnelFail.set(100);
    DamageLevelForHyperjumpFail.set(100);
    MinimumChunnelDistance.set(100);
    TowStrengthEngineScale.set(1);
    TowStrengthDistanceScale.set(19);
    AllowPriorityBuild.set(1);
    SBQBuildPALBoost.set(1);
    SBQNewBuildPALBoost.set(1);
    SBQPointsForAging.set(1200);
    SBQBuildChangePenalty.set(2147483647);
    SBQBoostExpX100.set(0);
    BuildChangeRelativePenalty.set(100);
    PALDecayPerTurn.set(20);
    PALPlayerRate.set(100);
    PALCombatAggressor.set(0);
    PALAggressorPointsPer10KT.set(2);
    PALOpponentPointsPer10KT.set(2);
    PALAggressorKillPointsPer10KT.set(10);
    PALOpponentKillPointsPer10KT.set(10);
    PALCombatPlanetScaling.set(50);
    PALCombatBaseScaling.set(80);
    PALShipCapturePer10Crew.set(5);
    PALRecyclingPer10KT.set(4);
    PALBoardingPartyPer10Crew.set(3);
    PALGroundAttackPer100Clans.set(100);
    PALGloryDevice.set(100);
    PALGloryDevicePer10KT.set(0);
    PALGloryDamagePer10KT.set(2);
    PALGloryKillPer10KT.set(0);
    PALImperialAssault.set(100);
    PALRGA.set(10);
    PALPillage.set(10);
    PALIncludesESB.set(1);
    FilterPlayerMessages.set(0);
    AlternativeAntiCloak.set(0);
    AntiCloakImmunity.set("Yes,Yes,Yes,No,No,No,No,No,No,No,No");
    AllowMoreThan500Minefields.set(0);
    NumMinefields.set(500);
    PALShipMinekillPer10KT.set(0);
    MaximumMinefieldsPerPlayer.set(10000);
    MineIdNeedsPermission.set(0);
    BuildQueue.set("PAL");
    PBPCostPer100KT.set(200);
    PBPMinimumCost.set(400);
    PBPCloneCostRate.set(200);
    AllowShipNames.set(1);
    BuildPointReport.set(2 /* Allies */);
    AlternativeMinesDestroyMines.set(0);
    NumShips.set(500);
    ExtendedSensorSweep.set(1);
    ColonistCombatSurvivalRate.set(0);
    NewNativesPerTurn.set(1);
    NewNativesPopulationRange.set("2500,5000");
    NewNativesRaceRate.set(1);
    NewNativesGovernmentRate.set(1);
    PlayerSpecialMission.copyFrom(PlayerRace);
    TowedShipsCooperate.set(1);
    WrmScanRange.set(100);
    FuelUsagePerFightFor100KT.set(0);
    FuelUsagePerTurnFor100KT.set(0);
    CPEnableEnemies.set(1);
    CPEnableShow.set(1);
    CPEnableRefit.set(1);
    AllowIncompatibleConfiguration.set(0);
    FreeFighterCost.set("T3 M2");
    StarbaseCost.set("T402 D120 M340 $900");
    BaseFighterCost.set("T3 M2 $100");
    ShipFighterCost.set("T3 M2 S5");
    MaximumFightersOnBase.set(60);
    MaximumDefenseOnBase.set(200);
    NumExperienceLevels.set(0);
    ExperienceLevels.set("750,1500,3000,6000");
    ExperienceLevelNames.set("Recruit,Soldier,Experienced,Elite,Ultra Elite");
    ExperienceLimit.set(1000000);
    EPRecrewScaling.set(30);
    EPShipAging.set(15);
    EPPlanetAging.set(25);
    EPPlanetGovernment.set(0);
    EPShipMovement100LY.set(40);
    EPShipHyperjump.set(30);
    EPShipChunnel.set(30);
    EPShipIonStorm100MEV.set(80);
    EPCombatKillScaling.set(800);
    EPCombatDamageScaling.set(200);
    EPShipAlchemy100KT.set(5);
    EPCombatBoostRate.set(100);
    EPCombatBoostLevel.set(0);
    EPTrainingScale.set(70);
    ExactExperienceReports.set(0);
    EModBayRechargeRate.set("1,2,3,4");
    EModBayRechargeBonus.set("0,0,0,0");
    EModBeamRechargeRate.set("0,0,0,0");
    EModBeamRechargeBonus.set("0,0,0,0");
    EModTubeRechargeRate.set("1,2,3,8");
    EModBeamHitFighterCharge.set("0,0,0,0");
    EModTorpHitOdds.set("9,18,27,35");
    EModBeamHitOdds.set("0,0,0,0");
    EModBeamHitBonus.set("0,0,0,0");
    EModStrikesPerFighter.set("1,2,3,4");
    EModFighterBeamExplosive.set("0,0,0,0");
    EModFighterBeamKill.set("0,0,0,0");
    EModFighterMovementSpeed.set("0,0,0,0");
    EModMaxFightersLaunched.set("0,0,0,0");
    EModTorpHitBonus.set("0,0,0,0");
    EModTubeRechargeBonus.set("0,0,0,0");
    EModExtraFighterBays.set("0,0,0,0");
    EModEngineShieldBonusRate.set("0,0,0,0");
    EModShieldDamageScaling.set("0,0,0,0");
    EModShieldKillScaling.set("0,0,0,0");
    EModHullDamageScaling.set("0,0,0,0");
    EModCrewKillScaling.set("-5,-10,-15,-20");
    EModPlanetaryTorpsPerTube.set("0,0,0,0");
    EModMineHitOddsBonus.set("5,10,15,20");
    GameName.set("Nameless Game");
    BayRechargeRate.set(52);
    BayRechargeBonus.set(2);
    BeamRechargeRate.set(8);
    BeamRechargeBonus.set(0);
    TubeRechargeRate.set(45);
    BeamHitFighterCharge.set(500);
    BeamHitShipCharge.set(600);
    TorpFiringRange.set(30000);
    BeamFiringRange.set(20000);
    TorpHitOdds.set(65);
    BeamHitOdds.set(100);
    BeamHitBonus.set(0);
    StrikesPerFighter.set(7);
    FighterKillOdds.set(20);
    FighterBeamExplosive.set(2);
    FighterBeamKill.set(2);
    ShipMovementSpeed.set(75);
    FighterMovementSpeed.set(235);
    BayLaunchInterval.set(3);
    MaxFightersLaunched.set(19);
    AllowAlternativeCombat.set(0);
    StandoffDistance.set(3000);
    PlanetsHaveTubes.set(0);
    PlanetaryTorpsPerTube.set(3);
    FireOnAttackFighters.set(0);
    TorpHitBonus.set(0);
    TubeRechargeBonus.set(0);
    ShieldDamageScaling.set(80);
    HullDamageScaling.set(80);
    CrewKillScaling.set(80);
    ShieldKillScaling.set(0);
    ExtraFighterBays.set("3,0,0,0,0,0,0,0,0,0,0");
    BeamHitFighterRange.set(100000);
    FighterFiringRange.set(3000);
    AllowVPAFeatures.set(1);
    MinimumHappiness.set(20);
    ColonistCombatCaptureRate.set(100);
    EPAcademyScale.set(400);
    UseBaseTorpsInCombat.set(1);
    BaseTechCost.set(100);
    EPShipBuild1000TorpUnits.set(0);
    EPShipBuild10Fighters.set(0);

    markAllOptionsUnset();
}

/** Assign dependant options. If an option has the default "same as other option",
    and has not been explicitly set in the config file, copy its value from the other
    option. */
void
GGameConfig::assignDependantOptions()
{
    if (!DarkSenseRange.wasSet())
        DarkSenseRange.copyFrom(SensorRange);
    if (!AllowWebMines.wasSet())
        AllowWebMines.copyFrom(AllowMinefields);
    if (!WebMineDecayRate.wasSet())
        WebMineDecayRate.copyFrom(MineDecayRate);
    if (!MaximumWebMinefieldRadius.wasSet())
        MaximumWebMinefieldRadius.copyFrom(MaximumMinefieldRadius);
    if (!MineScanRange.wasSet())
        MineScanRange.copyFrom(SensorRange);
    if (!AllowAnonymousMessages.wasSet())
        AllowAnonymousMessages.copyFrom(AllowPlayerMessages);
    if (!UnitsPerWebRate.wasSet())
        UnitsPerWebRate.copyFrom(UnitsPerTorpRate);
    if (!AllowHyperjumpGravWells.wasSet())
        AllowHyperjumpGravWells.copyFrom(AllowGravityWells);
    if (!AllowPrivateerTowCapture.wasSet())
        AllowPrivateerTowCapture.copyFrom(AllowBoardingParties);
    if (!AllowCrystalTowCapture.wasSet())
        AllowCrystalTowCapture.copyFrom(AllowBoardingParties);
    if (!CloakedMineTravelSafeWarp.wasSet())
        CloakedMineTravelSafeWarp.copyFrom(MineTravelSafeWarp);
    if (!WebMineTravelSafeWarp.wasSet())
        WebMineTravelSafeWarp.copyFrom(MineTravelSafeWarp);
    if (!NativeTaxRate.wasSet())
        NativeTaxRate.copyFrom(ColonistTaxRate);
    if (!PlayerSpecialMission.wasSet())
        PlayerSpecialMission.copyFrom(PlayerRace);
}

/** Import 11 WORDs from HCONFIG image (per-player settings). */
static void
importArray16(ConfigStdOption& opt, const char data[], std::size_t index)
{
    for (int i = 1; i <= 11; ++i) {
        opt.set(i, getInt16(data, index));
        index += 2;
    }
}

/** Import 8 DWORDs from HCONFIG image (meteor settings). */
static void
importArray32(ConfigIntArrayOption<8>& opt, const char data[], std::size_t index)
{
    for (int i = 1; i <= 8; ++i) {
        opt.set(i, getInt32(data, index));
        index += 4;
    }
}

/** Assign all options from a HCONFIG.HST image.
    \param data [in] The image
    \param size [in] Size of image
    \param src  [in] Source marker for these options */
void
GGameConfig::assignFromHConfigImage(const char data[], std::size_t size, ConfigOption::Source src)
{
    /* Instead of checking each option's position, we only check known
       version boundaries. */
    if (size >= 10) {
        RecycleRate.set(getUint16(data, 0)); RecycleRate.setSource(src);
        RandomMeteorRate.set(getUint16(data, 2)); RandomMeteorRate.setSource(src);
        AllowMinefields.set(getUint16(data, 4)); AllowMinefields.setSource(src);
        AllowAlchemy.set(getUint16(data, 6)); AllowAlchemy.setSource(src);
        DeleteOldMessages.set(getUint16(data, 8)); DeleteOldMessages.setSource(src);
    }
    if (size >= 186) {
        DisablePasswords.set(getUint16(data, 10)); DisablePasswords.setSource(src);
        importArray16(GroundKillFactor, data, 14); GroundKillFactor.setSource(src);
        importArray16(GroundDefenseFactor, data, 38); GroundDefenseFactor.setSource(src);
        importArray16(FreeFighters, data, 62); FreeFighters.setSource(src);
        importArray16(RaceMiningRate, data, 86); RaceMiningRate.setSource(src);
        importArray16(ColonistTaxRate, data, 110); ColonistTaxRate.setSource(src);
        RebelsBuildFighters.set(getUint16(data, 132)); RebelsBuildFighters.setSource(src);
        ColoniesBuildFighters.set(getUint16(data, 134)); ColoniesBuildFighters.setSource(src);
        RobotsBuildFighters.set(getUint16(data, 136)); RobotsBuildFighters.setSource(src);
        CloakFailureRate.set(getUint16(data, 138)); CloakFailureRate.setSource(src);
        RobCloakedShips.set(getUint16(data, 140)); RobCloakedShips.setSource(src);
        ScanRange.set(getUint16(data, 142)); ScanRange.setSource(src);
        DarkSenseRange.set(getUint16(data, 144)); DarkSenseRange.setSource(src);
        AllowHiss.set(getUint16(data, 146)); AllowHiss.setSource(src);
        AllowRebelGroundAttack.set(getUint16(data, 148)); AllowRebelGroundAttack.setSource(src);
        AllowSuperRefit.set(getUint16(data, 150)); AllowSuperRefit.setSource(src);
        AllowWebMines.set(getUint16(data, 152)); AllowWebMines.setSource(src);
        CloakFuelBurn.set(getUint16(data, 154)); CloakFuelBurn.setSource(src);
        SensorRange.set(getUint16(data, 156)); SensorRange.setSource(src);
        AllowNewNatives.set(getUint16(data, 158)); AllowNewNatives.setSource(src);
        AllowPlanetAttacks.set(getUint16(data, 160)); AllowPlanetAttacks.setSource(src);
        BorgAssimilationRate.set(getUint16(data, 162)); BorgAssimilationRate.setSource(src);
        WebMineDecayRate.set(getUint16(data, 164)); WebMineDecayRate.setSource(src);
        MineDecayRate.set(getUint16(data, 166)); MineDecayRate.setSource(src);
        MaximumMinefieldRadius.set(getUint16(data, 168)); MaximumMinefieldRadius.setSource(src);
        TransuraniumDecayRate.set(getUint16(data, 170)); TransuraniumDecayRate.setSource(src);
        StructureDecayPerTurn.set(getUint16(data, 172)); StructureDecayPerTurn.setSource(src);
        AllowEatingSupplies.set(getUint16(data, 174)); AllowEatingSupplies.setSource(src);
        AllowNoFuelMovement.set(getUint16(data, 176)); AllowNoFuelMovement.setSource(src);
        MineHitOdds.set(getUint16(data, 178)); MineHitOdds.setSource(src);
        WebMineHitOdds.set(getUint16(data, 180)); WebMineHitOdds.setSource(src);
        MineScanRange.set(getUint16(data, 182)); MineScanRange.setSource(src);
        AllowMinesDestroyMines.set(getUint16(data, 184)); AllowMinesDestroyMines.setSource(src);
    }
    if (size >= 288) {
        AllowEngineShieldBonus.set(getUint16(data, 186)); AllowEngineShieldBonus.setSource(src);
        EngineShieldBonusRate.set(getUint16(data, 188)); EngineShieldBonusRate.setSource(src);
        AllowColoniesSweepWebs.set(getUint16(data, 192)); AllowColoniesSweepWebs.setSource(src);
        MineSweepRate.set(getUint16(data, 194)); MineSweepRate.setSource(src);
        WebMineSweepRate.set(getUint16(data, 196)); WebMineSweepRate.setSource(src);
        HissEffectRate.set(getUint16(data, 198)); HissEffectRate.setSource(src);
        RobFailureOdds.set(getUint16(data, 200)); RobFailureOdds.setSource(src);
        PlanetsAttackRebels.set(getUint16(data, 202)); PlanetsAttackRebels.setSource(src);
        PlanetsAttackKlingons.set(getUint16(data, 204)); PlanetsAttackKlingons.setSource(src);
        MineSweepRange.set(getUint16(data, 206)); MineSweepRange.setSource(src);
        WebMineSweepRange.set(getUint16(data, 208)); WebMineSweepRange.setSource(src);
        AllowScienceMissions.set(getUint16(data, 210)); AllowScienceMissions.setSource(src);
        MineHitOddsWhenCloakedX10.set(getUint16(data, 212)); MineHitOddsWhenCloakedX10.setSource(src);
        DamageLevelForCloakFail.set(getUint16(data, 214)); DamageLevelForCloakFail.setSource(src);
        AllowFedCombatBonus.set(getUint16(data, 216)); AllowFedCombatBonus.setSource(src);
        MeteorShowerOdds.set(getUint16(data, 218)); MeteorShowerOdds.setSource(src);
        importArray32(MeteorShowerOreRanges, data, 220); MeteorShowerOreRanges.setSource(src);
        LargeMeteorsImpacting.set(getUint16(data, 252)); LargeMeteorsImpacting.setSource(src);
        importArray32(LargeMeteorOreRanges, data, 254); LargeMeteorOreRanges.setSource(src);
        AllowMeteorMessages.set(getUint16(data, 286)); AllowMeteorMessages.setSource(src);
    }
    if (size >= 298) {
        AllowOneEngineTowing.set(getUint16(data, 288)); AllowOneEngineTowing.setSource(src);
        AllowHyperWarps.set(getUint16(data, 290)); AllowHyperWarps.setSource(src);
        ClimateDeathRate.set(getUint16(data, 292)); ClimateDeathRate.setSource(src);
        AllowGravityWells.set(getUint16(data, 294)); AllowGravityWells.setSource(src);
        CrystalsPreferDeserts.set(getUint16(data, 296)); CrystalsPreferDeserts.setSource(src);
    }
    if (size >= 302) {
        AllowMinesDestroyWebs.set(getUint16(data, 298)); AllowMinesDestroyWebs.setSource(src);
        ClimateLimitsPopulation.set(getUint16(data, 300)); ClimateLimitsPopulation.setSource(src);
    }
    if (size >= 328) {
        MaxPlanetaryIncome.set(getUint32(data, 302)); MaxPlanetaryIncome.setSource(src);
        IonStormActivity.set(getUint16(data, 306)); IonStormActivity.setSource(src);
        AllowChunneling.set(getUint16(data, 308)); AllowChunneling.setSource(src);
        AllowDeluxeSuperSpy.set(getUint16(data, 310)); AllowDeluxeSuperSpy.setSource(src);
        IonStormsHideMines.set(getUint16(data, 312)); IonStormsHideMines.setSource(src);
        AllowGloryDevice.set(getUint16(data, 314)); AllowGloryDevice.setSource(src);
        AllowAntiCloakShips.set(getUint16(data, 316)); AllowAntiCloakShips.setSource(src);
        AllowGamblingShips.set(getUint16(data, 318)); AllowGamblingShips.setSource(src);
        AllowCloakedShipsAttack.set(getUint16(data, 320)); AllowCloakedShipsAttack.setSource(src);
        AllowShipCloning.set(getUint16(data, 322)); AllowShipCloning.setSource(src);
        AllowBoardingParties.set(getUint16(data, 324)); AllowBoardingParties.setSource(src);
        AllowImperialAssault.set(getUint16(data, 326)); AllowImperialAssault.setSource(src);
    }
    if (size >= 336) {
        RamScoopFuelPerLY.set(getUint16(data, 328)); RamScoopFuelPerLY.setSource(src);
        AllowAdvancedRefinery.set(getUint16(data, 330)); AllowAdvancedRefinery.setSource(src);
        AllowBioscanners.set(getUint16(data, 332)); AllowBioscanners.setSource(src);
        HullTechNotSlowedByMines.set(getUint16(data, 334)); HullTechNotSlowedByMines.setSource(src);
    }
    if (size >= 340) {
        AllowVPAFeatures.set(getUint16(data, 338)); AllowVPAFeatures.setSource(src);
    }
}

/** Load PCONFIG.SRC.
    \param pconf pconfig.src file
    \param shiplist shiplist.txt file, may be null. */
static void
loadPConfig(Ptr<Stream> pconf, Ptr<Stream> shiplist, ConfigOption::Source src)
{
    ConfigFileParser parser(config, true, src);
    parser.setCharacterSet(getGameCharacterSet());
    console.write(LOG_INFO, format(_("Reading configuration from %s..."), pconf->getName()));
    parser.configure("phost", true);
    parser.parseFile(*pconf);
    if (shiplist) {
        console.write(LOG_INFO, format(_("Reading configuration from %s..."), shiplist->getName()));
        parser.configure("phost", false);
        parser.parseFile(*shiplist);
    }
    config.assignDependantOptions();
    host.setHavePConfig();
}

/** Load HCONFIG.HST.
    \todo do host version guessing here? */
static void
loadHConfig(Stream& hconf, ConfigOption::Source src)
{
    Stream::fpos_t size = hconf.getSize();
    if (size < 0 || size > MAX_HCONFIG_SIZE)
        throw FileFormatException(hconf, _("File has invalid size"));

    ArrayPtr<char> data (size);
    hconf.readT(data.ptr(), size);

    console.write(LOG_INFO, format(_("Reading configuration from %s..."), hconf.getName()));
    config.assignFromHConfigImage(data.ptr(), size, src);
    config.assignDependantOptions();
    host.setHaveHConfig(size);
}

/** Load default config. Initializes the config variable with all
    default values. The defaults are the default values from PHost,
    which should also be the THost defaults. */
static void
loadDefaultConfig()
{
    config.assignDefaults();
}

static void
loadRaceMapping(Stream& s)
{
    char mapping[2*NUM_PLAYERS];
    if (s.read(mapping, sizeof(mapping)) == sizeof(mapping)) {
        for (int i = 1; i <= NUM_PLAYERS; ++i) {
            config.PlayerRace.set(i, getInt16(mapping, 2*(i-1)));
        }
        config.PlayerSpecialMission.copyFrom(config.PlayerRace);
        config.PlayerRace.setSource(ConfigOption::srcGame);
        config.PlayerSpecialMission.setSource(ConfigOption::srcGame);
        host.host_type = ht_SRace;
    }
}

/** Initialize configuration. Loads all configuration files. */
void
initConfig()
{
    Ptr<Stream> ps;

    loadDefaultConfig();

    // This used to load some files from spec_file_dir.
    // Now use game_file_dir exclusively; this agrees with PCC1, PHost, and c2ng.
    ps = game_file_dir->openFileNT("pconfig.src", Stream::C_READ);
    if (ps.ptr()) {
        loadPConfig(ps, game_file_dir->openFileNT("shiplist.txt", Stream::C_READ), ConfigOption::srcGame);
        return;
    } else {
        // Set Tim-Host defaults
        config.RoundGravityWells.set(1);
        config.CPEnableRemote.set(0);
    }

    ps = game_file_dir->openFileNT("friday.dat", Stream::C_READ);
    if (ps.ptr()) {
        loadRaceMapping(*ps);
    }

    ps = game_file_dir->openFileNT("hconfig.hst", Stream::C_READ);
    if (ps.ptr()) {
        loadHConfig(*ps, ConfigOption::srcGame);
        return;
    }
    console.write(LOG_WARN, _("No host configuration file found -- using defaults"));
}

/** Get name of an experience level.
    \param nr level number (0..NUM_EXPERIENCE_LEVELS) */
string_t
getExperienceLevelName(int nr)
{
    string_t s = config.ExperienceLevelNames(nr);
    if (s.empty()) {
        s = format(_("Level %d"), nr);
    }
    return s;
}

/** Get experience bonus.
    \param param Experience option to test
    \param level Experience level
    \return Value */
int
getExperienceBonus(const ConfigExpOption& param, int level)
{
    if (level <= 0 || level > 10)
        return 0;
    else
        return param(level);
}

/** Get set of all players of a particular race. */
GPlayerSet
getPlayersOfRace(int nr)
{
    return getPlayersWhere(config.PlayerRace, nr);
}

/** Get set of all players where an option is enabled.
    \param opt Boolean option */
GPlayerSet
getPlayersWhereEnabled(const ConfigStdOption& opt)
{
    GPlayerSet result;
    for (int i = 1; i <= NUM_PLAYERS; ++i)
        if (opt(i) != 0)
            result += i;
    return result;
}

/** Get set of all players where an option has a given value.
    \param opt Option
    \param value Value to check for */
GPlayerSet
getPlayersWhere(const ConfigStdOption& opt, int value)
{
    GPlayerSet result;
    for (int i = 1; i <= NUM_PLAYERS; ++i)
        if (opt(i) == value)
            result += i;
    return result;
}

extern const ModuleInitializer mod_config[] = {
    // Load default config upon startup and when leaving a directory
    { OnStartup | OnEnterNoGame, loadDefaultConfig },
    // Load config on entering directory
    { OnEnterGame, initConfig },
    { 0, 0 }
};
