/**
  *  \file game/msgtemplate.h
  *  \brief Message Template
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_MSGTEMPLATE_H
#define PCC_V2_GAME_MSGTEMPLATE_H

#include <vector>
#include "cpluslib/types.h"
#include "game/parser.h"
#include "cpluslib/string.h"

typedef std::vector<string_t> GMessageLines;

void splitMessage(GMessageLines& out, const string_t& in);

/** Message template. This object contains information to parse a single
    message into a set of variables. */
class GMessageTemplate {
 public:
    // Instructions
    struct Instruction {
        uint8_t opcode;
        int8_t offset;
        uint16_t index;
        uint16_t count;
        Instruction(uint8_t opcode, int8_t offset, uint16_t index, uint16_t count)
            : opcode(opcode), offset(offset), index(index), count(count)
            { }
    };

    // Instruction opcodes
    enum {
        // Simple instructions
        iMatchKind  = 0x00,                 // index=chr(key)
        iMatchSubId = 0x01,                 // index=chr(subid)
        iMatchBigId = 0x02,                 // index=BigId (integer)
        iValue      = 0x03,                 // index=index into string table

        // Instructions that have a scope
        iCheck      = 0x10,                 // plus scope, index=index into string table
        iFail       = 0x20,                 // plus scope, index=index into string table
        iFind       = 0x30,                 // plus scope, index=index into string table
        iParse      = 0x40,                 // plus scope, index=index into string table, count=number of variables
        iArray      = 0x50,                 // plus scope, index=index into string table, count=number of variables

        iMask       = 0xF0
    };

    // Scopes
    enum {
        sAny        = 0,
        sRelative   = 1,
        sFixed      = 2
    };

    GMessageTemplate(GMessageObject message_object, string_t name);
    ~GMessageTemplate();
    void addMatchInstruction(uint8_t opcode, uint16_t value);
    void addValueInstruction(uint8_t opcode, string_t value);
    void addCheckInstruction(uint8_t opcode, int8_t offset, string_t value);
    void addVariable(string_t name);
    void addVariables(string_t names);
    void setContinueFlag(bool flag);
    bool getContinueFlag() const;

    int getNumVariables() const;
    int getNumWildcards() const;
    int getNumRestrictions() const;
    int getVariableSlotByName(string_t name) const;
    string_t getVariableName(int index) const;
    string_t getTemplateName() const;
    GMessageObject getMessageObject() const;

    bool match(const GMessageLines& message, int player, std::vector<string_t>& values) const;

    void dump() const;

    static int32_t getIntValue(const string_t& value);

 private:
    bool check(const GMessageLines& message, uint16_t& line, const Instruction& insn) const;
    bool matchLine(const string_t& line, uint_t index, uint_t nvar, uint_t valueIndex, std::vector<string_t>& values) const;
    bool matchPart(const string_t& line, uint_t startAt, uint_t index, uint_t nvar, std::vector<string_t>& values) const;
    void consolidateArray(std::vector<string_t>& values, uint_t nvar, uint_t nelems) const;

    GMessageObject message_object;
    string_t name;
    bool continue_flag;
    std::vector<Instruction> instructions;
    std::vector<string_t> strings;
    std::vector<string_t> variables;
    std::vector<string_t> types;
};


inline bool
GMessageTemplate::getContinueFlag() const
{
    return continue_flag;
}

inline GMessageObject
GMessageTemplate::getMessageObject() const
{
    return message_object;
}


enum GMessageHeaderInformation {
    MsgHdrKind,                 ///< Message kind. Character code.
    MsgHdrSubId,                ///< Message SubId. Character code.
    MsgHdrId,                   ///< Message Id. Integer.
    MsgHdrBigId,                ///< Message BigId, SubId and Id concatenated. Integer.
    MsgHdrAge                   ///< Message age. Flag, 0 for current, 1 for old.
};

int getMessageHeaderInformation(const GMessageLines& msg, GMessageHeaderInformation what);

#endif
