/**
  *  \file game/pvcr.cc
  *  \brief PVCR Implementation
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/pvcr.h"
#include "game/config.h"
#include "game/spec.h"
#include "game/formula.h"

/** \class VcrPlayerPHost
    \brief PHost VCR Player

    This is the PHost VCR player, started as a port from the PCC
    v1.0.15 player, and went through serious tweaking.

    This player supports all PHost versions from 2.4 upwards, up to at
    least version 4.0.

    Things remaining to do:
    - some things should be done in subroutines (killFighter...)
    - cache weapon specs

    Believe me or not, I have not copied code off the PDK or PHost.
    The PDK is still (or, "again") faster... */

/*
  This seems to be pretty optimisation-resistant.

  All tests run on a vanilla Pentium 200. Each test is the mean of at
  least two runs.

  + original:                   AC: 15.74,  Non-AC: 3.87
  + inlining:                   AC: 15.27,  Non-AC: 3.72
  + PVCR_INTEGER:               AC: 15.25,  Non-AC: 3.74(!)
  + f_getDamage():              AC: 16.11(!) [not used]
  + caching config:             AC: 13.58,  Non-AC: 3.59
  + fighterMove:                AC: 11.91,  Non-AC: 3.41
  + beamFindNearestFighter:     AC:  6.70,  Non-AC: 2.04
  + template<int side>:         AC:  5.36,  Non-AC: 1.64  (currently not used)

  For comparison, PDK 4.2:      AC: 9.50,   Non-AC: 2.44
  PDK 4.3 (beamFindNearest.)    AC: 5.53,   Non-AC: 1.72

  The one reason why the PDK is faster is probably that the PDK
  contains for loops (for(side=0; side<2; ++side)), where this code
  contains two function calls for each phase. gcc can optimise that
  better. The speedup gained by templatizing seems to support that
  supposition.

  More optimisation: 100 test runs on Athlon XP 2800+ (2 GHz):
  (testvcr tests/pvcr/vcr5.dat 100)

  + start with:                 AC: 4.72,   Non-AC: 1.52
  + ShMovSpeed=0 for planets:   AC: 4.86,   Non-AC: 1.49  (not used)
  + fast forward:               AC: 4.82,   Non-AC: 1.45
  + randomRange100              AC: 4.53,   Non-AC: 1.30
  + uint32 seed                 AC: 4.33,   Non-AC: 1.27
  + fighterIntercept opt.: no obervable speedup in this test
    case, but 5% faster (4.4 -> 4.2) for carrier/carrier fight
  + PVCR_PREPARED_RNG           AC: 3.78,   Non-AC: 1.13
  + 20120122                    AC: 3.65,   Non-AC: 1.09
  + randomRange100LT, RunningStatus int16 for some values:
                                AC: 3.41,   Non-AC: 1.03
  + store seed<<16              AC: 3.19,   Non-AC: 0.97
  + Status contains VCR         AC: 2.64,   Non-AC: 0.83
  + side->Status&               AC: 2.57,   Non-AC: 0.82

  Disabling VCR_STATIC:
  + 20120616                    AC: 2.86,   Non-AC: 0.87
  + 20120707                    AC: 2.93,   Non-AC: 0.90
  + templatizing hit            AC: 2.81,   Non-AC: 0.87

  Note that FF applies only in 2 of 15 battles of the test case; for
  those where it applies it yields a huge speedup, so I put it in.

  Some things which do not get anything:
  + PRNG skip tables for Fast Forward: precompute how N iterations
    of randomRange(100) modify the seed for a subset of M of all
    seeds; if current seed is one of those and we have still N or
    more to go, make big jump instead of small step. With M*N=65536,
    this should get a theoretical speedup of 2 for the FF operation,
    but has no measurable overall effect. Tested with N=128.

  New tests (i5 3700 MHz, gcc-4.9, x64), 1000 runs
  + 20160325                    AC: 4.34,   Non-AC: 1.25 */

/** Bitmask of VCR capabilities supported by this player. */
static const unsigned SUPPORTED_CAP = vcrcap_DeathRays | vcrcap_Experience | vcrcap_Beams;

/** Movement timer. We check whether standoff distance has been
    reached every this many ticks (it could have been done a little
    simpler, but this way integrates neatly into the timer framework
    we need for regular interval checks). */
static const int DET_MOVEMENT_TIMER = 100;

/** Inactivity timer. We check whether combat made progress every this
    many ticks after standoff distance has been reached. A simple
    heuristic increases the value if necessary. */
static const int DET_INACTIVITY_TIMER = 5000;


/** Distance. Shortcut for later optimisations. */
#define DISTANCE() (long(status[1].r.obj_x) - status[0].r.obj_x)

/** Get experience-modified value of an option.
    \param opt primary option
    \param exp experience modificator (EMod) option
    \param min,max range for result */
static int
getExperienceMod(const ConfigStdOption& opt, const ConfigExpOption& exp, const GVcrObject& obj, int min, int max)
{
    int32_t sum = opt(obj.getOwner());
    if (obj.getExperienceLevel() != 0) {
        sum += exp(obj.getExperienceLevel());
    }
    if (sum < min) {
        return min;
    } else if (sum > max) {
        return max;
    } else {
        return sum;
    }
}

/********************************** RNG **********************************/

/*
 *  The PRNG has been optimized to generate the best possible x86 machine
 *  code with my compiler.
 *  - scale everything by 65536; saves one masking operation
 *  - "a+b*-c" instead of "a-b*c"
 */

/** Random Number Generator. Basic linear congruence. */
uint32
VcrPlayerPHost::random64k()
{
    seed = (13489*65536 + seed * -259U);
    return seed;
}

#ifdef PVCR_PREPARED_RNG
/** Random Number Generator. Returns a random number uniformly in [0,max),
    where max is the parameter passed to PreparedRNG::operator=. */
int
VcrPlayerPHost::randomRange(const PreparedRNG& rng)
{
    uint32_t max = rng.limit;
    uint32_t i;
    do
        i = random64k();
    while (i >= max);
    return i / rng.divi;
}

/** Prepare random number generator. The PreparedRNG can later be
    passed to randomRange to return a number in [0,max). Doing it this
    way avoids one division instruction (or more) per RNG invocation. */
void
VcrPlayerPHost::PreparedRNG::operator=(uint32_t max)
{
    if (max == 0) {
        // Pathological case. These RNGs will never be called.
        divi = 65536U*65535;
        limit = 65536U*65535;
    } else {
        divi  = (65535U / max) * 65536;
        limit = divi * max;
    }
}
#else
/** Random Number Generator. Returns a random number uniformly
    in [0, max) */
int
VcrPlayerPHost::randomRange(uint32_t max)
{
    uint32_t divisor = (65535U / max) * 65536;
    uint32_t i;
    do
        i = random64k() / divisor;
    while (i >= max);
    return i;
}
#endif

/** Random Number Generator. Same as randomRange(100), optimized to
    avoid divisions. */
int
VcrPlayerPHost::randomRange100()
{
    uint32_t i;
    do
        i = random64k();
    while (i >= 65500U*65536U);
    return i / (65536U*655U);
}

/** Random Number Generator. Compute a random number uniformly in [0,100)
    and return true iff it is smaller than comp. */
bool
VcrPlayerPHost::randomRange100LT(int comp)
{
    uint32_t i;
    do
        i = random64k();
    while (i >= 65500U*65536U);

    // comp could be >100, which would make 65536*655*comp overflow.
    // Add one extra shift instruction instead.
    i >>= 16;
    return i < uint32_t(comp * 655);
}

/************************** Standard "Hit" Stuff *************************/

/** \name Standard "Hit" Stuff */
//@{

/** Damage formulas for non-alternative (Regular) Combat. */
struct VcrPlayerPHost::RegularFormula {
#ifdef PVCR_INTEGER
    /** Compute shield damage, regular combat, integer version. Returns actual value. */
    static inline int32_t
    computeShieldDamageS(int expl, int kill, const Status& st)
        {
            int32_t damage = radiv(st.f.ShieldDamageScaling * int32(expl) + st.f.ShieldKillScaling * int32(kill),
                                   st.f.mass_plus1) + 1;
            return std::min(damage, 10000);
        }
    /** Compute hull damage, regular combat, integer version. Returns actual value. */
    static inline int32_t
    computeHullDamageS(int expl, int kill, const Status& st)
        {
            int32_t d = radiv(computeShieldDamageS(expl, kill, st) * st.f.HullDamageScaling,
                              st.f.mass_plus1) + 1;
            return std::min(d, 10000);
        }
    /** Compute killed crew, regular combat, integer version. Returns actual value. */
    static inline int32_t
    computeCrewKilledS(int kill, bool death, const Status& st)
        {
            int32_t rv = radiv(st.f.CrewKillScaling * int32(kill), st.f.mass_plus1);
            if (death && rv == 0)
                return 1;
            else
                return rv;
        }
#else
    /** Compute shield damage, regular combat, float version. Returns actual value. */
    static inline damage_t
    computeShieldDamage(int expl, int kill, const Status& st)
        {
            double damage = (st.f.ShieldDamageScaling * double(expl)
                             + st.f.ShieldKillScaling * double(kill))
                / double(st.f.mass_plus1);
            if (damage > 10000)
                damage = 10000;
            return (int) (damage + 1.5);
        }
    /** Compute hull damage, regular combat, float version. Returns actual value. */
    static inline damage_t
    computeHullDamage(int expl, int kill, const Status& st)
        {
            double d = computeShieldDamage(expl, kill, st) * st.f.HullDamageScaling / double(st.f.mass_plus1);
            if (d > 10000)
                return 10000;
            else
                return (int) (d + 1.5);
        }
    /** Compute killed crew, regular combat, float version. Returns actual value. */
    static inline damage_t
    computeCrewKilled(int kill, bool death, const Status& st)
        {
            double d = st.f.CrewKillScaling * double(kill) / double(st.f.mass_plus1);
            int i = int(d + 0.5);
            if (death && i == 0)
                return 1;
            else
                return i;
        }
#endif
};

/** Damage formulas for Alternative Combat. */
struct VcrPlayerPHost::AlternativeFormula {
#ifdef PVCR_INTEGER
    /** Compute shield damage, alternative combat, integer version. Returns value sacled by mass+1. */
    static inline int32_t
    computeShieldDamageS(int expl, int kill, const Status& st)
        {
            int32_t damage = (st.f.ShieldDamageScaling * int32(expl) + st.f.ShieldKillScaling * int32(kill));
            return std::min(damage, int32(st.f.max_scaled));
        }
    /** Compute hull damage, alternative combat, integer version. Returns value scaled by mass+1. */
    static inline int32_t
    computeHullDamageS(int expl, int /*kill*/, const Status& st)
        {
            int32_t d = int32_t(expl) * st.f.HullDamageScaling;
            return std::min(d, int32(st.f.max_scaled));
        }
    /** Compute killed crew, alternative combat, integer version. Returns value scaled by mass+1. */
    static inline int32_t
    computeCrewKilledS(int kill, bool /*death*/, const Status& st)
        {
            return st.f.CrewKillScaling * int32_t(kill);
        }
#else
    /** Compute shield damage, alternative combat, float version. Returns actual value. */
    static inline damage_t
    computeShieldDamage(int expl, int kill, const Status& st)
        {
            double damage = (st.f.ShieldDamageScaling * double(expl)
                             + st.f.ShieldKillScaling * double(kill))
                / double(st.f.mass_plus1);
            if (damage > 10000)
                damage = 10000;
            return damage;
        }
    /** Compute hull damage, alternative combat, float version. Returns actual value. */
    static inline damage_t
    computeHullDamage(int expl, int kill, const Status& st)
        {
            double d = double(expl) * st.f.HullDamageScaling / double(st.f.mass_plus1);
            if (d > 10000)
                return 10000;
            else
                return d;
        }
    /** Compute killed crew, alternative combat, float version. Returns actual value. */
    static inline damage_t
    computeCrewKilled(int kill, bool death, const Status& st)
        {
            return st.f.CrewKillScaling * double(kill) / double(st.f.mass_plus1);
        }
#endif
};


/** Hit object, back-end.
    \param Formula formula set in use (AlternativeFormula, RegularFormula).
    \param st    object being hit
    \param kill  anti-life (x-ray) power of weapon
    \param expl  explosive power of weapon
    \param is_death_ray true iff this weapon emits death rays. */
template<typename Formula>
inline bool
VcrPlayerPHost::hitT(Status& st, int kill, int expl, bool is_death_ray)
{
#ifdef PVCR_INTEGER
    int damage_rate = 100;

    if (!is_death_ray) {
        /* Hit shields */
        if (st.r.shield_scaled > 0) {
            int32 damage_s = Formula::computeShieldDamageS(expl, kill, st);
            if (st.r.shield_scaled < damage_s) {
                // Shields are completely worn down by this hit
                damage_rate = (damage_s - st.r.shield_scaled) * 100 / damage_s;
                st.r.shield_scaled = 0;
            } else {
                // Shields still hold
                damage_rate = 0;
                st.r.shield_scaled -= damage_s;
            }
        }

        /* Shields are down -- do damage */
        if (damage_rate <= 0)
            return false;

        st.r.damage_scaled2 += Formula::computeHullDamageS(expl, kill, st) * damage_rate;

        if (st.r.damage_scaled2 >= st.f.damage_limit_scaled) {
            return true;
        }
    }

    if (!st.r.obj.isPlanet()) {
        int32 killed_s = Formula::computeCrewKilledS(kill, is_death_ray, st) * damage_rate;
        st.r.crew_scaled2 -= killed_s;
        if (st.r.crew_scaled2 < 50*st.f.scale) {
            st.r.crew_scaled2 = 0;
            return true;
        }
    }

    return false;
#else
    int damage_rate = 100;

    if (!is_death_ray) {
        /* Hit shields */
        if (st.r.shield > 0) {
            double damage = Formula::computeShieldDamage(expl, kill, st);
            if (st.r.shield <= damage) {
                // Shields are completely worn down by this hit
                damage_rate = int((damage - st.r.shield) * 100 / damage);
                st.r.shield = 0;
            } else {
                // Shields still hold
                damage_rate = 0;
                st.r.shield -= damage;
            }
        }

        /* Shields are down -- do damage */
        if (damage_rate <= 0)
            return false;

        st.r.damage += Formula::computeHullDamage(expl, kill, st) * damage_rate / 100.0;

        if (int(st.r.damage + 0.5) >= st.f.damage_limit) {
            return true;
        }
    }

    if (!st.r.obj.isPlanet()) {
        double killed = Formula::computeCrewKilled(kill, is_death_ray, st) * damage_rate / 100.0;
        st.r.crew -= killed;
        if (st.r.crew < 0.5) {
            st.r.crew = 0;
            return true;
        }
    }
    return false;
#endif
}

/** Hit object.
    \param st    object being hit
    \param kill  anti-life (x-ray) power of weapon
    \param expl  explosive power of weapon
    \param is_death_ray true iff this weapon emits death rays. "hit"
    can't itself check for this, because a beam may end up here with
    expl=0 when fired while partially charged */
bool
VcrPlayerPHost::hit(Status& st, int kill, int expl, bool is_death_ray)
{
    if (kill <= 0) {
        kill = 1;
    }
    if (expl <= 0) {
        expl = 1;
    }
    if ((capabilities & vcrcap_DeathRays) == 0) {
        is_death_ray = false;
    }

    if (config.AllowAlternativeCombat()) {
        return hitT<AlternativeFormula>(st, kill, expl, is_death_ray);
    } else {
        return hitT<RegularFormula>(st, kill, expl, is_death_ray);
    }
}
//@}

/******************************** Fighters *******************************/

/** \name Fighters */
//@{

/** Compute bay recharge rate. Documented formula. Used for initialisation. */
int
VcrPlayerPHost::computeBayRechargeRate(int num, const GVcrObject& obj)
{
    int i = getExperienceMod(config.BayRechargeBonus, config.EModBayRechargeBonus, obj, -500, 500) * num
        + getExperienceMod(config.BayRechargeRate, config.EModBayRechargeRate, obj, 0, 16384);
    return i > 1 ? i : 1;
}

/** Recharge Fighter Bays. */
void
VcrPlayerPHost::fighterRecharge(Status& st)
{
    register const int mx = st.r.obj.getBayCount();
    for (int i = 0; i < mx; ++i)
        if (st.r.bay_status[i] < 1000)
            st.r.bay_status[i] += randomRange(st.f.bay_recharge);
}

/** Launch Fighters. */
void
VcrPlayerPHost::fighterLaunch(Status& st)
{
    /* can we launch a fighter? */
    if (st.r.obj.getFighterCount() == 0
        || st.r.launched_fighters >= st.f.MaxFightersLaunched
        || st.r.launch_countdown > 0)
        return;

    /* yes, we can! */
    int bay_mx = st.r.obj.getBayCount();
    for (int bay = 0; bay < bay_mx; ++bay) {
        if (st.r.bay_status[bay] >= 1000) {
            for (int track = 0; track < st.f.MaxFightersLaunched; ++track) {
                if (st.r.fighter_status[track] == 0) {
                    /* okay, we have a bay which is ready, and an empty track. */
                    st.r.fighter_status[track]     = 1;
                    st.r.fighter_x[track]          = st.r.obj_x;
                    st.r.fighter_shots_left[track] = st.f.StrikesPerFighter;
                    st.r.bay_status[bay]           = 0;
                    st.r.launched_fighters++;
                    st.r.obj.addFighters(-1);
                    st.r.launch_countdown = st.f.BayLaunchInterval;
                    vis->startFighter(st.f.side, track);
                    setStatFighters(st.f.side, st.r.obj.getFighterCount());
                    return;
                }
            }
        }
    }
}

/** Move Fighters. Takes back returned fighters. */
void
VcrPlayerPHost::fighterMove(Status& st)
{
    if (st.r.launched_fighters == 0)
        return;

    int side = st.f.side;
    int dir = (side == 0) ? st.f.FighterMovementSpeed : -st.f.FighterMovementSpeed;

    for (int track = 0, limit = st.f.MaxFightersLaunched; track < limit; ++track)
        if (st.r.fighter_status[track] == 1) {
            st.r.fighter_x[track] += dir;
        } else if (st.r.fighter_status[track] == 2) {
            st.r.fighter_x[track] -= dir;
            if ((side == 0)
                ? st.r.fighter_x[track] < st.r.obj_x
                : st.r.fighter_x[track] > st.r.obj_x)
            {
                /* fighter comes back to baseship */
                st.r.launched_fighters--;
                st.r.obj.addFighters(+1);
                vis->landFighter(side, track);
                st.r.fighter_status[track] = 0;
            }
        }
}

/** Do fighter intercepts. */
void
VcrPlayerPHost::fighterIntercept()
{
    static const int NEVER = -0x4000;

    /* Fighter intercept only happens if both have fighters */
    if (status[0].r.launched_fighters == 0 || status[1].r.launched_fighters == 0)
        return;

    /* Compute screen positions. The actual fighter intercept code compares fighter
       positions, with coarse granularity, because exact matches never happen. By
       ignoring the lowest 7 bits of the X coordinate, we effectively compare just
       the upper 9 bits. Fighter intercept processes all fighter pairs which is very
       expensive. We use the following optimisations:
       - precompute the fighter positions (this optimisation is also in PVCR)
       - process only occupied slots
       - process only situations where we know to have fighter pairs. Since we have
         just 9 bits that specify a position, we can easily keep track of which
         points have a left fighter, and only process right fighters at these points.
         Note that eliminating left fighters does not work even if we know that they
         have no counterpart because we must call randomRange() an appropriate number
         of times. */
    int16_t lmatch[VCR_MAX_FTRS], rmatch[VCR_MAX_FTRS];
    int8_t  lslot[VCR_MAX_FTRS],  rslot[VCR_MAX_FTRS];
    int     lcount = 0,           rcount = 0;

    uint32_t bins[16];  /* That's 512 bits */
    for (int i = 0; i < 16; ++i)
        bins[i] = 0;

    /* Compute left positions and mark occupied slots */
    for (int i = 0, limit = status[0].f.MaxFightersLaunched; i < limit; ++i) {
        if (status[0].r.fighter_status[i] != 0) {
            int hash = status[0].r.fighter_x[i] >> 7;
            lslot[lcount] = i;
            lmatch[lcount] = hash;
            lcount++;
            bins[(hash >> 5) & 15] |= 1 << (hash & 31);
        }
    }

    /* Compute right positions, but eliminate impossible slots */
    for (int i = 0, limit = status[1].f.MaxFightersLaunched; i < limit; ++i) {
        if (status[1].r.fighter_status[i] != 0) {
            int hash = status[1].r.fighter_x[i] >> 7;
            if (bins[(hash >> 5) & 15] & (1 << (hash & 31))) {
                rslot[rcount] = i;
                rmatch[rcount] = hash;
                rcount++;
            }
        }
    }

    /* Degenerate case: no match */
    if (!rcount) {
        for (int ls = 0; ls < lcount; ++ls)
            randomRange100();
        return;
    }

    /* Full version */
    for (int ls = 0; ls < lcount; ++ls) {
        if (randomRange100LT(one_f)) {
            for (int rs = 0; rs < rcount; ++rs) {
                if (rmatch[rs] == lmatch[ls]) {
                    int lf = lslot[ls];
                    int rf = rslot[rs];
                    /* two fighters at the same place, intercepting */
                    /* theoretically, here is a slight imbalance. If
                       the "right" fighter survives, it can fire again
                       while a "left" one can not. Whether this is
                       relevant in practice is unknown. */
                    if (randomRange100LT(right_probab)) {
                        vis->fireBeam(1, rf, lf, 1, status[1].f.FighterBeamExplosive, status[1].f.FighterBeamKill);
                        status[0].r.launched_fighters--;
                        vis->killFighter(0, lf);
                        status[0].r.fighter_status[lf] = 0;
                    } else {
                        vis->fireBeam(0, lf, rf, 1, status[0].f.FighterBeamExplosive, status[0].f.FighterBeamKill);
                        status[1].r.launched_fighters--;
                        vis->killFighter(1, rf);
                        status[1].r.fighter_status[rf] = 0;
                        rmatch[rs] = NEVER;
                    }
                    break;
                }
            }
        }
    }
}

/** Fighters attack enemy. */
bool
VcrPlayerPHost::fighterAttack(Status& st, Status& opp)
{
    /* only if we have fighters */
    if (st.r.launched_fighters == 0)
        return false;

    int enemy_x = opp.r.obj_x;
    for (int i = 0, limit = st.f.MaxFightersLaunched; i < limit; ++i) {
        if (st.r.fighter_status[i] == 1) {
            if (std::abs(int32(st.r.fighter_x[i]) - enemy_x) <= st.f.FighterFiringRange) {
                st.r.fighter_shots_left[i]--;
                if (!st.r.fighter_shots_left[i])
                    st.r.fighter_status[i] = 2;

                bool hitres = hit(opp, st.f.FighterBeamKill, st.f.FighterBeamExplosive, false);
                vis->fireBeam(st.f.side, i, -1, 1, st.f.FighterBeamExplosive, st.f.FighterBeamKill);
                if (hitres)
                    return true;
            } else if ((capabilities & vcrcap_Beams) != 0) {
                /* if fighter moved past enemy, retreat */
                if (st.f.side == 0
                    ? st.r.fighter_x[i] > enemy_x + st.f.FighterFiringRange
                    : st.r.fighter_x[i] < enemy_x - st.f.FighterFiringRange)
                {
                    st.r.fighter_shots_left[i] = 0;
                    st.r.fighter_status[i] = 2;
                }
            }
        }
    }
    return false;
}
//@}

/********************************* Beams *********************************/

/** \name Beams */
//@{

/** Compute beam hit odds. Documented formula. Used during initialisation. */
int
VcrPlayerPHost::computeBeamHitOdds(int beam, const GVcrObject& obj)
{
    GBeam& b = getBeam(beam);
    int i = getExperienceMod(config.BeamHitBonus, config.EModBeamHitBonus, obj, -4095, 4095)
              * (b.getKill() + b.getDamage()) / 100
        + getExperienceMod(config.BeamHitOdds, config.EModBeamHitOdds, obj, 0, 100);
    return i < 0 ? 0 : i;
}

/** Compute beam recharge rate. Documented formula. Used during
    initialisation. */
int
VcrPlayerPHost::computeBeamRechargeRate(int beam, const GVcrObject& obj)
{
    GBeam& b = getBeam(beam);
    int i = (((b.getKill() + b.getDamage()) * getExperienceMod(config.BeamRechargeBonus, config.EModBeamRechargeBonus, obj, -4095, 4095)) / 100
             + getExperienceMod(config.BeamRechargeRate, config.EModBeamRechargeRate, obj, 0, 16384))
        * obj.getBeamChargeRate();
    return i < 1 ? 1 : i;
}

/** Recharge beams. */
void
VcrPlayerPHost::beamRecharge(Status& st)
{
    register const int mx = st.r.obj.getBeamCount();
    for (int i = 0; i < mx; ++i) {
        if (st.r.beam_status[i] < 1000) {
            st.r.beam_status[i] += randomRange(st.f.beam_recharge);
            vis->updateBeam(st.f.side, i);
        }
    }
}

/** Find nearest-possible fighter. Returns fighter index [0,VCR_MAX_FTRS),
    or -1 if none. */
int
VcrPlayerPHost::beamFindNearestFighter(const Status& st, const Status& opp)
{
    long mindist = 100000;
    int fighter = -1;

    /* When enemy does not have a fighter, don't fire at one... */
    if (opp.r.launched_fighters == 0)
        return -1;

    long my_x = st.r.obj_x;

    /* Look for a matching fighters. When we're asked to fire on
       attacking fighters only, accept only those first. */
    for (int i = 0, limit = opp.f.MaxFightersLaunched; i < limit; ++i)
        if (opp.r.fighter_status[i] == 1
            || (!config.FireOnAttackFighters() && opp.r.fighter_status[i] != 0))
        {
            long d = std::abs(my_x - opp.r.fighter_x[i]);
            if (d < mindist) {
                mindist = d;
                fighter = i;
            }
        }

    /* No fighter found. Check retreating fighters. Any of them does.
       Interestingly enough, PHost only checks distances for the first
       pass, not for the second one. */
    if (fighter < 0 && config.FireOnAttackFighters()) {
        for (int i = 0, limit = opp.f.MaxFightersLaunched; i < limit; ++i)
            if (opp.r.fighter_status[i] != 0) {
                fighter = i;
                break;
            }
        return fighter;
    } else {
        /* Found a fighter on first pass. Must be in range. */
        if (mindist > st.f.BeamHitFighterRange)
            return -1;
        else
            return fighter;
    }
}

/** Fire beams on /side/. \returns true iff battle ends. */
bool
VcrPlayerPHost::beamFire(Status& st, Status& opp)
{
    const int beam_mx = st.r.obj.getBeamCount();
    if (beam_mx <= 0) {
        return false;
    }

    const GBeam& gb = getBeam(st.r.obj.getBeamType());
    for (int beam = 0; beam < beam_mx; ++beam) {
        /* Can we fire at a fighter? */
        if (st.r.beam_status[beam] >= st.f.BeamHitFighterCharge) {
            /* PVCR tests `&& distance <= 100000', but that's not needed */
            bool missing = !randomRange100LT(st.f.beam_hit_odds);
            int fighter = beamFindNearestFighter(st, opp);
            if (fighter >= 0) {
                /* We fire at a fighter. */
                vis->fireBeam(st.f.side, -1-beam, fighter, missing ? -1 : +1,
                              gb.getDamage(), gb.getKill());
                st.r.beam_status[beam] = 0;
                vis->updateBeam(st.f.side, beam);
                if (!missing) {
                    vis->killFighter(opp.f.side, fighter);
                    opp.r.fighter_status[fighter] = 0;
                    opp.r.launched_fighters--;
                }
                return false;
            }
        }

        /* Can we fire at the enemy? */
        if ((opp.r.launched_fighters <= 0 || (capabilities & vcrcap_Beams) != 0)
            && st.r.beam_status[beam] >= st.f.BeamHitShipCharge
            && DISTANCE() <= st.f.BeamFiringRange)
        {
            bool missing = !randomRange100LT(st.f.beam_hit_odds);
            int kill = (gb.getKill()   * (st.r.beam_status[beam]/10) / 100) * st.r.obj.getBeamKillRate();
            int dest = gb.getDamage() * (st.r.beam_status[beam]/10) / 100;

            st.r.beam_status[beam] = 0;
            vis->updateBeam(st.f.side, beam);

            if (!missing) {
                bool hitr = hit(opp, kill, dest, gb.getDamage() == 0);
                vis->fireBeam(st.f.side, -1-beam, -1, 1, dest, kill);
                if (hitr)
                    return true;
            } else {
                vis->fireBeam(st.f.side, -1-beam, -1, -1, dest, kill);
            }
            return false;
        }
    }
    return false;
}
//@}

/******************************* Torpedoes *******************************/

/** \name Torpedoes */
//@{

/** Compute torpedo hit odds. Documented formula. Used in initialisation. */
int
VcrPlayerPHost::computeTorpHitOdds(int torp, const GVcrObject& obj)
{
    const GTorpLauncher& tl = getTorp(torp);
    int i = ((getExperienceMod(config.TorpHitBonus, config.EModTorpHitBonus, obj, -4095, 4095) * (tl.getKill() + tl.getDamage())) / 100
             + getExperienceMod(config.TorpHitOdds, config.EModTorpHitOdds, obj, 0, 100));
    return i < 0 ? 0 : i;
}

/** Compute torpedo recharge rate. Documented formula. Used in
    initialisation. */
int
VcrPlayerPHost::computeTubeRechargeRate(int torp, const GVcrObject& obj)
{
    const GTorpLauncher& tl = getTorp(torp);
    int i = (((getExperienceMod(config.TubeRechargeBonus, config.EModTubeRechargeBonus, obj, -4095, 4095) * (tl.getKill() + tl.getDamage())) / 100
             + getExperienceMod(config.TubeRechargeRate, config.EModTubeRechargeRate, obj, 0, 16384)))
        * obj.getTorpChargeRate();
    return i < 1 ? 1 : i;
}

/** Recharge torpedo launchers. */
void
VcrPlayerPHost::torpsRecharge(Status& st)
{
    register const int mx = st.r.obj.getTorpLauncherCount();
    for (int i = 0; i < mx; ++i) {
        if (st.r.torp_status[i] < 1000) {
            st.r.torp_status[i] += randomRange(st.f.torp_recharge);
            vis->updateTorp(st.f.side, i);
        }
    }
}

/** Fire torpedoes. */
bool
VcrPlayerPHost::torpsFire(Status& st, Status& opp)
{
    if (DISTANCE() > st.f.TorpFiringRange || st.r.obj.getTorpCount() == 0)
        return false;

    for (int launcher = 0, limit = st.r.obj.getTorpLauncherCount(); launcher < limit; ++launcher)
        if (st.r.torp_status[launcher] >= 1000) {
            /* we're firing a torpedo */
            int rr = randomRange100();

            st.r.obj.addTorpedoes(-1);
            st.r.torp_status[launcher] = 0;
            vis->updateTorp(st.f.side, launcher);
            if (rr <= st.f.torp_hit_odds) {
                /* Scaling factor for torpedo effect. Tim scales with 2 for some reason. */
                const GTorpLauncher& tl = getTorp(st.r.obj.getTorpType());
                int kill = tl.getKill();
                int damage = tl.getDamage();
                if (!config.AllowAlternativeCombat()) {
                    kill *= 2;
                    damage *= 2;
                }

                /* we hit the enemy */
                bool hitr = hit(opp, kill, damage, damage == 0);
                setStatTorpHit(st.f.side);
                vis->fireTorp(st.f.side, rr, launcher);
                return hitr;
            } else {
                /* miss */
                vis->fireTorp(st.f.side, -1-rr, launcher);
                return false;
            }
        }
    return false;
}
//@}

/******************************** Control ********************************/

/** \name Main Stuff */
//@{

/** Move objects towards each other. */
inline void
VcrPlayerPHost::moveObjects()
{
    long remain = DISTANCE() - config.StandoffDistance();
    if (remain <= 0)
        return;

    /* move objects. Ensure StandoffDistance is not violated */
    long move_left = std::min (remain, long(status[0].f.ShipMovementSpeed));
    status[0].r.obj_x += move_left;
    remain -= move_left;

    if (!status[1].r.obj.isPlanet()) {
        long move_right = std::min (remain, long(status[1].f.ShipMovementSpeed));
        status[1].r.obj_x -= move_right;
    }
}

/** Check whether object still has offensive capabilities. */
inline bool
VcrPlayerPHost::canStillFight(const Status& st, const Status& opp)
{
    const bool drcheck = !(capabilities & vcrcap_DeathRays) || !opp.r.obj.isPlanet();
    return (st.r.obj.getBeamCount() > 0 && (drcheck || getBeam(st.r.obj.getBeamType()).getDamage()))
        || (st.r.obj.getFighterCount() > 0 && st.r.obj.getBayCount() > 0)
        || (st.r.launched_fighters > 0)
        || (st.r.obj.getTorpCount() > 0 && (drcheck || getTorp(st.r.obj.getTorpType()).getDamage()));
}

/** Initialize inactivity detection. This is called at initialisation
    of the battle and resets the detector. */
void
VcrPlayerPHost::initActivityDetector()
{
    det_valid = false;
    det_timer = DET_MOVEMENT_TIMER;
}

/** Compare inactivity detector status for one side.
    \param a     DetStatus object
    \param side  Side to compare with
    \return true iff DetStatus represents current status */
inline bool
VcrPlayerPHost::compareDetStatus(const DetStatus& a, const Status& st)
{
    return a.fighters == st.r.launched_fighters + st.r.obj.getFighterCount()
        && a.torps    == st.r.obj.getTorpCount()
#ifdef PVCR_INTEGER
        && a.shield_scaled == st.r.shield_scaled
        && a.damage_scaled == st.r.damage_scaled2
        && a.crew_scaled   == st.r.crew_scaled2
#else
        && a.shield  == st.r.shield
        && a.damage  == st.r.damage
        && a.crew    == st.r.crew
#endif
        ;
}

/** Fill in inactivity detector status for one side.
    \param a     [out] DetStatus object
    \param side  Side to fill in */
void
VcrPlayerPHost::fillInDetStatus(DetStatus& a, const Status& st)
{
    a.fighters = st.r.launched_fighters + st.r.obj.getFighterCount();
    a.torps    = st.r.obj.getTorpCount();
#ifdef PVCR_INTEGER
    a.shield_scaled = st.r.shield_scaled;
    a.damage_scaled = st.r.damage_scaled2;
    a.crew_scaled   = st.r.crew_scaled2;
#else
    a.shield  = st.r.shield;
    a.damage  = st.r.damage;
    a.crew    = st.r.crew;
#endif
}

/** Check whether there is combat activity. This feature was added in
    PHost 4.1a to avoid infinite loops on accidentally / maliciously
    modified combat configurations. We need not link it to particular
    host versions, though, because it just avoids infinite loops and
    does not change actual results.

    This detector works by taking a snapshot of the relevant
    parameters of a ship, and periodically checking whether these
    parameters changed. In case they don't, it assumes an infinite
    loop without any progress and ends the fight. The interval between
    checks is chosen to minimize the possibility of false positives.
    It has the known deficiency of yielding false positives when
    weapon hit and recharge rates are unattractively low (recharge
    rate 2 + hit rate 1 yields one hit on average after 100'000
    ticks).

    \return true if combat shall continue, false if it shall end */
bool
VcrPlayerPHost::checkCombatActivity()
{
    /* re-check timer expired? If not, don't check. */
    if (det_timer > time)
        return true;

    /* still moving? If yes, there's progress. */
    if (DISTANCE() > config.StandoffDistance()) {
        det_timer = time + DET_MOVEMENT_TIMER;
        return true;
    }

    /* Movement has stopped. Has there been any progress since last check? */
    if (det_valid) {
        if (compareDetStatus(det_status[0], status[0]) && compareDetStatus(det_status[1], status[1])) {
            /* No progress */
            return false;
        }
    }

    /* Combat still runs. Compute re-check time. */
    fillInDetStatus(det_status[0], status[0]);
    fillInDetStatus(det_status[1], status[1]);
    det_valid = true;

    int32 interval = DET_INACTIVITY_TIMER;
    for (int i = 0; i < 2; ++i) {
        if (det_status[i].fighters != 0) {
            int32 L = status[i].f.BayLaunchInterval + 100;  // 100 = fuzz factor for safety
            if (status[i].f.FighterMovementSpeed > 0)
                L += 2 * config.StandoffDistance() / status[i].f.FighterMovementSpeed;
            if (L > interval)
                interval = L;
        }
    }
    det_timer = time + interval;
    return true;
}

/** Play one cycle. Slightly different than in PCC 1.x due to
    more-precisely specified semantics. */
bool
VcrPlayerPHost::playCycle()
{
    if (!canStillFight(status[0], status[1]) && !canStillFight(status[1], status[0]))
        done = true;
    else if (!checkCombatActivity())
        done = true;

    if (done)
        return false;

    ++time;
    if (status[0].r.launch_countdown > 0)
        --status[0].r.launch_countdown;
    if (status[1].r.launch_countdown > 0)
        --status[1].r.launch_countdown;

    fighterRecharge(status[0]);
    fighterRecharge(status[1]);
    beamRecharge(status[0]);
    beamRecharge(status[1]);
    torpsRecharge(status[0]);
    torpsRecharge(status[1]);
    fighterLaunch(status[0]);
    fighterLaunch(status[1]);

    if (fighterAttack(status[0], status[1]) || fighterAttack(status[1], status[0])
        || torpsFire(status[0], status[1]) || torpsFire(status[1], status[0])
        || beamFire(status[0], status[1]) || beamFire(status[1], status[0]))
    {
        done = true;
    } else {
        fighterIntercept();
        fighterMove(status[0]);
        fighterMove(status[1]);
        moveObjects();
    }
    return true;
}

/** Initialize player.
    \param phost3 true iff PHost 3 or later algorithm */
VcrPlayerPHost::VcrPlayerPHost(bool phost3)
    : seed(0),
      done(false),
      one_f(0),
      right_probab(0),
      capabilities(0),
      det_valid(false),
      det_timer(0),
      phost3(phost3)
{ }

/** Verify one side of VCR. */
bool
VcrPlayerPHost::checkVcrSide(GClassicVcr& vcr, int side)
{
    bool err = false;
    GVcrObject& o = vcr(side);

    /* owner and types must be in range */
    if (o.getOwner() <= 0 || o.getOwner() > 12)
        o.setOwner(12);

    if (o.getBeamType() < 0 || o.getBeamType() > NUM_BEAMS) {
        o.setBeamType(0);
        o.setBeamCount(0);
        err = true;
    }

    if (o.getTorpType() < 0 || o.getTorpType() > NUM_TORPS) {
        o.setTorpType(0);
        o.setTorpLauncherCount(0);
        err = true;
    }

    /* validate weapon counts */
    if (o.getBeamCount() > VCR_MAX_BEAMS) {
        o.setBeamCount(VCR_MAX_BEAMS);
        err = true;
    }

    if (o.getTorpLauncherCount() > VCR_MAX_TORPS) {
        o.setTorpLauncherCount(VCR_MAX_TORPS);
        err = true;
    }

    if (o.getBayCount() > VCR_MAX_BAYS) {
        o.setBayCount(VCR_MAX_BAYS);
        err = true;
    }

    /* ensure experience level is consistent with configuration */
    if (o.getExperienceLevel()) {
        if (!(capabilities & vcrcap_Experience) || o.getExperienceLevel() > config.NumExperienceLevels()) {
            o.setExperienceLevel(0);
            err = true;
        }
    }

    return err;
}

/** Check and correct VCR. Returns true iff an error was corrected,
    false if the VCR was ok. */
bool
VcrPlayerPHost::checkVcr(GClassicVcr& vcr)
{
    bool l = checkVcrSide(vcr, 0);
    bool r = checkVcrSide(vcr, 1);
    return l || r;
}

/** Initialize VCR Player. Creates shadow copies of configuration
    parameters and pre-charges weapons. */
void
VcrPlayerPHost::initVcr(const GClassicVcr& avcr)
{
    GClassicVcr tmpVcr = avcr;
    if (checkVcr(tmpVcr)) {
        status_word = VCRS_INVALID;
        done = true;
        return;
    }

    /* Initialize Playback */
    status_word = 0;
    time = 0;
    seed = avcr.getSeed()*65536;
    done = false;

    status[0].r.obj_x = -29000;
    status[1].r.obj_x = +29000;
    for (int side = 0; side < 2; ++side) {
        Status& st = status[side];
        st.f.side = side;
        st.r.obj = avcr(side);
#ifdef PVCR_INTEGER
        if (config.AllowAlternativeCombat())
            st.f.scale = st.r.obj.getMass() + 1;
        else
            st.f.scale = 1;
        st.r.shield_scaled  = st.r.obj.getShield() * int32(st.f.scale);
        st.r.crew_scaled2   = st.r.obj.getCrew() * int32(st.f.scale) * 100;
        st.r.damage_scaled2 = st.r.obj.getDamage() * int32(st.f.scale) * 100;
        st.f.max_scaled     = int32(10000) * st.f.scale;
        st.f.mass_plus1     = st.r.obj.getMass() + 1;
#else
        st.r.shield     = st.r.obj.getShield();
        st.r.crew       = st.r.obj.getCrew();
        st.r.damage     = st.r.obj.getDamage();
        st.f.mass_plus1 = st.r.obj.getMass() + 1;
#endif

        if (st.r.obj.getBeamCount() > 0) {
            st.f.beam_hit_odds = computeBeamHitOdds     (st.r.obj.getBeamType(), st.r.obj);
            st.f.beam_recharge = computeBeamRechargeRate(st.r.obj.getBeamType(), st.r.obj);
        } else {
            st.f.beam_hit_odds = 0;
            st.f.beam_recharge = 0;
        }
        if (st.r.obj.getTorpLauncherCount() > 0) {
            st.f.torp_hit_odds = computeTorpHitOdds(st.r.obj.getTorpType(), st.r.obj);
            st.f.torp_recharge = computeTubeRechargeRate(st.r.obj.getTorpType(), st.r.obj);
        } else {
            st.f.torp_hit_odds = 0;
            st.f.torp_recharge = 0;
        }
        if (st.r.obj.getBayCount() > 0) {
            st.f.bay_recharge = computeBayRechargeRate(st.r.obj.getBayCount(), st.r.obj);
        } else {
            st.f.bay_recharge = 0;
        }

        int charge = (st.r.obj.getShield() >= 100) ? 1000 : 0;
        std::fill_n(st.r.beam_status, int(VCR_MAX_BEAMS), charge);
        std::fill_n(st.r.torp_status, int(VCR_MAX_TORPS), charge);
        std::fill_n(st.r.bay_status,  int(VCR_MAX_BAYS),  0); // !!!
        std::fill_n(st.r.fighter_status, int(VCR_MAX_FTRS), 0);
        std::fill_n(st.r.fighter_x, int(VCR_MAX_FTRS), 0);  // not needed?
        std::fill_n(st.r.fighter_shots_left, int(VCR_MAX_FTRS), 0);  // not needed?
        st.r.launch_countdown = st.r.launched_fighters = 0;

#ifdef PVCR_INTEGER
        /* Original comparison was:
            if (st.r.damage_scaled2 + 50 * st.f.scale >=
                st.f.damage_limit * 100 * st.f.scale)          */
        st.f.damage_limit_scaled =
            raceId(st.r.obj.getOwner()) == 2
            ? (150 * 2 - 1) * 50 * st.f.scale
            : (100 * 2 - 1) * 50 * st.f.scale;
#else
        st.f.damage_limit = (raceId(st.r.obj.getOwner()) == 2 ? 150 : 100);
#endif

        const int owner = st.r.obj.getOwner();
        st.f.ShieldDamageScaling  = getExperienceMod(config.ShieldDamageScaling,  config.EModShieldDamageScaling,  st.r.obj, 0, 32767);
        st.f.ShieldKillScaling    = getExperienceMod(config.ShieldKillScaling,    config.EModShieldKillScaling,    st.r.obj, 0, 32767);
        st.f.HullDamageScaling    = getExperienceMod(config.HullDamageScaling,    config.EModHullDamageScaling,    st.r.obj, 0, 32767);
        st.f.MaxFightersLaunched  = getExperienceMod(config.MaxFightersLaunched,  config.EModMaxFightersLaunched,  st.r.obj, 0, VCR_MAX_FTRS);
        st.f.StrikesPerFighter    = getExperienceMod(config.StrikesPerFighter,    config.EModStrikesPerFighter,    st.r.obj, 1, 100);
        st.f.BayLaunchInterval    = config.BayLaunchInterval(owner);
        st.f.FighterMovementSpeed = getExperienceMod(config.FighterMovementSpeed, config.EModFighterMovementSpeed, st.r.obj, 1, 10000);
        st.f.FighterBeamExplosive = getExperienceMod(config.FighterBeamExplosive, config.EModFighterBeamExplosive, st.r.obj, 1, 1000);
        st.f.FighterBeamKill      = getExperienceMod(config.FighterBeamKill,      config.EModFighterBeamKill,      st.r.obj, 1, 1000);
        st.f.FighterFiringRange   = config.FighterFiringRange(owner);
        st.f.BeamHitFighterRange  = config.BeamHitFighterRange(owner);
        st.f.BeamHitFighterCharge = getExperienceMod(config.BeamHitFighterCharge, config.EModBeamHitFighterCharge, st.r.obj, 1, 1000);
        st.f.BeamFiringRange      = config.BeamFiringRange(owner);
        st.f.BeamHitShipCharge    = config.BeamHitShipCharge(owner);
        st.f.TorpFiringRange      = config.TorpFiringRange(owner);
        st.f.ShipMovementSpeed    = config.ShipMovementSpeed(owner);

        st.f.CrewKillScaling =
            radiv((100-st.r.obj.getCrewDefenseRate()) * getExperienceMod(config.CrewKillScaling, config.EModCrewKillScaling, st.r.obj, 0, 32767),
                  100);
    }

    /* pre-compute fighter intercept probabilities */
    if (phost3) {
        /* PHost 3 or 4 */
        int left_odds  = config.FighterKillOdds(status[0].r.obj.getOwner());
        int right_odds = config.FighterKillOdds(status[1].r.obj.getOwner());
        int left_f   = (100 - left_odds) * right_odds;
        int right_f  = (100 - right_odds) * left_odds;
        one_f        = (left_f + right_f) / 100;
        if (one_f == 0)
            right_probab = 50;
        else
            right_probab = left_f / one_f;
    } else {
        /* In PHost 2, combat options were not arrayized. Hence, for
           a valid pconfig, all FighterKillOdds values are the same
           and we can pick any one */
        one_f        = config.FighterKillOdds(1);
        right_probab = 50;
    }

    initActivityDetector();
    initStat();
    vis->init();
}

/** Finish up VCR. */
void
VcrPlayerPHost::doneVcr(GClassicVcr& avcr)
{
    ASSERT(status_word != VCRS_INVALID);
    ASSERT(done);

    for (int side = 0; side < 2; ++side) {
        Status& st = status[side];
#ifdef PVCR_INTEGER
        if (config.AllowAlternativeCombat()) {
            st.r.obj.setDamage(radiv(st.r.damage_scaled2, st.f.scale * 100));
            st.r.obj.setCrew  (radiv(st.r.crew_scaled2,   st.f.scale * 100));
            st.r.obj.setShield(radiv(st.r.shield_scaled,  st.f.scale));
        } else {
            st.r.obj.setDamage(radiv(st.r.damage_scaled2, 100));
            st.r.obj.setCrew  (radiv(st.r.crew_scaled2, 100));
            st.r.obj.setShield(st.r.shield_scaled);
        }
#else
        st.r.obj.setDamage(int(st.r.damage + 0.5));
        st.r.obj.setCrew  (int(st.r.crew + 0.5));
        st.r.obj.setShield(int(st.r.shield + 0.5));
#endif
        if (st.r.obj.getDamage() > 100)   // FIXME: take out?
            st.r.obj.setDamage(100);

#ifdef PVCR_INTEGER
        if (canStillFight(st, status[!side]) && st.r.obj.getDamage() < 100)
#else
        if (canStillFight(st, status[!side]) && st.r.damage < 99.5)
#endif
        {
            for (int i = 0; i < st.f.MaxFightersLaunched; ++i) {
                if (st.r.fighter_status[i]) {
                    st.r.obj.addFighters(+1);
                    --st.r.launched_fighters;
                    vis->landFighter(side, i);
                    st.r.fighter_status[i] = 0;
                }
            }
            ASSERT(st.r.launched_fighters == 0);
        }
        st.r.launched_fighters = 0;
    }

    status_word = 0;

    /* Lizards fight up to 150, but explode afterwards. */
    if (status[0].r.obj.getDamage() >= 100)
        status_word |= VCRS_LEFT_DESTROYED;
    else if (status[0].r.obj.getCrew() <= 0)
        status_word |= VCRS_LEFT_CAPTURED;

    if (status[1].r.obj.getDamage() >= 100)
        status_word |= VCRS_RIGHT_DESTROYED;
    else if (!status[1].r.obj.isPlanet() && status[1].r.obj.getCrew() <= 0)
        status_word |= VCRS_RIGHT_CAPTURED;

    if (status_word & VCRS_LEFT_DESTROYED)
        vis->killObject(0);
    if (status_word & VCRS_RIGHT_DESTROYED)
        vis->killObject(1);

    if (status_word == 0)
        status_word = VCRS_STALEMATE;

    avcr(0) = status[0].r.obj;
    avcr(1) = status[1].r.obj;
}

bool
VcrPlayerPHost::setCapabilities(uint16 cap)
{
    if (cap & ~SUPPORTED_CAP)
        return false;
    capabilities = cap;
    return true;
}

/** Fast forward. The PHost 'Fast Forward' scheme is a little more
    complicated than the one for THost because most values are
    configurable, and because the PRNG is more complicated. */
void
VcrPlayerPHost::playFastForward()
{
    if (time == 0
        && status[0].r.obj.getBayCount() == 0 && status[1].r.obj.getBayCount() == 0
        && status[0].r.obj.getShield() == 100 && status[1].r.obj.getShield() == 100)
    {
        /* no carrier in play, both fully charged. Hence, no recharge
           work to be done. First interesting thing will happen when
           we are in beam range. PHost still polls the PRNG once for
           every beam and tick, as long as beams are charged above
           BeamHitFighterCharge. */
        /* FIXME: same applies if one ship is a freighter */
        long dist_now = DISTANCE();
        int target_dist = config.StandoffDistance();
        if (target_dist < status[0].f.BeamFiringRange)
            target_dist = status[0].f.BeamFiringRange;
        if (target_dist < status[1].f.BeamFiringRange)
            target_dist = status[1].f.BeamFiringRange;
        if (status[0].r.obj.getTorpLauncherCount() > 0)
            if (target_dist < status[0].f.TorpFiringRange)
                target_dist = status[0].f.TorpFiringRange;
        if (status[1].r.obj.getTorpLauncherCount() > 0)
            if (target_dist < status[1].f.TorpFiringRange)
                target_dist = status[1].f.TorpFiringRange;

        int speed = status[0].f.ShipMovementSpeed;
        if (!status[1].r.obj.isPlanet())
            speed += status[1].f.ShipMovementSpeed;

        if (target_dist < dist_now && speed > 0) {
            /* We'll move towards each other at a nonzero speed */
            int adv_time = (dist_now - target_dist) / speed;
            if (adv_time > 2) {
                /* less 2 ticks, for safety */
                adv_time -= 2;

                /* advance time */
                time += adv_time;
                status[0].r.obj_x += status[0].f.ShipMovementSpeed * adv_time;
                if (!status[1].r.obj.isPlanet())
                    status[1].r.obj_x -= status[1].f.ShipMovementSpeed * adv_time;

                /* advance PRNG. Although there is a formula how to
                   advance a linear congruence *fast*, we cannot use
                   it because randomRange() advances the seed twice in
                   36 out of 65536 cases. */
                /* For the simulator, a more adventurous FF scheme
                   would be to *not* advance the PRNG, and calculate
                   the initial seed backwards when the user wants to
                   see this instance. */
                /* The maximum num_pulls is around 10000 for regular
                   configurations, so there's no point in optimizing
                   for >65500. */
                int num_pulls = 0;
                if (status[0].f.BeamHitFighterCharge <= 1000)
                    num_pulls += status[0].r.obj.getBeamCount();
                if (status[1].f.BeamHitFighterCharge <= 1000)
                    num_pulls += status[1].r.obj.getBeamCount();
                num_pulls *= adv_time;
                while (num_pulls-- > 0)
                    randomRange100();
            }
        }
    }
}
//@}

/** \name Accessor Interface */
//@{

int
VcrPlayerPHost::getBeamStatus(int side, int id)
{
    return status[side].r.beam_status[id] / 10;
}

int
VcrPlayerPHost::getTorpStatus(int side, int id)
{
    return status[side].r.torp_status[id] / 10;
}

int
VcrPlayerPHost::getTorpCount(int side)
{
    return status[side].r.obj.getTorpCount();
}

int
VcrPlayerPHost::getFighterCount(int side)
{
    return status[side].r.obj.getFighterCount();
}

int
VcrPlayerPHost::getShields(int side)
{
    /* Round, because driver code also rounds. */
#ifdef PVCR_INTEGER
    return radiv(status[side].r.shield_scaled, status[side].f.scale);
#else
    return int(status[side].r.shield + 0.5);
#endif
}

int
VcrPlayerPHost::getDamage(int side)
{
#ifdef PVCR_INTEGER
    return radiv(status[side].r.damage_scaled2, status[side].f.scale * 100);
#else
    return int(status[side].r.damage + 0.5);
#endif
}

int
VcrPlayerPHost::getCrew(int side)
{
#ifdef PVCR_INTEGER
    return radiv(status[side].r.crew_scaled2, status[side].f.scale * 100);
#else
    return int(status[side].r.crew + 0.5);
#endif
}

int
VcrPlayerPHost::getFighterX(int side, int id)
{
    return (status[side].r.fighter_x[id]) / 128 + 320;
}

int
VcrPlayerPHost::getFighterStatus(int side, int id)
{
    return status[side].r.fighter_status[id];
}

int
VcrPlayerPHost::getObjectX(int side)
{
    return (status[side].r.obj_x) / 128 + 320;
}

long
VcrPlayerPHost::getDistance()
{
    return status[1].r.obj_x - status[0].r.obj_x;
}

/** Save status. Easy mindless way, just save everything. Actually, we
    could re-compute status[].f and one_f/right_probab from the VCR
    record, but I'm too lazy to do that now. */
VcrStatus*
VcrPlayerPHost::getStatusToken()
{
    StatusToken* s = new StatusToken(time);

    s->status[0]    = status[0];
    s->status[1]    = status[1];
    s->seed         = seed;
    s->status_word  = status_word;
    s->done         = done;
    s->one_f        = one_f;
    s->right_probab = right_probab;

    return s;
}

void
VcrPlayerPHost::setStatus(VcrStatus& s)
{
    StatusToken* t = checked_cast<StatusToken*>(&s);
    status[0]    = t->status[0];
    status[1]    = t->status[1];
    seed         = t->seed;
    status_word  = t->status_word;
    done         = t->done;
    one_f        = t->one_f;
    right_probab = t->right_probab;
    time         = t->time;
}

//@}

VcrPlayerPHost::StatusToken::StatusToken(time_type t)
    : VcrStatus(t)
{ }
