/**
  *  \file game/ship.cc
  *  \brief Ships
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "game/ship.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/formula.h"       // getHeadingDeg
#include "game/host.h"
#include "game/hullfunc.h"
#include "game/planet.h"
#include "game/racename.h"
#include "game/ship-int.h"
#include "game/spec.h"
#include "game/univ.h"
#include "int/data.h"
#include "int/intglobal.h"
#include "int/value.h"
#include "util/nls.h"
#include "util/recode.h"
#include "game/parser.h"

static const TShip blank_ship = {
    0, 0,                                       // ID, Owner
    { UNKNOWNC, UNKNOWNC, UNKNOWNC },           // FCode
    UNKNOWNP,                                   // Warp
    UNKNOWNS, UNKNOWNS,                         // Waypoint-DX, DY
    UNKNOWNP, UNKNOWNP,                         // X, Y
    UNKNOWNP,                                   // Engine
    UNKNOWNP,                                   // Hull
    UNKNOWNP, UNKNOWNP,                         // Beam Type, Count
    UNKNOWNP,                                   // Bay Count
    UNKNOWNP,                                   // Torp Type
    UNKNOWNP,                                   // Ammo
    UNKNOWNP,                                   // Torp Launcher Count
    UNKNOWNP,                                   // Mission
    UNKNOWNP,                                   // PE
    UNKNOWNP,                                   // Tow
    UNKNOWNP, UNKNOWNP,                         // Damage, Crew
    UNKNOWNP,                                   // Colonists
    { 0 },                                      // Name
    { UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP }, // Ore
    UNKNOWNP,                                   // Supplies
    { { 0, 0, 0, 0 }, 0, 0, 0 },                // Unload
    { { 0, 0, 0, 0 }, 0, 0, 0 },                // Transfer
    UNKNOWNP,                                   // Intercept
    UNKNOWNP,                                   // Cash
};

static const TDbShipTrackEntry blank_ship_track = {
    UNKNOWNP, UNKNOWNP,                         // x, y
    UNKNOWNP,                                   // Speed
    UNKNOWNP,                                   // Heading
    UNKNOWNP                                    // Mass
};

static void
updateField16(int16_t& field_time, int16_t time,
              int16_t& field_value, int32_t value)
{
    if (field_time <= time || !mp16_t(field_value).isKnown()) {
        field_value = value;
        if (field_time < time)
            field_time = time;
    }
}

/** Check for dummy name. PHost can filter out ship names; we detect
    such names to avoid overwriting a known name by a dummy.
    \param name [in] Name, 20 bytes
    \param ship_id [in] Ship Id
    \return true iff it is a dummy name
    \todo maybe recognize other client's dummy names? */
static bool
isDummyName(const char* name, int ship_id)
{
    return getBasicStringN(name, 20) == format("Ship %d", ship_id);
}

/** Fix dull transfer: clear target of a null transfer to avoid confusing things */
static void fixDullTransfer(TShipTransfer& t)
{
    if (t.ore[0] == 0 && t.ore[1] == 0 && t.ore[2] == 0 && t.ore[3] == 0
        && t.colonists == 0 && t.supplies == 0)
    {
        t.id = 0;
    }
}

/******************************** MapInfo ********************************/

GShip::MapInfo::MapInfo(int id)
    : id(id),
      scanned_mass(),
      special_functions(),
      orbit(0),
      rc_flag(0),
      kind(NoShip),
      fleet_nr(0),
      fleet_name(),
      predicted_pos(0, 0)
{ }

GShip::MapInfo::~MapInfo()
{ }

/******************************** ShipInfo *******************************/

GShip::ShipInfo::ShipInfo(int id)
{
    init(id);
}

void
GShip::ShipInfo::init(int id)
{
    data         = blank_ship;
    data.ship_id = id;
    time[0]      = 0;
    time[1]      = -1;
    track_turn   = 0;
    for (int i = 0; i < NUM_SHIP_TRACK_ENTRIES; ++i) {
        track[i] = blank_ship_track;
    }
    heading = -1;
}

void
GShip::ShipInfo::clear()
{
    init(data.ship_id);
}

/********************************* GShip *********************************/

/** Constructor. Creates a blank ship object for ship of a given Id.
    \param id ship id */
GShip::GShip(int id)
    : GMapObject(),
      unit_scores(),
      cargo_arbiter(),
      ship_info(0),
      map_info(id),
      associated_messages(),
      ship_source(),
      target_source(),
      xy_source()
{
    transporter_in_use[0] = transporter_in_use[1] = false;
}

GShip::GShip(const GShip& other)
    : GMapObject(),
      unit_scores(other.unit_scores),
      cargo_arbiter(),
      ship_info(0),
      map_info(other.map_info),
      ship_source(other.ship_source),
      target_source(other.target_source),
      xy_source(other.xy_source)
{
    if (other.ship_info != 0) {
        ship_info = new ShipInfo(*other.ship_info);
    }
    transporter_in_use[0] = transporter_in_use[1] = false;
}

GShip::~GShip()
{
    delete ship_info;
}


/*
 *  Load and Save
 */

/** Add SHIP.DAT entry. This assumes that if we see a ship through
    several .dat files, we get the same one each time.
    \param data [in] Parsed .dat file entry
    \param source [in] Source flags, i.e. bitfield of which file this record is from */
void
GShip::addShipData(const TShip& data, GPlayerSet source)
{
    // Set hull through regular setter to get reset behaviour
    setHullId(mp16_t(data.hull));

    ShipInfo& info = createShipInfo();
    info.data = data;

    // Adjust SRace
    if (host.isSRace() && info.data.mission == 1) {
        info.data.mission = 9;
    }

    // Fix erros
    fixDullTransfer(info.data.unload);
    fixDullTransfer(info.data.transfer);
    if (data.warp < 0) {
        info.data.warp = 0;         // Lizards with >100% damage
    }

    ship_source |= source;
}

/** Add TARGET.DAT entry. This assumes that if we see a ship through
    several .dat files, we get the same one each time. This does, however,
    handle the case that one record is seen through PHost's ship name filter
    and the other one isn't, and uses the better ship name.
    \param target [in] Parsed TARGET.DAT file entry
    \param source [in] Source flags, i.e. bitfield of which file this record is from */
void
GShip::addTargetData(const TShipTarget& target, GPlayerSet source)
{
    // Record that we know it
    target_source |= source;

    // Update ship unless we already saw it this turn
    if (ship_source.empty()) {
        // Set hull through regular setter to get reset behaviour
        setHullId(mp16_t(target.hull));

        // Set remainder
        ShipInfo& info = createShipInfo();
        info.data.owner       = target.owner;
        info.data.warp        = target.warp;
        info.data.x           = target.x;
        info.data.y           = target.y;
        info.data.waypoint_dx = mn16_t().getRawValue();
        info.data.waypoint_dy = mn16_t().getRawValue();
        info.heading          = target.heading;

        // If the new target is better than the one we have, use that
        if (!isDummyName(target.name, map_info.id)) {
            std::memcpy(info.data.name, target.name, sizeof(target.name));
        }
    }
}

/** Add SHIPXY.DAT entry. */
void
GShip::addShipXYData(TShipXY xy, GPlayerSet source)
{
    // Record that we know it
    xy_source |= source;

    // Update ship.
    if (ship_source.empty()) {
        ShipInfo& info        = createShipInfo();
        info.data.x           = xy.x;
        info.data.y           = xy.y;
        info.data.waypoint_dx = mn16_t().getRawValue();
        info.data.waypoint_dy = mn16_t().getRawValue();
        info.data.owner       = xy.owner;
        map_info.scanned_mass = mp16_t(xy.mass);
    }
}

/** Add remote control flag.
    \param flag=0 for non-RC'd ships
    \param flag<0 if RC is forbidden
    \param flag>0 if to report the ship's real owner */
void
GShip::addRCEntry(int flag)
{
    map_info.rc_flag = flag;
}

/** Add ship history data.
    For the time being, we only accept history data when no current data is present.
    Since the history database file is loaded before the game data, this works out nicely.
    When the order is changed, we'd have to merge here and avoid overwriting target/xy
    data with history. */
void
GShip::addShipHistoryData(const TDbShip& data)
{
    if (ship_source.empty() && target_source.empty() && xy_source.empty()) {
        // Remember data
        ShipInfo& info = createShipInfo();
        info.data    = data.ship;
        info.time[0] = data.time[0];
        info.time[1] = data.time[1];

        // Must clear location info
        info.data.x           = mp16_t().getRawValue();
        info.data.y           = mp16_t().getRawValue();
        info.data.waypoint_dx = mn16_t().getRawValue();
        info.data.waypoint_dy = mn16_t().getRawValue();

        // Sanitize hull
        if (data.ship.hull == 0) {
            info.data.hull = UNKNOWNP;
        }
    }
}

/** Add ship track data.
    \param entry [in] Ship Track entry from database
    \param turn  [in] Turn number (absolute value) */
void
GShip::addShipTrackEntry(const TDbShipTrackEntry& entry, int turn)
{
    if (entry.x != UNKNOWNP && entry.y != UNKNOWNP) {
        // Adjust ship position
        adjustShipTrack(turn);

        // Fill in new entry if possible
        ShipInfo& info = createShipInfo();
        int slot = info.track_turn - turn;
        if (slot < NUM_SHIP_TRACK_ENTRIES) {
            info.track[slot] = entry;
        }
    }
}

/** Add ship function.
    \param function Identifier of function. Can be a basic hull function identifier
    (hf_XXX), or a host function Id (see GHullFunctionData::getIdFromHostId). */
void
GShip::addShipFunction(int function)
{
    map_info.special_functions.push_back(function);
}

void
GShip::addMessageInformation(const GMessageInformation& info, int msgNr)
{
    TDbShipTrackEntry track;
    track.x = UNKNOWNP;
    track.y = UNKNOWNP;
    track.speed = 0;
    track.heading = -1;
    track.mass = -1;

    for (GMessageInformation::iterator i = info.begin(); i != info.end(); ++i) {
        if (GMessageIntValue* iv = dynamic_cast<GMessageIntValue*>(*i)) {
            switch (iv->getIndex()) {
             case mi_ShipRemoteFlag:
                addRCEntry(iv->getValue());
                break;

             case mi_X:
                if (ship_source.empty() && target_source.empty() && xy_source.empty()) {
                    track.x = iv->getValue();
                }
                break;

             case mi_Y:
                if (ship_source.empty() && target_source.empty() && xy_source.empty()) {
                    track.y = iv->getValue();
                }
                break;

             case mi_Owner:
                if (ship_source.empty() && target_source.empty() && xy_source.empty()) {
                    if (iv->getValue() != 0) {
                        ShipInfo& si = createShipInfo();
                        if (info.getTurnNumber() >= si.time[sts_Rest] || si.data.owner <= 0) {
                            si.data.owner = iv->getValue();
                            if (si.time[sts_Rest] < info.getTurnNumber()) {
                                si.time[sts_Rest] = info.getTurnNumber();
                            }
                        }
                    }
                }
                break;

             case mi_ShipHull:
                if (ship_source.empty() && target_source.empty()) {
                    if (iv->getValue() != 0) {
                        ShipInfo& si = createShipInfo();
                        updateField16(si.time[sts_Rest], info.getTurnNumber(), si.data.hull, iv->getValue());
                    }
                }
                break;

             case mi_ShipEngineType:
                if (ship_source.empty()) {
                    if (iv->getValue() != 0) {
                        ShipInfo& si = createShipInfo();
                        updateField16(si.time[sts_Rest], info.getTurnNumber(), si.data.engine, iv->getValue());
                    }
                }
                break;

             case mi_ShipBeamType:
                if (ship_source.empty()) {
                    ShipInfo& si = createShipInfo();
                    updateField16(si.time[sts_ArmsDamage], info.getTurnNumber(), si.data.beam_type, iv->getValue());
                }
                break;

             case mi_ShipNumBeams:
                if (ship_source.empty()) {
                    ShipInfo& si = createShipInfo();
                    updateField16(si.time[sts_ArmsDamage], info.getTurnNumber(), si.data.beam_count, iv->getValue());
                }
                break;

             case mi_ShipNumBays:
                if (ship_source.empty()) {
                    ShipInfo& si = createShipInfo();
                    updateField16(si.time[sts_ArmsDamage], info.getTurnNumber(), si.data.bay_count, iv->getValue());
                }
                break;

             case mi_ShipLauncherType:
                if (ship_source.empty()) {
                    ShipInfo& si = createShipInfo();
                    updateField16(si.time[sts_ArmsDamage], info.getTurnNumber(), si.data.torp_type, iv->getValue());
                }
                break;

             case mi_ShipAmmo:
                if (ship_source.empty()) {
                    ShipInfo& si = createShipInfo();
                    updateField16(si.time[sts_ArmsDamage], info.getTurnNumber(), si.data.ammo, iv->getValue());
                }
                break;

             case mi_ShipNumLaunchers:
                if (ship_source.empty()) {
                    ShipInfo& si = createShipInfo();
                    updateField16(si.time[sts_ArmsDamage], info.getTurnNumber(), si.data.torp_launcher_count, iv->getValue());
                }
                break;

             case mi_Damage:
                if (ship_source.empty()) {
                    ShipInfo& si = createShipInfo();
                    updateField16(si.time[sts_ArmsDamage], info.getTurnNumber(), si.data.damage, iv->getValue());
                }
                break;

             default:
                break;
            }
        }
        if (GMessageStringValue* sv = dynamic_cast<GMessageStringValue*>(*i)) {
            switch (sv->getIndex()) {
             case ms_Name:
                if (ship_source.empty() && target_source.empty()) {
                    if (!isDummyName(sv->getValue().c_str(), getId())) {
                        setName(sv->getValue());
                    }
                }
                break;

             default:
                break;
            }
        }
    }

    /* If we produced a ship track entry, add it (addShipTrackEntry will filter) */
    addShipTrackEntry(track, info.getTurnNumber());

    /* Associate message */
    if (msgNr >= 0) {
        if (associated_messages.empty() || associated_messages.back() != msgNr) {
            associated_messages.push_back(msgNr);
        }
    }

    /* mark ship dirty, just in case */
    markDirty();
}

/** Get ship data.
    \param data [out] Ship data to store in SHIP.DAT file */
void
GShip::getShipData(TShip& data) const
{
    ASSERT(ship_info);
    data = ship_info->data;
    if (host.isSRace()) {
        if (data.mission == 1) {
            data.mission = 0;
        } else if (data.mission == 9) {
            data.mission = 1;
        }
    }
}

/** Get header for ship track data in history file.
    \param header [out] Header
    \return Number of track entries (0=none, no header returned in this case) */
int
GShip::getShipTrackHeader(TDbShipTrack& header) const
{
    if (ship_info != 0) {
        header.id = map_info.id;
        header.ref_turn = ship_info->track_turn;
        return NUM_SHIP_TRACK_ENTRIES;
    } else {
        return 0;
    }
}

/** Get ship track entry by slot.
    \param entry [out] Entry
    \param slot [in] Slot to access, [0, getShipTrackHeader(...)) */
void
GShip::getShipTrackEntry(TDbShipTrackEntry& entry, int slot) const
{
    ASSERT(slot >= 0 && slot < NUM_SHIP_TRACK_ENTRIES);
    ASSERT(ship_info != 0);
    entry = ship_info->track[slot];
}

/** Get ship history data.
    \param the_data [out] Data to store in history file */
void
GShip::getShipHistoryData(TDbShip& data) const
{
    if (ship_info != 0) {
        data.ship    = ship_info->data;
        data.time[0] = ship_info->time[0];
        data.time[1] = ship_info->time[1];
    } else {
        data.ship    = blank_ship;
        data.time[0] = 0;
        data.time[1] = 0;
    }
}

/** Do internal checks for this ship.
    Internal checks do not require a partner to interact with.
    This will fix the problems, and display appropriate messages.
    It will also fill in the ship kind. */
void
GShip::internalCheck()
{
    /* figure out what kind we are */
    if (ship_source.nonempty()) {
        /* current ship */
        map_info.kind = CurrentShip;
    } else if (target_source.nonempty()) {
        /* current target */
        map_info.kind = CurrentTarget;
    } else if (xy_source.nonempty()) {
        /* nonvisual contact */
        map_info.kind = CurrentUnknown;
    } else if (ship_info != 0) {
        /* history ship */
        map_info.kind = HistoryShip;
    } else {
        /* unknown */
        map_info.kind = NoShip;
    }

    /* owner needs to be known */
    if (ship_info != 0 && ship_info->data.owner <= 0) {
        /* Ships without owner are generated from explosion records (TUtil1Bang).
           If anything else reports an unowned ship, that's an error in the data files. */
        // Reset ship type
        map_info.kind = NoShip;

        // Avoid that anyone trusts this data
        ship_source = target_source = xy_source = GPlayerSet();

        // Remove known data. If this data leaks into chartX.cc files, PCC2 < 2.0.7 will crash.
        delete ship_info;
        ship_info = 0;
    }
}

/** Combined checks, phase 1.
    This will do all post-processing which needs a partner to interact with.
    It requires the playability to be filled in. */
void
GShip::combinedCheck1(GUniverse& univ)
{
    // If we see the ship, it must exist even if history data says otherwise.
    // The next condition might otherwise delete it.
    if (!ship_source.empty() || !target_source.empty() || !xy_source.empty()) {
        if (ship_info != 0 && ship_info->data.damage > 150) {
            ship_info->data.damage = 0;
        }
    }

    // If ship claims to exists, but we don't have current data, it's destroyed. Remove it.
    // (But don't upgrade a non-existant ship, e.g. explosion-only, to HistoryShip.)
    if ((univ.isData(getOwner()) && ship_source.empty())
        || (getOwner() > 0 && ship_info != 0 && ship_info->data.damage > 150))
    {
        // Clear current data
        map_info.scanned_mass = mp16_t();
        ship_info->data.x = -1;
        ship_info->data.y = -1;
        ship_info->data.warp = -1;

        // Clear current turn's history data, we know it does not exist this turn
        if (ship_info->track_turn == univ.getTurnNumber()) {
            ship_info->track[0].x = UNKNOWNP;
            ship_info->track[0].y = UNKNOWNP;
            ship_info->track[0].speed = -1;
            ship_info->track[0].heading = -1;
            ship_info->track[0].mass = -1;
        }

        map_info.kind = HistoryShip;
    }

    // If this is a history ship, but we have current-turn data, it is a guessed ship instead
    if (map_info.kind == HistoryShip
        && ship_info != 0
        && ship_info->data.owner > 0
        && ship_info->track_turn == univ.getTurnNumber()
        && ship_info->track[0].x != UNKNOWNP
        && ship_info->track[0].y != UNKNOWNP)
    {
        map_info.kind = GuessedShip;
        ship_info->data.x = ship_info->track[0].x;
        ship_info->data.y = ship_info->track[0].y;
    }

    // Compute orbit
    map_info.orbit = univ.getPlanetAt(getPos());
    if (map_info.orbit) {
        univ.getPlanet(map_info.orbit).setShipInOrbit(*this);
    }

    // Update ages
    if (hasFullShipData()) {
        ship_info->time[0] = ship_info->time[1] = univ.getTurnNumber();
    }

    // Update ship track.
    // The simplest way is to generate a whole new record and have the regular code assimilate it.
    if (isVisible()) {
        TDbShipTrackEntry e;
        e.x       = ship_info->data.x;
        e.y       = ship_info->data.y;
        e.speed   = getWarp().getRawValue();
        e.heading = getHeading().getRawValue();
        e.mass    = getMass().getRawValue();
        addShipTrackEntry(e, univ.getTurnNumber());
    }
}


/*
 *  GMapObject interface
 */
string_t
GShip::getName(Name which) const
{
    // Figure out plain name
    string_t plain;
    if (ship_info != 0 && isKnownBasicString(ship_info->data.name)) {
        plain = convertGameToUtf8(getBasicString(ship_info->data.name));
    }

    // Is that just what we wanted?
    if (which == PlainName) {
        return plain;
    }

    // Add detail
    string_t result;
    if (plain.empty()) {
        result = format(_("Ship #%d"), map_info.id);
    } else {
        result = format(_("Ship #%d: %s"), map_info.id, convertGameToUtf8(getBasicString(ship_info->data.name)));
    }

    if (which == DetailedName) {
        IntValue* iv = getShipProperties(map_info.id)[sp_Comment];
        if (getBoolValue(iv) > 0) {
            result += ": ";
            result += iv->toString(false);
        } else if (getOwner() > 0) {
            mp16_t hull = getHullId();
            if (hull.isKnown()) {
                result += format(" (%s %s)", player_racenames.getAdjName(getOwner()), getHull().getShortName());
            } else {
                result += format(_(" (%s starship)"), player_racenames.getAdjName(getOwner()));
            }
        } else {
            // no decoration
        }
    }
    return result;
}

int
GShip::getId() const
{
    return map_info.id;
}

int
GShip::getOwner() const
{
    return ship_info != 0 ? ship_info->data.owner : 0;
}

GPoint
GShip::getPos() const
{
    const TShip& sh = getDisplayedShip();
    return GPoint(sh.x, sh.y);
}


/*
 *  Status inquiry
 */

/** Check whether this ship is visible.
    If it is visible, it is displayed on the map. */
bool
GShip::isVisible() const
{
    return map_info.kind == CurrentShip
        || map_info.kind == CurrentTarget
        || map_info.kind == CurrentUnknown
        || map_info.kind == GuessedShip;
}

/** Check whether this ship is reliably visible (to a player).
    A ship can be unreliably visible if it guessed.
    It can also be reliably visible to one player but not another one if they are not allied.

    \param forPlayer Player to ask question for: is this ship known to that player,
    and will host accept orders relating to it?
    If zero, check whether ship is seen reliably by anyone. */
bool
GShip::isReliablyVisible(const int forPlayer) const
{
    GPlayerSet set = target_source | ship_source | xy_source;
    if (forPlayer == 0) {
        return set.nonempty();
    } else {
        return set.contains(forPlayer);
    }
}

/** Get ship source flags.
    This is the set of players whose SHIP file contains a copy of this ship (usually a unit set). */
GPlayerSet
GShip::getShipSource() const
{
    return ship_source;
}

/** Get kind of this ship.
    The kind determines how complete and reliable this ship's data is. */
GShip::Kind
GShip::getShipKind() const
{
    return map_info.kind;
}

/** Check whether we have any data about this ship. */
bool
GShip::hasAnyShipData() const
{
    return ship_info != 0;
}

/** Check whether we have full, playable data. */
bool
GShip::hasFullShipData() const
{
    return ship_source.nonempty();
}


/*
 *  History accessors
 */

/** Get history timestamp. 0 means not set. */
int
GShip::getHistoryTimestamp(TDbShipTimestamp which_one) const
{
    if (ship_info != 0) {
        return ship_info->time[which_one];
    } else {
        return 0;
    }
}

/** Get ship history entry for a turn.
    \param turn Turn number */
const TDbShipTrackEntry&
GShip::getHistoryEntry(int turn) const
{
    if (ship_info != 0 && turn <= ship_info->track_turn && turn > ship_info->track_turn - NUM_SHIP_TRACK_ENTRIES) {
        return ship_info->track[ship_info->track_turn - turn];
    } else {
        return blank_ship_track;
    }
}

/** Get newest history turn. This is the newest turn for which we
    (may) have information. Iteration can use
    "for (t = getHistoryNewestTurn(); t > getHistoryOldestTurn(); --t)". */
int
GShip::getHistoryNewestTurn() const
{
    if (ship_info != 0) {
        return ship_info->track_turn;
    } else {
        return 0;
    }
}

/** Get oldest history turn. This is the turn before the oldest one
    for which we may have information.

    Note that this may return zero or a negative number if the history
    buffer reaches back to before the game start.

    \see getHistoryNewestTurn() */
int
GShip::getHistoryOldestTurn() const
{
    if (ship_info != 0) {
        return ship_info->track_turn - NUM_SHIP_TRACK_ENTRIES;
    } else {
        return 0;
    }
}


/*
 *  Type Accessors
 */

/** Get ship mass. */
mp16_t
GShip::getMass() const
{
    if (map_info.kind == CurrentShip) {
        return mp16_t(getShipMass(ship_info->data));
    } else {
        return map_info.scanned_mass;
    }
}

/** Get hull Id. */
mp16_t
GShip::getHullId() const
{
    return mp16_t(getDisplayedShip().hull);
}

/** Set hull Id. */
void
GShip::setHullId(mp16_t h)
{
    // Sanitize. Hull=0 is in some databases.
    if (h.getRawValue() == 0) {
        h = mp16_t();
    }

    if (ship_info != 0 || h.isKnown()) {
        ShipInfo& info = createShipInfo();
        if (mp16_t(info.data.hull).isKnown() && h.isKnown() && h.getRawValue() != info.data.hull) {
            // It's a hull change. Everything you know is wrong.
            info.clear();
        }
        info.data.hull = h.getRawValue();
        markDirty();
    }
}

/** Get hull object. */
GHull&
GShip::getHull() const
{
    mp16_t n = getHullId();
    ASSERT(n.isKnown());
    ASSERT(n > 0);
    return ::getHull(n);
}


/*
 *  Location accessors
 */

/** Get Id of planet we're orbiting.
    \return planet id, 0 if none. */
int
GShip::getOrbitPlanetId() const
{
    return map_info.orbit;
}

/** Check whether ship wants to be cloned.
    This does not whether cloning actually is permitted (registeredness,
    hull restrictions, player restrictions). */
bool
GShip::isCloningAt(const GPlanet& planet) const
{
    return ship_info != 0
        && planet.getPos() == getPos()
        && (ship_info->data.fcode[0] == 'c' && ship_info->data.fcode[1] == 'l' && ship_info->data.fcode[2] == 'n');
}


/*
 *  Owner accessors
 */

/** Get real owner of ship. */
int
GShip::getRealOwner() const
{
    if (map_info.rc_flag > 0) {
        return map_info.rc_flag;
    } else {
        return getOwner();
    }
}

/** Get ship's remote control flag. */
int
GShip::getRemoteControlFlag() const
{
    return map_info.rc_flag;
}


/*
 *  Course accessors
 */

/** Get waypoint.
    If the waypoint is not known, the GPoint's components will be unknown if interpreted as mp16_t. */
GPoint
GShip::getWaypoint() const
{
    const TShip& data = getDisplayedShip();
    return GPoint(data.x != UNKNOWNP && data.waypoint_dx != UNKNOWNS ? data.x + data.waypoint_dx : UNKNOWNP,
                  data.y != UNKNOWNP && data.waypoint_dy != UNKNOWNS ? data.y + data.waypoint_dy : UNKNOWNP);

}

/** Set waypoint.
    We can set a waypoint only if the ship is visible, i.e. has coordinates.
    If the ship is not visible, the call is ignored. */
void
GShip::setWaypoint(GPoint pt)
{
    if (isVisible()) {
        const TShip& data = getDisplayedShip();
        int16_t newDX = pt.x - data.x;
        int16_t newDY = pt.y - data.y;
        if (newDX != data.waypoint_dx || newDY != data.waypoint_dy) {
            ship_info->data.waypoint_dx = newDX;
            ship_info->data.waypoint_dy = newDY;
            markDirty();
        }
    }
}

/** Get waypoint X displacement. */
mn16_t
GShip::getWaypointDX() const
{
    return mn16_t(getDisplayedShip().waypoint_dx);
}

/** Get waypoint Y displacement. */
mn16_t
GShip::getWaypointDY() const
{
    return mn16_t(getDisplayedShip().waypoint_dy);
}

/** Get ship's heading vector.
    \return value from [0,360), corresponding to VGAP's heading angles.
    Returns unknown value if ship doesn't move or heading not known */
mp16_t
GShip::getHeading() const
{
    if (ship_info != 0) {
        if (ship_info->data.waypoint_dx != UNKNOWNS && ship_info->data.waypoint_dy != UNKNOWNS) {
            if (ship_info->data.waypoint_dx == 0 && ship_info->data.waypoint_dy == 0) {
                return mp16_t();
            } else {
                return mp16_t(int(getHeadingDeg(ship_info->data.waypoint_dx, ship_info->data.waypoint_dy)));
            }
        } else {
            return mp16_t(ship_info->heading);
        }
    } else {
        return mp16_t();
    }
}

/** Get warp factor.
    \return warp factor or unknown */
mp16_t
GShip::getWarp() const
{
    return mp16_t(getDisplayedShip().warp);
}

/** Set warp factor. */
void
GShip::setWarp(mp16_t warp)
{
    if (ship_info != 0 || warp.isKnown()) {
        createShipInfo().data.warp = warp.getRawValue();
        markDirty();
    }
}

/** Check whether ship is hyperdriving.
    \return true iff ship has hyperdrive and tries to activate it */
bool
GShip::isHyperdriving() const
{
    return map_info.kind == CurrentShip
        && canDoSpecial(hf_Hyperdrive)
        && ship_info->data.warp > 0
        && std::memcmp("HYP", ship_info->data.fcode, 3) == 0;
}


/*
 *  Equipment accessors
 */

/** Get engine type.
    \return engine type [1,NUM_ENGINES], or unknown */
mp16_t
GShip::getEngineType() const
{
    return mp16_t(getDisplayedShip().engine);
}

/** Set engine type (for history).
    \param engine engine type [1,NUM_ENGINES], or unknown */
void
GShip::setEngineType(mp16_t engine)
{
    createShipInfo().data.engine = engine.getRawValue();
    markDirty();
}

/** Get beam type.
    \return beam type [1,NUM_BEAMS], or unknown */
mp16_t
GShip::getBeamType() const
{
    return mp16_t(getDisplayedShip().beam_type);
}

/** Set beam type (for history).
    \param type beam type [1,NUM_BEAMS], or unknown */
void
GShip::setBeamType(mp16_t type)
{
    createShipInfo().data.beam_type = type.getRawValue();
    markDirty();
}

/** Get number of beams.
    \return beam count, or unknown */
mp16_t
GShip::getNumBeams() const
{
    const TShip& sh = getDisplayedShip();
    if (sh.beam_type == 0) {
        return mp16_t(0);
    } else {
        return mp16_t(sh.beam_count);
    }
}

/** Set number of beams (for history).
    \param count beam count, or unknown */
void
GShip::setNumBeams(mp16_t count)
{
    createShipInfo().data.beam_count = count.getRawValue();
    markDirty();
}

/** Get bay count.
    \return bay count, or unknown */
mp16_t
GShip::getNumBays() const
{
    return mp16_t(getDisplayedShip().bay_count);
}

/** Set bay count (for history).
    \param count bay count, or unknown */
void
GShip::setNumBays(mp16_t count)
{
    createShipInfo().data.bay_count = count.getRawValue();
    markDirty();
}

/** Get torpedo type.
    \return torpedo type [1,NUM_TORPS], or unknown */
mp16_t
GShip::getTorpType() const
{
    return mp16_t(getDisplayedShip().torp_type);
}

/** Set torpedo type (for history).
    \param type beam type [1,NUM_TORPS], or unknown */
void
GShip::setTorpType(mp16_t type)
{
    createShipInfo().data.torp_type = type.getRawValue();
    markDirty();
}

/** Get torpedo launcher count.
    \return torpedo launcher count, or unknown */
mp16_t
GShip::getNumTorpLaunchers() const
{
    const TShip& sh = getDisplayedShip();
    if (sh.torp_type == 0) {
        return mp16_t(0);
    } else {
        return mp16_t(sh.torp_launcher_count);
    }
}

/** Set torpedo launcher count (for history).
    \param count torpedo launcher count, or unknown */
void
GShip::setNumTorpLaunchers(mp16_t count)
{
    createShipInfo().data.torp_launcher_count = count.getRawValue();
    markDirty();
}


/*
 *  Mission accessors
 */

/** Set ship name.
    \param str New name */
void
GShip::setName(const string_t& str)
{
    storeBasicString(createShipInfo().data.name, convertUtf8ToGame(str));
    markDirty();
}

/** Get ship mission.
    \return mission number, or unknown */
mp16_t
GShip::getMission() const
{
    return mp16_t(getDisplayedShip().mission);
}

/** Set ship mission.
    \param m Mission number
    \param i Intercept number
    \param t Tow number

    FIXME: should this take a TMissionData object? */
void
GShip::setMission(mp16_t m, mp16_t i, mp16_t t)
{
    if (ship_info != 0 || m.isKnown() || i.isKnown() || t.isKnown()) {
        ShipInfo& info = createShipInfo();
        if (m.getRawValue() != info.data.mission
            || i.getRawValue() != info.data.mission_intercept
            || t.getRawValue() != info.data.mission_tow)
        {
            info.data.mission           = m.getRawValue();
            info.data.mission_intercept = i.getRawValue();
            info.data.mission_tow       = t.getRawValue();
            markDirty();
        }
    }
}

/** Get tow number. */
mp16_t
GShip::getTowId() const
{
    return mp16_t(getDisplayedShip().mission_tow);
}

/** Get intercept number. */
mp16_t
GShip::getInterceptId() const
{
    return mp16_t(getDisplayedShip().mission_intercept);
}

/** Get primary enemy. */
mp16_t
GShip::getPrimaryEnemy() const
{
    return mp16_t(getDisplayedShip().primary_enemy);
}

/** Set primary enemy.
    \param pe New primary enemy, [0,NUM_PLAYERS] */
void
GShip::setPrimaryEnemy(mp16_t pe)
{
    if (ship_info != 0 || pe.isKnown()) {
        createShipInfo().data.primary_enemy = pe.getRawValue();
        markDirty();
    }
}

/** Get ship damage.
    \return damage level, or unknown */
mp16_t
GShip::getDamage() const
{
    return mp16_t(getDisplayedShip().damage);
}

/** Set ship damage.
    \param damage damage level, or unknown */
void
GShip::setDamage(mp16_t damage)
{
    if (ship_info != 0 || damage.isKnown()) {
        createShipInfo().data.damage = damage.getRawValue();
        markDirty();
    }
}

/** Get ship crew.
    \return number of crewmen, or unknown */
mp16_t
GShip::getCrew() const
{
    return mp16_t(getDisplayedShip().crew);
}

/** Set ship crew.
    \param crew number of crewmen, or unknown */
void
GShip::setCrew(mp16_t crew)
{
    if (ship_info != 0 || crew.isKnown()) {
        createShipInfo().data.crew = crew.getRawValue();
        markDirty();
    }
}

/** Get friendly code. */
string_t
GShip::getFCode() const
{
    return convertGameToUtf8(getBasicString(getDisplayedShip().fcode));
}

/** Set friendly code. */
void
GShip::setFCode(const string_t& str)
{
    storeBasicString(createShipInfo().data.fcode, convertUtf8ToGame(str));
    markDirty();
}

/** Check whether fcode is known. */
bool
GShip::isFCodeKnown() const
{
    return ship_info != 0
        && (ship_info->data.fcode[0] != UNKNOWNC
            || ship_info->data.fcode[1] != UNKNOWNC
            || ship_info->data.fcode[2] != UNKNOWNC);
}


/*
 *  Cargo accessors
 */

/** Get ammunition on ship, raw version. This does not honor reservations.
    It returns the amount of torpedoes/fighters, regardless of their type. */
mp16_t
GShip::getAmmoRaw() const
{
    return mp16_t(getDisplayedShip().ammo);
}

/** Get ammunition on ship.
    This honors the cargo reservations, i.e. the return value does not include cargo
    that has been promised to ongoing transactions. */
mn32_t
GShip::getAmmo() const
{
    /* The cleanest way to implement this would be to derive a type, and then call getCargo().
       This implementation returns a valid value if the ship has ammo, but no known weapons,
       which is normally an error but can possibly happen through history. Thus, the mass
       computations still work out.

       Actually, we need the arbiter computations only for own ships for which we know everything,
       only those will have reserved cargo. However, exploiting that fact doesn't save us any
       code here. */
    const TShip& dpy = getDisplayedShip();
    if (mp16_t(dpy.ammo).isKnown()) {
        int32_t result = dpy.ammo;
        if (dpy.bay_count > 0) {
            result -= cargo_arbiter.getReservedCargo(el_Fighters);
        } else if (dpy.torp_launcher_count > 0 && dpy.torp_type > 0 && dpy.torp_type <= NUM_TORPS) {
            result -= cargo_arbiter.getReservedCargo(getCargoTypeFromTorpType(dpy.torp_type));
        } else {
            // unknown type
        }
        return mn32_t(result);
    } else {
        return mn32_t();
    }
}

/** Get cargo amount on ship, raw version. This does not honor reservations. */
mp16_t
GShip::getCargoRaw(GCargoType type) const
{
    const TShip& dpy = getDisplayedShip();
    switch (type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        return mp16_t(dpy.ore_cargo[type]);
     case el_Fighters:
        if (dpy.bay_count > 0) {
            return mp16_t(dpy.ammo);
        } else if (dpy.bay_count == 0) {
            return mp16_t(0);
        } else {
            return mp16_t();
        }
     case el_Colonists:
        return mp16_t(ship_info->data.colonists);
     case el_Supplies:
        return mp16_t(ship_info->data.supplies);
     case el_Money:
        return mp16_t(ship_info->data.money);
     default:
        ASSERT(type >= el_Torps && type < el_COUNT);
        if (dpy.torp_type == (type - el_Torps+1) && dpy.torp_launcher_count > 0) {
            // We know it has this type of torpedoes
            return mp16_t(dpy.ammo);
        } else if (dpy.torp_type >= 0 || dpy.torp_launcher_count == 0) {
            // We know it has torpedoes, but not this type
            // -or- it has no torpedoes
            return mp16_t(0);
        } else {
            return mp16_t();
        }
    }
}

/** Get cargo amount on ship.
    This honors the cargo reservations, i.e. the return value does not include cargo
    that has been promised to ongoing transactions. */
mn32_t
GShip::getCargo(GCargoType type) const
{
    mp16_t have = getCargoRaw(type);
    if (have.isKnown()) {
        return mn32_t(have - cargo_arbiter.getReservedCargo(type));
    } else {
        return mn32_t();
    }
}

/** Change amount that is on ship. */
void
GShip::changeCargoRaw(GCargoType type, int amount)
{
    if (amount == 0)
        return;

    // FIXME: we can partially allow this on non-current ships.
    // FIXME: this trusts the type, i.e. does not refuse adding torps to a carrier.
    // FIXME: we need a way to directly set a ship's cargo, for
    // foreign ships. This is needed for various scans:
    // - web drain complete (=set N to 0)
    // - combat result (=set ammo)
    ASSERT(hasFullShipData());
    markDirty();

    switch (type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        ship_info->data.ore_cargo[type] += amount;
        break;
     case el_Fighters:
        ship_info->data.ammo += amount;
        break;
     case el_Colonists:
        ship_info->data.colonists += amount;
        break;
     case el_Supplies:
        ship_info->data.supplies += amount;
        break;
     case el_Money:
        ship_info->data.money += amount;
        break;
     default:
        ASSERT(type >= el_Torps && type < el_COUNT);
        ship_info->data.ammo += amount;
        break;
    }
}

mn32_t
GShip::getFreeCargo() const
{
    if (map_info.kind == CurrentShip) {
         return mn32_t(getHull().getCargo()
                       - getCargo(el_Tritanium)
                       - getCargo(el_Duranium)
                       - getCargo(el_Molybdenum)
                       - getAmmo()
                       - getCargo(el_Colonists)
                       - getCargo(el_Supplies));
    } else {
        return mn32_t();
    }
}


/*
 *  Transporter access
 */

/** Access a transporter.
    \param which which_one transporter to access. */
TShipTransfer&
GShip::getTransporter(Transporter which_one) throw()
{
    return which_one == t_Transfer ? ship_info->data.transfer : ship_info->data.unload;
}

/** Access a transporter.
    \param which which_one transporter to access. */
const TShipTransfer&
GShip::getTransporter(Transporter which_one) const throw()
{
    return which_one == t_Transfer ? ship_info->data.transfer : ship_info->data.unload;
}

/** Check whether a transporter is active. An active transporter has a
    full transfer order. */
bool
GShip::isTransporterActive(Transporter which_one) const throw()
{
    if (hasFullShipData()) {
        const TShipTransfer& t = getTransporter(which_one);
        return (t.ore[0] | t.ore[1] | t.ore[2] | t.ore[3] | t.colonists | t.supplies) != 0;
    } else {
        return false;
    }
}

/** Check whether a transporter is cancellable. A transporter is
    cancellable if its content can be moved back into main cargo room
    without overloading the ship. */
bool
GShip::isTransporterCancellable(Transporter which_one) const throw()
{
    // FIXME: this does not guarantee correctness in all cases: Take a
    // ship with 10 cargo room left, and 10 kt in the transporter. A
    // pending transaction adds 10 to the ship (i.e. it got the
    // guarantee that it'll be able to add this). When we now do
    // cancelTranporter() and commit that transaction, the ship is
    // overloaded.
    if (hasFullShipData()) {
        const TShipTransfer& t = getTransporter(which_one);
        long sum = (long) t.ore[0] + t.ore[1] + t.ore[2] + t.ore[3] + t.supplies + t.colonists;
        return sum == 0 || sum <= getFreeCargo();
    } else {
        return false;
    }
}

/** Cancel a transporter. Moves its contents back to main cargo room.
    \pre isTransporterCancellable(which_one) */
void
GShip::cancelTranporter(Transporter which_one) throw()
{
    /* Because the transporter is guaranteed to be cancellable,
       we know that this ship has full data. */
    ASSERT(isTransporterCancellable(which_one));
    TShipTransfer& t = getTransporter(which_one);
    ship_info->data.ore_cargo[0] += t.ore[0];
    ship_info->data.ore_cargo[1] += t.ore[1];
    ship_info->data.ore_cargo[2] += t.ore[2];
    ship_info->data.ore_cargo[3] += t.ore[3];
    ship_info->data.colonists += t.colonists;
    ship_info->data.supplies += t.supplies;
    t.ore[0] = t.ore[1] = t.ore[2] = t.ore[3] = t.supplies = t.colonists = t.id = 0;

    markDirty();
}

/** Check whether transporter is in use. A transporter is in use if
    the user is currently preparing a transport order. This means we
    must deny all other attempts of starting to prepare such an
    order. */
bool
GShip::isTransporterInUse(Transporter which_one) const throw()
{
    return transporter_in_use[which_one];
}

/** Set transporter in-use mark.
    \see isTransporterInUse */
void
GShip::setTransporterInUse(Transporter which_one, bool in_use) throw()
{
    transporter_in_use[which_one] = in_use;
}


/*
 *  Fleet accessors
 */

/** Set number of the fleet this ship is in.
    This function just sets the internal flag.
    Do not use this function directly; use ::setFleetNumber(GUniverse&,int,int) instead.
    That function will update all dependant information. */
void
GShip::setFleetNumber(int fno)
{
    this->map_info.fleet_nr = fno;
}

/** Get number of the fleet this ship is in.
    \return fleet number (a ship Id), or 0 */
int
GShip::getFleetNumber() const
{
    return map_info.fleet_nr;
}

/** Set name of the fleet led by this ship. */
void
GShip::setFleetName(string_t name)
{
    this->map_info.fleet_name = name;
}

/** Get name of the fleet led by this ship (if any). */
const string_t&
GShip::getFleetName() const
{
    return map_info.fleet_name;
}

/** Check for fleet leader. */
bool
GShip::isFleetLeader() const
{
    return map_info.fleet_nr == map_info.id;
}

/** Check for fleet member.
    \return true if this is a fleet member (but not a leader) */
bool
GShip::isFleetMember() const
{
    return map_info.fleet_nr != 0 && map_info.fleet_nr != map_info.id;
}

void
GShip::setPredictedPos(GPoint pt)
{
    map_info.predicted_pos = pt;
}

GPoint
GShip::getPredictedPos() const
{
    return map_info.predicted_pos;
}


/*
 *  Function accessors
 */

/** Check whether this ship can do special function.
    \param basic_function A basic ship function, e.g. hf_Cloak. */
bool
GShip::canDoSpecial(int basic_function) const
{
    /* Do we know the hull?
       FIXME: we could check racial specials */
    if (!getHullId().isKnown())
        return false;

    /* Figure out experience level */
    int level = unit_scores.getScore(ship_score_definitions.lookupScore(ScoreId_ExpLevel));
    if (level < 0)
        level = 0;

    /* Check class functions */
    if (getHull().canDoSpecial(basic_function, getRealOwner(), true, GExpLevelSet(level)))
        return true;

    /* Check this ship's functions */
    const GHullFunctionData& fd = GHullFunctionData::instance;
    if (map_info.special_functions.size() != 0) {
        /* Ship has special functions. Figure out levels. */
        GExpLevelSet allLevels = GExpLevelSet::allUpTo(config.NumExperienceLevels());
        GExpLevelSet thisLevel = GExpLevelSet(level);

        /* Check functions */
        for (std::size_t i = 0; i < map_info.special_functions.size(); ++i) {
            GHullFunction f;
            if (fd.getFunctionDefinition(fd.getIdFromHostId(map_info.special_functions[i]), f)
                && f.getBasicFunctionId() == basic_function
                && (f.getLevels().contains(allLevels) || f.getLevels().containsAnyOf(thisLevel)))
            {
                return true;
            }
        }
    }

    /* Not found */
    return false;
}

/** Enumerate this ship's special functions (only functions assigned to this
    ship, not including class functions). */
void
GShip::enumShipSpecificSpecials(GHullFunctionList& list) const
{
    const GHullFunctionData& fd = GHullFunctionData::instance;
    for (std::size_t i = 0; i < map_info.special_functions.size(); ++i) {
        GHullFunction f;
        if (fd.getFunctionDefinition(fd.getIdFromHostId(map_info.special_functions[i]), f)) {
            list.add(f);
        }
    }
}

/** Check whether this ship has any ship-specific functions. */
bool
GShip::hasAnyShipSpecificSpecials() const
{
    return map_info.special_functions.size() != 0;
}

const std::vector<int>&
GShip::getAssociatedMessages() const
{
    return associated_messages;
}


/*
 *  Data access
 */

/** Create ShipInfo if it doesn't exist yet. Returns reference to it. */
GShip::ShipInfo&
GShip::createShipInfo()
{
    if (!ship_info) {
        ship_info = new ShipInfo(map_info.id);
        if (map_info.kind == NoShip) {
            map_info.kind = HistoryShip;
        }
    }
    return *ship_info;
}

/** Get read-only ship data. */
const TShip&
GShip::getDisplayedShip() const
{
    return ship_info != 0 ? ship_info->data : blank_ship;
}

/** Adjust ship track info to start no later than turn.
    \post turn <= info.track_turn */
void
GShip::adjustShipTrack(int turn)
{
    ShipInfo& info = createShipInfo();
    if (turn > info.track_turn) {
        int adjust = turn - info.track_turn;
        for (int i = NUM_SHIP_TRACK_ENTRIES-1; i >= 0; --i) {
            info.track[i] = (i >= adjust
                             ? info.track[i - adjust]
                             : blank_ship_track);
        }
        info.track_turn = turn;
    }
}
