/**
  *  \file resmgr/resdir.cc
  *  \brief ResDirectory resource provider class
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "resmgr/resdir.h"
#include "gfx/font.h"
#include "gfx/blit.h"
#include "gfx/bmfont.h"
#include "io/textfile.h"
#include "util/console.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "ui/uiglobal.h"
#include "arch/platform.h"
#include "util/misc.h"

namespace {
    class Scanner {
        string_t str;
        string_t::size_type pos;
     public:
        Scanner(string_t str)
            : str(str),
              pos(0)
            { }
        bool scanString(const char* s);
        bool scanInt(int& out);
        bool scanEnd()
            { return pos == str.size(); }
    };
}

bool
Scanner::scanString(const char* s)
{
    using std::strlen;
    string_t::size_type len = strlen(s);
    if (str.size() - pos >= len && str.compare(pos, len, s, len) == 0) {
        pos += len;
        return true;
    } else {
        return false;
    }
}

bool
Scanner::scanInt(int& out)
{
    string_t::size_type err, err2;
    if (stringValue(str.substr(pos), out, err)) {
        // completely valid
        pos = str.size();
        return true;
    } else if (err != 0 && stringValue(str.substr(pos, err), out, err2)) {
        // part is valid
        pos += err;
        return true;
    } else {
        // invalid
        return false;
    }
}


/** Add constant to all bytes in a pixmap. It must be a palettized pixmap. */
static void
addToPixelValue(GfxPixmap& pix, int toAdd)
{
    GfxSystemSurface sfc = pix.getSystemSurface();
    uint8_t* pixels = static_cast<uint8_t*>(sfc->pixels);
    uint32_t w = sfc->w;
    uint32_t h = sfc->h;
    uint32_t pitch = sfc->pitch;
    for (uint32_t y = 0; y < h; ++y) {
        for (uint32_t x = 0; x < w; ++x) {
            pixels[x] += toAdd;
        }
        pixels += pitch;
    }
}


/** Convert resource identifier string value into file name template.
    For example, "foo.bar90" is converted to "foo/bar90.", which tells
    openResourceFile() to look for a file "bar90" with any extension
    within directory "foo".
    \param name [in/out] Resource name */
static void
convertName(string_t& name)
{
    string_t::size_type n = 0;
    while ((n = name.find('.', n)) != string_t::npos)
        name[n] = Platform::PATH_SEPARATOR, ++n;
    name += '.';
}

/****************************** ResDirectory *****************************/

/*! \class ResDirectory

    This manages resources stored in individual files. It provides
    management of alias definitions (index.txt) as well as heuristics
    for guessing file names from resource Ids: a resource "foo.bar" is
    looked up in file "foo/bar.EXT", where EXT is an extension
    matching the resource type.

    For graphics, it supports some additional tricks. */

/** Constructor. */
ResDirectory::ResDirectory(Ptr<Directory> dir)
    : dir(dir)
{
    Ptr<Stream> s = dir->openFileNT("index.txt", Stream::C_READ);
    if (s) {
        loadAliases(*s);
    }
}

ResDirectory::~ResDirectory()
{ }

Ptr<GfxPixmap>
ResDirectory::loadPixmap(ResId id)
{
    /* Fetch definition for this item */
    string_t name = id.toString();
    alias_map_t::iterator i = alias_map.find("g." + name);
    if (i != alias_map.end()) {
        name = i->second.name;
    } else {
        convertName(name);
    }

    /* Parse it */
    string_t file = strRTrim(strFirst(name, "|"));
    Ptr<Stream> stream = openResourceFile(*dir, file, gfx_suffix_list);
    if (!stream)
        return 0;
    Ptr<GfxPixmap> pix = loadImage(*stream);
    if (!pix)
        return pix;
    stream = 0;

    /* Parse modifications */
    while (strRemove(name, "|")) {
        string_t op = strTrim(strFirst(name, "|"));
        Scanner ops(op);
        if (op.size() == 0) {
            /* Blank */
        } else if (ui_screen != 0 && ops.scanString("size:")) {
            /* Resize */
            int width = -1, height = -1;
            if (ops.scanString("screen")) {
                width  = ui_screen->getSize().x;
                height = ui_screen->getSize().y;
            } else {
                if (ops.scanInt(width)) {
                    if (ops.scanString("%")) {
                        width = ui_screen->getSize().x * width / 100;
                    }
                }
                if (ops.scanString(",") && ops.scanInt(height)) {
                    if (ops.scanString("%")) {
                        height = ui_screen->getSize().y * height / 100;
                    }
                }
            }

            if (ops.scanEnd() && width > 0 && height > 0) {
                Ptr<GfxPixmap> npix = new GfxPixmap(width, height, *pix);
                doStretchRotateBlit(*pix, *npix,
                                    GfxRect(*pix), GfxRect(*npix),
                                    0, 0,
                                    width, 0,
                                    0, height);
                pix = npix;
            }
        } else if (ops.scanString("add:")) {
            /* Add value to each pixel */
            int toAdd;
            if (ops.scanInt(toAdd) && ops.scanEnd() && pix->getBitsPerPixel() == 8) {
                addToPixelValue(*pix, toAdd);
            }
        } else {
            if (Ptr<Stream> stream = openResourceFile(*dir, op, gfx_suffix_list)) {
                Ptr<GfxPixmap> ovl = loadImage(*stream);
                if (ovl)
                    ovl->blit(*pix, 0, 0);
            }
        }
    }

    return pix;
}

SfxSampleData*
ResDirectory::loadSound(ResId id)
{
    string_t name = id.toString();
    alias_map_t::iterator i = alias_map.find("s." + name);
    if (i != alias_map.end()) {
        name = i->second.name;
    } else {
        convertName(name);
    }

    Ptr<Stream> st = openResourceFile(*dir, name, sfx_suffix_list);
    if (st) {
        // FIXME: exceptions?
        try {
            return new SfxSampleData(*st);
        }
        catch (...)
        { }
    }
    return 0;
}

Ptr<GfxAnimation>
ResDirectory::loadAnimation(ResId /*id*/)
{
    return 0;
}

Ptr<GfxFont>
ResDirectory::getFont(ResId id)
{
    string_t name = id.toString();
    alias_map_t::iterator i = alias_map.find("f." + name);
    if (i != alias_map.end()) {
        name = i->second.name;
    } else {
        convertName(name);
    }

    Ptr<Stream> s = openResourceFile(*dir, name, fnt_suffix_list);
    if (s) {
        Ptr<GfxBitmapFont> res = new GfxBitmapFont();
        res->load(*s, 0);
        return res;
    } else {
        return 0;
    }
}

/** Load alias table from given stream. */
void
ResDirectory::loadAliases(Stream& in)
{
    TextFile tf(in);
    string_t line, l, r;
    while (tf.getLine(line)) {
        /* strip comment */
        string_t::size_type p = line.find_first_of(";#");
        if (p != line.npos)
            line.erase(p);
        line = strTrim(line);
        if (line.empty())
            continue;

        /* must be "normal" line */
        if (strSplit(line, l, r, "=")) {
            alias_map[strTrim(l)].name = strTrim(r);
        } else {
            console.write(LOG_ERROR, format(_("%s:%d: file format error -- line ignored"),
                                            in.getName(), tf.getLineNr()));
        }
    }
}
