#
#  A dead-simple module to parse XML.
#
#  This is intended for fast transformations of XML to HTML, assuming
#  the input-XML is not too far from HTML. It reads XML input into a
#  nice data structure.
#
#  It supports macros (entities) from inline DTD fragments and SYSTEM
#  DTDs, and well-formed XML input. It is rather fast.
#
#  It does NOT support: namespaces (do it yourself if you insist),
#  "proper" entities (i.e. "&#32;" should normally come out as a
#  space; undefined entities should be reported instead of copied to
#  output), character set recoding (characters >=128 are in no way
#  handled specially), correct whitespace trimming rules.
#
package SimpleXML;

use 5.000;
use Exporter;
use vars qw{ @ISA @EXPORT };

@ISA = ('Exporter');
@EXPORT = qw{ lex_set_text lex_set_macro lex parse_document expand_macros
              $lex_token $lex_value %lex_attribs @lex_dtd_path
              lex_set_position lex_count_lines lex_pis lex_forget_pis };

use vars ('$lex_text',                   # text to parse
          '$lex_token',                  # token type ('open', 'close', 'text')
          '$lex_value',                  # token name, or text
          '%lex_attribs',                # for open tags, the attributes
          '%lex_macros',                 # macros
          '$lex_filename',
          '$lex_linenumber',
          '$lex_pending_close',
          '@lex_dtd_path',
          '@lex_pis',
         );
@lex_dtd_path = ('.');

# Set the text to parse. Must be a complete XML document.
# For example, use
#   open FILE, "...";
#   lex_set_text (<FILE>);
sub lex_set_text {
    $lex_text = join '', @_;
    $lex_pending_close = undef;
}

# Set logical position for parse
sub lex_set_position {
    $lex_filename = shift;
    $lex_linenumber = shift;
}

# Set a entity ("macro"), as in 'lex_set_macro("szlig", "&#223;")'.
# These entity will be expanded in input. Other entities will be left
# alone and therefore appear in output.
sub lex_set_macro {
    while (@_) {
        my $key = shift;
        $lex_macros{$key} = shift;
    }
}

sub lex_pis { @lex_pis }

sub lex_forget_pis {
    @lex_pis = ();
}

# sub lex_read_file {
#     my $fd = shift;
#     $lex_text = join '', <$fd>;
# }

# Read one token. Returns
# - $lex_token    type of token (open/close/text/eof)
# - $lex_value    value of token (tag, or text)
# - %lex_attribs  attributes, if it is an opening tag
sub lex {
    if (defined $lex_pending_close) {
        $lex_value = $lex_pending_close;
        $lex_token = 'close';
        $lex_pending_close = undef;
        return $lex_token;
    }

    while (1) {
        if ($lex_text =~ m{\G<!--(.*?)-->}sgc) {
            # empty
            $lex_linenumber += lex_count_lines($1)
              if defined $lex_linenumber;
        } elsif ($lex_text =~ m{\G<\?([-\w:]+)(.*?)>}sgc) {
            # processing instruction. Just collect these
            my $pi_name = $1;
            my %pi_attribs = ();
            my $attribs = $2;
            $lex_linenumber += lex_count_lines($2)
              if defined $lex_linenumber;
            while ($attribs =~ m{([-\w:]+)="(.*?)"}sgc) { $pi_attribs{$1} = expand_macros($2); }
            push @lex_pis, [ $pi_name, \%pi_attribs ];
        } elsif ($lex_text =~ m{\G<!DOCTYPE(\s*)}sgci) {
            # doctype definition. Might be followed by a doctype
            # fragment or DTD definition
            my $origpos = pos($lex_text);
            $lex_linenumber += lex_count_lines($1)
              if defined $lex_linenumber;
            if ($lex_text =~ m{\G\w+\s*}sgc) { }
            if ($lex_text =~ m{\G(system|public)\s+}sgci) {
                my $is_sys = uc($1) eq 'SYSTEM';
                if ($lex_text =~ m{\G([-a-z0-9.:]+)\s*}sgc ||
                    $lex_text =~ m{\G"([^\"]*)"\s*}sgc ||
                    $lex_text =~ m{\G'([^\']*)'\s*}sgc)
                {
                    handle_dtd ($is_sys, $1);
                }
            }
            $lex_text =~ m{[^\[>]*}sgc; # eat up random garbage
            if ($lex_text =~ m{\G\[}sgc) {
                while (1) {
                    if ($lex_text =~ m{\G\].*?>}sgc) {
                        last;
                    } elsif ($lex_text =~ m{\G<!ENTITY\s+([a-z0-9]+)\s+("[^\"]*"|'[^\']*')\s*>}sgci) {
                        lex_set_macro ($1, expand_macros_and_nums (substr($2, 1, length($2)-2)));
                    } elsif ($lex_text =~ m{\G<[^>]*>}sgc) {
                    } elsif ($lex_text =~ m{\G[^<\]]+}sgc) {
                    } else {
                        die "invalid doctype fragment at '" . substr($lex_text, pos($lex_text), 30) . "'";
                    }
                }
            } elsif ($lex_text =~ m{\G.*?>}sgc) {
            } else {
                die "invalid doctype declaraion";
            }
            $lex_linenumber += lex_count_lines (substr($lex_text, $origpos, pos($lex_text)-$origpos))
              if defined($lex_linenumber);
            # print "doctype fragment ends at ", pos($lex_text), "\n";
        } elsif ($lex_text =~ m{\G</([-\w:]+)>}sgc) {
            # close tag
            $lex_token = 'close';
            $lex_value = $1;
            return $lex_token;
        } elsif ($lex_text =~ m{\G<([-\w:]+)(\s*)(.*?)>}sgc) {
            # open tag
            $lex_token = 'open';
            $lex_value = $1;
            %lex_attribs = ();
            my $whitespace = $2;
            my $attribs = $3;
            # FIXME: this sucks.
            while ($attribs =~ m{([-\w:]+)="(.*?)"}sgc) { $lex_attribs{$1} = expand_macros($2); }
            #while ($attribs =~ m{([-\w:]+)='(.*?)'}sgc) { $lex_attribs{$1} = expand_macros($2); }
            $lex_pending_close = $lex_value
              if $attribs =~ m|/$|;
            $lex_linenumber += lex_count_lines ($whitespace . $attribs)
              if defined ($lex_linenumber);
            return $lex_token;
        } elsif ($lex_text =~ m{\G([^<]+)}sgc) {
            # text
            my $text = $1;
            $lex_token = 'text';
            $lex_value = expand_macros($text);
            $lex_linenumber += lex_count_lines ($text)
              if defined ($lex_linenumber);
            return $lex_token;
        } elsif ($lex_text =~ m{\G\Z}sgc) {
            return ($lex_token = 'eof');
        } else {
            die "parse error at '" . substr($lex_text, pos($lex_text), 30) . "'";
        }
    }
}

# Parse a document. Assumes you have already read a token (as usual
# for recursive-descent parsers). Returns the document, as reference
# to a list of elements. Each element is
#  - text [input was text]
#  - reference to a sublist [input was a tag; sublist contains
#    name of tag, reference to attribs, reference to contents].
# Example: "This is <a href="target">a link</a>." will parse as
# ["This is ", ["a", { href => 'target' }, ["a link"]], "."].
sub parse_document {
    # A document is a list of
    # - text [cdata]
    # - ref [sublist, containing: name, reference to attribs, reference to contents]
    my @list = ();
    while (1) {
        if ($lex_token eq 'text') {
            push @list, $lex_value
                unless @list==0 && $lex_value !~ /\S/;
            lex ();
        } elsif ($lex_token eq 'open') {
            my %sub_attrs = %lex_attribs;
            my @sublist = ($lex_value, \%sub_attrs, undef, [ $lex_filename, $lex_linenumber ]);
            lex ();
            $sublist[2] = parse_document ();
            if ($lex_token ne 'close' || $lex_value ne $sublist[0]) {
                my $prefix = "";
                $prefix .= "$lex_filename: " if defined($lex_filename);
                $prefix .= "line $lex_linenumber: " if defined($lex_linenumber);
                my $what = ($lex_token eq 'close') ? "'</$lex_value>'" : "$lex_token \[$lex_value]";
                die "${prefix}expected '</$sublist[0]>', got $what ($.)";
            }
            push @list, \@sublist;
            lex ();
        } else {
            pop @list
                if @list && !ref($list[-1]) && $list[-1] !~ /\S/;
            return \@list;
        }
    }
}

# Expand entities / macros in a string.
sub expand_macros {
    my $text = shift;
    $text =~ s{&([^;]*);}{defined($lex_macros{$1}) ? $lex_macros{$1} : "&$1;"}ge;
    $text;
}

sub expand_macros_and_nums {
    my $text = expand_macros(@_);
#print STDERR "[[$text]]\n" if $text =~ m{&};
    $text =~ s{&\#([0-9]+);}{chr($1)}ge;
    $text;
}

# Process doctype
#   $_[0] = 1 if "system" DTD, 0 if "public" DTD
#   $_[1] = doctype
sub handle_dtd {
    my ($is_sys, $file) = @_;
    if ($is_sys) {
        foreach (@lex_dtd_path) {
            if (open DTD, "< $_/$file") {
                process_dtd (join '', <DTD>);
                close DTD;
                last;
            }
        }
    }
}

sub process_dtd {
    my $dtd = join '', @_;
    # FIXME? This is almost a copy of the respective code in lex
    while (1) {
        if ($dtd =~ m{\G\Z}sgc) {
            last;
        } elsif ($dtd =~ m{\G<!--.*?-->}sgc) {
        } elsif ($dtd =~ m{\G<!ENTITY\s+([a-z0-9]+)\s+("[^\"]*"|'[^\']*')\s*>}sgci) {
            lex_set_macro ($1, expand_macros_and_nums(substr($2, 1, length($2)-2)));
        } elsif ($dtd =~ m{\G<[^>]*>}sgc) {
        } elsif ($dtd =~ m{\G[^<\]]+}sgc) {
        } else {
            die "invalid doctype fragment at '" . substr($dtd, pos($dtd), 30) . "'";
        }
    }
}

sub lex_count_lines {
    my $count = 0;
    foreach (@_) {
        foreach (m{\n}g) { ++$count }
    }
    $count;
}

1;
