/**
  *  \file ui/widgets/abstractbutton.cc
  *  \brief UIAbstractButton class
  *
  *  (c) 2001-2024 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/widgets/abstractbutton.h"
#include "ui/commands.h"
#include "cpluslib/assert.h"

/**************************** UIAbstractButton ***************************/

/** Construct new button. The button will not react upon keys, just
    mouse-clicks. */
UIAbstractButton::UIAbstractButton()
    : UIWidget(),
      key(0)
{ }

/** Construct new key button. The button will post a UIKey event.
    \param key key that triggers this button. */
UIAbstractButton::UIAbstractButton(UIKey key)
    : UIWidget(),
      key(key)
{
    setState(bf_Key, true);
}

/** Construct new button.
    \param id widget id, used for emitting commands etc.
    \param key key that triggers this button */
UIAbstractButton::UIAbstractButton(int id, UIKey key)
    : UIWidget(id),
      key(key)
{ }

/** Destructor. */
UIAbstractButton::~UIAbstractButton()
{ }

/** Add command. The button will post a UICommand(id, arg) when pressed. */
UIAbstractButton&
UIAbstractButton::addCommand()
{
    sig_fired.add(this, &UIAbstractButton::postCommand);
    return *this;
}

/** Add command. The button will post a stop(id) when pressed. */
UIAbstractButton&
UIAbstractButton::addStop()
{
    sig_fired.add(this, &UIAbstractButton::callStop);
    return *this;
}

/** Make this a "key" button. */
UIAbstractButton&
UIAbstractButton::addKey()
{
    setState(bf_Key, true);
    return *this;
}

/** Change button state. When the button loses logical focus, it must
    be un-pressed. Certain events also need redraw. */
void
UIAbstractButton::onStateChange(int astate, bool enable)
{
    UIWidget::onStateChange(astate, enable);

    // FIXME: can we get rid of this status manipulation?
    if (astate & st_Disabled)
        status &= ~bf_Pressed;

    if ((status & (bf_Pressed | st_Selected)) == bf_Pressed)
        status &= ~bf_Pressed;

    if (astate & (bf_Pressed | st_Selected | bf_Highlight | st_Disabled | bf_PressOverride))
        drawWidget(false);
}

/** Event handler. Processes mouse and keyboard events, and calls
    fire() when needed. */
bool
UIAbstractButton::handleEvent(const UIEvent& e, bool second_pass)
{
    switch(e.type) {
     case ev_Key:
        if (!hasState(st_Disabled) && !hasState(bf_Key)
            && (e.key.code == key
                || (second_pass && (e.key.code == (key & ~ss_Alt)
                                    || (key == '#' && e.key.code == '\\')))))
        {
            setSelectedWidget();
            fire(e.key.arg);
            return true;
        }
        break;

     case ev_Mouse:
        if (isInside(GfxPoint(e.mouse.x, e.mouse.y)) && !hasState(st_Disabled)) {
            /* mouse is in widget. */
            setSelectedWidget();
            if (e.mouse.b) {
                /* mouse pressed in button */
                setState(bf_Pressed, true);
            } else if (hasState(bf_Pressed)) {
                /* mouse released in button */
                setState(bf_Pressed, false);
                int arg = consumePrefixArg();
                if (hasState(bf_Key) && key != 0) {
                    putKey(key | getCurrentModifierKeys(), arg);
                }
                fire(arg);
            }
            return true;
        } else {
            /* automatically lose selection when mouse is out. */
            if (hasState(st_Selected)) {
                clearSelectedWidget();
            }

            /* let others handle the event, too. */
            return false;
        }
        break;
     default:
        break;
    }
    return UIWidget::handleEvent(e, second_pass);
}

bool
UIAbstractButton::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (cmd.code == cm_INT_QueryReferencePoint) {
        // FIXME: do we need this? This is the default setting.
        ASSERT(cmd.infop);
        TReferencePointArg& ra = *static_cast<TReferencePointArg*>(cmd.infop);
        ra.x = getExtent().x;
        ra.y = getExtent().getBottomY();
        return true;
    } else if (cmd.code == cm_INT_Push) {
        setState(bf_PressOverride, true);
        return true;
    } else if (cmd.code == cm_INT_Unpush) {
        setState(bf_PressOverride, false);
        return true;
    } else {
        return UIWidget::handleCommand(cmd, sender);
    }
}

void
UIAbstractButton::postCommand(int prefix)
{
    handleCommand(UICommand(getId(), prefix), *this);
}

void
UIAbstractButton::callStop(int)
{
    stop(id);
}

/** Reaction on button click. */
void
UIAbstractButton::fire(int arg)
{
    sig_fired.raise(arg);
}
